/*------------------------------------------------------------------------
 * MAGNET customer agent bid manager components
 * Copyright (c) 1998 Regents of the University of Minnesota.
 * All Rights Reserved by the Regents of the University of Minnesota.
 * This file contains Original Code and/or Modifications of Original
 * Code as defined in the MAGNET Public Source License, which is
 * available at http://www.cs.umn.edu/magnet/source_license_v1.txt.
 * You may not use this file except in compliance with the License.
 * $Id: PlanTWAdjuster.java,v 1.7.2.1 2002/05/05 19:41:45 ashutosh Exp $
 *------------------------------------------------------------------------
 * Revision History:
 * 
 * Sat Nov 10 16:54:41 2001  John Collins
 *   javadoc repair
 * 
 * Wed Dec 13 21:58:16 2000  John Collins
 *   New task bid plan API
 * 
 * Sun Oct  1 11:35:05 2000  John Collins
 *   Change name to PlanTWAdjuster - it's for adjusting time windows
 *   This now works with a TaskBidPlan and TaskBids, rather than the
 *   old CPMPlan and CPMTask classes, which are now deprecated.
 * 
 * Wed Sep 27 15:08:57 2000  John Collins
 *   Move to bid package
 * 
 * Mon Sep 25 11:38:53 2000  John Collins
 *   - Remove percentage calculations
 *   - Allow spec of CPM slack on a per-task basis, using a visitor
 *
 * Tue Dec  7 18:07:57 1999  John E. Collins
 *   Remove assumption that all tasks have the same duration by
 *   computing the latest early finish time as makespan rather than
 *   assuming we can derive it from the latest early start time.
 * 
 * Sun Nov 14 08:51:19 1999  John Collins
 *   Use JobInfo duration data
 * 
 * Wed Jan  6 20:21:21 1999  John E. Collins
 *   Correct calculation of PlanSlack
 *
 * Sun Jan  3 14:42:19 1999  John E. Collins
 *   Correct planSlack computation
 * 
 * Sat Jan  2 13:51:27 1999  John E. Collins
 *   Move to plan package
 *
 * Sept 29, 1998 
 *   Initial Entry
 *   Created based on evalCPM by John Collins
 */

package edu.umn.magnet.client.bid;

import edu.umn.magnet.client.eval.TaskBidPlan;
import edu.umn.magnet.client.eval.TaskBid;
import edu.umn.magnet.protocol.cust.TaskPlan;
import edu.umn.magnet.protocol.cust.SubtaskRequest;

import org.apache.log4j.Category;

import java.util.Iterator;
import java.util.Date;
import java.util.Properties;

/**
 * The methods of this class are used to adjust the time windows in a
 * "raw" plan in preparation for constructing the RFQ.  Two types of
 * adjustments are supported: overall slack, known as "plan slack",
 * and per-task slack, known as "CPM slack".  The latter is
 * accomplished by the <tt>setTimeWindows</tt> method, which must be
 * overridden in each subclass.
 *<p>
 * The underlying plan is a TaskBidPlan, which may contain some tasks
 * that are already associated with a bid.  We don't do any adjustment
 * on those.
 * </p>
 * @author John Collins
 */
public abstract class PlanTWAdjuster
{
  static Category log = Category.getInstance(PlanTWAdjuster.class.getName());

  TaskDurationAdjuster taskAdjuster = null;

  TaskBidPlan m_plan = null;
  long m_maxEarlyFinish = 0L;
  Date m_planStartDate = new Date(0L);
  Date m_planDeadline = null;

  // Amount of time to add to deadline
  long m_planSlack = 0;
  
  /**
   * Default constructor - doesn't do anything interesting
   */
  public PlanTWAdjuster () {
    super();
  }

  // ------------------- access to properties ---------------------
  /**
   * Returns the <tt>{@link TaskDurationAdjuster}</tt> instance that
   * will be used to set individual task durations.  If the adjuster
   * has not yet been set, then a new instance of TaskDurationAdjuster
   * is created and returned.
   */
  protected TaskDurationAdjuster getTaskDurationAdjuster ()
  {
    if (taskAdjuster == null)
      taskAdjuster = new TaskDurationAdjuster();
    return taskAdjuster;
  }

  protected TaskBidPlan getPlan ()
  {
    return m_plan;
  }

  /**
   * Sets the  <tt>{@link TaskDurationAdjuster}</tt> instance to be
   * used to adjust the durations of individual tasks.  If this method
   * is not called, then the default adjuster is an instance of
   * <tt>TaskDurationAdjuster</tt>, which actually does nothing to
   * task durations.
   */
  public void setTaskDurationAdjuster (TaskDurationAdjuster newTDA)
  {
    taskAdjuster = newTDA;
  }

  /**
   * Sets the time windows on tasks that are not yet bid, according to
   * specifications.
   * @param plan is a TaskBidPlan, in which at least some tasks are
   *   not yet associated with any bid.  If all tasks have bids, then
   *   this thing will accomplish exactly nothing.
   * @param startDate is a date on which the work has to start
   * @param planSlack (typically > 1.0) is the desired ratio of job time to
   *   makespan.
   */
  public void setTimeWindows(TaskBidPlan plan,
			     Date startDate,
			     double planSlack)
  {
    log.debug("setTimeWindows: ps = " + planSlack);
    m_plan = plan;
    m_planStartDate = startDate;
    
    // To produce a plan with slack, two passes are required
    // During the first pass, we are only interested in the latest
    // early finish date, so we only need a forward pass.
    for (Iterator nodes = m_plan.getTaskBids().iterator();
	 nodes.hasNext(); ) {
      ((TaskBid)nodes.next()).initCPMStart(m_planStartDate);
    }
    log.debug("CPM...Mapping Early Start");
    TaskDurationAdjuster firstAdjuster = new TaskDurationAdjuster();
    for (Iterator nodes = m_plan.getTaskBids().iterator();
	 nodes.hasNext(); ) {
      TaskBid node = (TaskBid)nodes.next();
      //log.debug("Compute early start for node " + node.getTaskId());
      computeEarlyStart(node, firstAdjuster);
    }
    
    // Second pass uses data from the tight plan to relax it.
    // Now compute the slack factor and the plan slack time
    long makespan = m_maxEarlyFinish - m_planStartDate.getTime();
    m_planDeadline = new Date(m_planStartDate.getTime()
			      + Math.round(makespan * planSlack));
    log.debug("PlanCPM: Start = "
	      + new Date(m_planStartDate.getTime())
	      + ", Earliest end = "
	      + new Date(m_planStartDate.getTime() + makespan)
	      + ", Latest end = "
	      + new Date(m_planDeadline.getTime()));
    setTW(getTaskDurationAdjuster());
    recordData();
  }

  /**
   * Sets the time windows on tasks that are not yet bid, according to
   * specifications.
   * @param plan is a TaskBidPlan, in which at least some tasks are
   *   not yet associated with any bid.  If all tasks have bids, then
   *   this thing will accomplish exactly nothing.
   * @param startDate is a Date on which the work has to start
   * @param deadline is a Date on which the work has to be completed
   */
  public void setTimeWindows(TaskBidPlan plan,
			     Date startDate,
			     Date deadline)
  {
    log.debug("setTimeWindows: deadline = " + deadline);
    m_plan = plan;
    m_planStartDate = startDate;
    m_planDeadline = deadline;
    log.debug("PlanCPM: Start = "
	      + new Date(m_planStartDate.getTime())
	      + ", Latest end = "
	      + new Date(m_planDeadline.getTime()));
    setTW(getTaskDurationAdjuster());
    recordData();
  }

  /**
   * Sets the {@link edu.umn.magnet.protocol.cust.SubtaskRequest
   * SubtaskRequest} time windows according to a subclass-determined
   * method.  This method must be overridden in all subclasses.
   */
  protected abstract void setTW (TaskDurationAdjuster adj);

  // ------------ common methods for subclasses ----------------

  /**
   * Updates the start/finish times in the node graph.
   */
  protected void updateCPM (TaskDurationAdjuster adjuster)
  {
    log.debug("updateCPM");

    // 1. Clear out the existing start & finish times. Set the start
    // time of each root node to the start time of the plan, and the
    // end time of each leaf node to the end time of the plan.
    for (Iterator nodes = m_plan.getTaskBids().iterator();
	 nodes.hasNext(); ) {
      TaskBid node = (TaskBid)nodes.next();
      node.initCPMStart(m_planStartDate);
      node.initCPMFinish(m_planDeadline);
    }

    // 2. Map through the nodes in any order. For each node,
    // recursively set the early start times of each of the parents, then
    // set the start time of the child. Also, recursively set the late
    // finish time of each child, then the late finish time of the
    // parent.
    log.debug("CPM...Mapping Early Start");
    for (Iterator nodes = m_plan.getTaskBids().iterator();
	 nodes.hasNext(); ) {
      TaskBid node = (TaskBid)nodes.next();
      computeEarlyStart(node, adjuster);
    }

    log.debug("CPM...Mapping Late Finish");
    for (Iterator nodes = m_plan.getTaskBids().iterator();
	 nodes.hasNext(); ) {
      TaskBid node = (TaskBid)nodes.next();
      
      computeLateFinish(node, adjuster);
      if (log.isInfoEnabled()) {
	Date es = node.getEarlyStartCPM();
	Date fs = node.getLateFinishCPM();
	log.info(node.getTask().getId() + ": "
		 + es.toString() + "->" + fs.toString());
      }
    }
  }

  /**
   * Recursively computes EarlyStart time of the <tt>node</tt>.  Early
   * Start is given by the <tt>earlyStartCPM</tt> property if it's
   * non-null, otherwise as the latest early finish time of the
   * predecesors of <tt>node</tt>, or the start time of the plan if
   * there are no predecessors.  As a side effect, the early start
   * times <tt>node</tt> and of all nodes in the transitive closure of
   * the precedence relation of <tt>node</tt> will also be computed
   * and stored in the <tt>earlyStartCPM</tt> property.
   * @param node a {@link TaskBid} for which we wish to compute the early
   *   start time.
   * @param adjuster a {@link TaskDurationAdjuster} that will be used
   *   to adjust the duration of nodes.
   * @return the early finish time of <tt>node</tt>.  */
  protected Date computeEarlyStart(TaskBid node,
				 TaskDurationAdjuster adjuster)
  {
    // This actually computes and records early start, and returns
    // the early finish time.  That is what allows the simple
    // recursive operation you see here.
    Date es = node.getEarlyStartCPM();
    if (es == null) {
      // this task needs work...
      // The early start of a node is the later of the latest early
      // finish among the predecessors of node and the start date of the
      // whole plan

      es = m_planStartDate;

      for (Iterator predecessors = m_plan.findPredecessors(node.getTask());
	   predecessors.hasNext(); ) {
	Date next = computeEarlyStart
	               (m_plan.findTaskBid
			 ((String)predecessors.next()),
			adjuster);
	if (next.after(es))
	  es = next;
      }

      // If the node has a bid or a specified early start time that is
      // later than es, we use that instead.
      Date bid = node.getEarlyStart();
      if (bid != null && bid.after(es)) {
	//log.debug("Bid Early Start = " + bid);
	es = bid;
      }
      // Now we have the "real" early start date, so we record it.
      node.setEarlyStartCPM(es);
    }
    //Update the maximum early finish date.
    long earlyFinish = es.getTime() + node.getAdjustedDuration(adjuster);
    if (earlyFinish > m_maxEarlyFinish) {
      //log.debug("Max Early Finish = " + new Date(earlyFinish));
      m_maxEarlyFinish = earlyFinish;
    }
    return new Date(es.getTime() + node.getAdjustedDuration(adjuster));     
  }

  /**
   * Recursively computes LateFinish's of nodes in the order of succession
   */
  protected Date computeLateFinish(TaskBid node,
				 TaskDurationAdjuster adjuster)
  {
    // This actually computes and records late finish, and returns
    // the late start time.  That is what allows the simple
    // recursive operation you see here.
    Date lf = node.getLateFinishCPM();
    if (lf == null) {
      // this one needs work...
      // The late finish of a node is the earlier of the earliest late
      // start among the successors of node and the plan late finish date = 
      // max(EarlyStartTime of all tasks)+TaskDuration+slack.

      if (m_planDeadline == null) {
	lf = new Date(m_maxEarlyFinish);
	m_planDeadline = lf;
      } else {
	lf = m_planDeadline;
      }

      for (Iterator successors =
	       m_plan.findSuccessors((SubtaskRequest)node.getTask());
	   successors.hasNext(); ) {
	Date next = computeLateFinish
	               (m_plan.findTaskBid
			 ((String)successors.next()),
			adjuster);
	if (next.before(lf))
	  lf = next;
      }

      // If the node has a bid or a specified late finish, we don't
      // set a later late finish time.
      Date bid = node.getLateFinish();
      if (bid != null && bid.before(lf))
	lf = bid;

      // Now we have the "real" late finish date, so we record it.
      node.setLateFinishCPM(lf);
    }
    return new Date(lf.getTime() - node.getAdjustedDuration(adjuster));      
  }

  /**
   * For those tasks that do not have bids, copies the time window
   * recorded in the TaskBid into the SubtaskRequest.
   */
  private void recordData ()
  {
    // Now we modify the actual SubtaskRequest data for those tasks
    // that have no bids.
    for (Iterator nodes = m_plan.getTaskBids().iterator();
	 nodes.hasNext(); ) {
      TaskBid node = (TaskBid)nodes.next();
      if (node.getBid() == null) {
	log.debug("Node " + node.getTask().getId()
		  + " has no bid: setting time window.");
	Date es = node.getEarlyStartCPM();
	Date fs = node.getLateFinishCPM();
	try {
	  node.getTask().setEarliestTimeToStart(es);
	  node.getTask().setDeadline(fs);
	} 
	catch (Exception e) {
	  e.printStackTrace();
	}
      }
    }
  }

  // --------------------- static stuff ------------------------
  /**
   * Static factory for plan time window adjusters and subclasses.
   * @param props A properties object in which the setup properties
   *   are expected to reside.
   */
  public static PlanTWAdjuster generate (Properties props)
  {
    String adjusterType = props.getProperty("PlanTWA.Type", null);
    log.info("prop: Plan TW Adjuster type = "
	     + adjusterType);
    if (adjusterType == null) {
      log.warn("Adjuster type not specified.  Using default");
      return new CPMTWAdjuster();
    }
    else {      
      try {
	String classname = "edu.umn.magnet.client.bid." + adjusterType;
	Class target = Class.forName(classname);
	PlanTWAdjuster result
	    = (PlanTWAdjuster)target.newInstance();
	//result.init(props);
	return result;
      }
      catch (ClassNotFoundException cnf) {
	log.error("specified adjuster not found.  Using default");
	return new CPMTWAdjuster();
      }
      catch (InstantiationException ie) {
	log.error("specified adjuster can't be instantiated.  Using default");
	return new CPMTWAdjuster();
      }
      catch (IllegalAccessException iae) {
	log.error("Illegal access exception.  Using default");
	return new CPMTWAdjuster();
      }
    }
  }
}
