/*------------------------------------------------------------------------
 * MAGNET Customer Agent Bid Management components
 * $Id: SchedFwdBySteps.java,v 1.3.2.1 2002/05/05 19:41:46 ashutosh Exp $
 * Copyright (c) 2000 Regents of the University of Minnesota.
 * All Rights Reserved by the Regents of the University of Minnesota.
 * This file contains Original Code and/or Modifications of Original
 * Code as defined in the MAGNET Public Source License, which is
 * available at http://www.cs.umn.edu/magnet/source_license_v1.txt.
 * You may not use this file except in compliance with the License.
 *------------------------------------------------------------------------
 * Revision History:
 * 
 * Fri Sep 29 09:31:54 2000  John Collins
 *   Initial Entry
 */

package edu.umn.magnet.client.bid;

import edu.umn.magnet.protocol.cust.TaskPlan;
import edu.umn.magnet.protocol.cust.SubtaskRequest;
import edu.umn.magnet.client.eval.TaskBidPlan;

import edu.umn.magnet.util.time.AgentTime;

import org.apache.log4j.Category;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.HashSet;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.LinkedList;
import java.util.Properties;

/**
 * Instances are responsible for evaluating a bidding situation and
 * producing a set of BidAgendaItem instances for controlling the
 * actual bidding process.  This version splits the bidding process
 * into a specified number of approximately equal steps, by finding
 * the longest path through the precedence network, splitting that
 * into steps, and then using the count of tasks per step along that
 * path as the maximum number of successive tasks for other paths,
 * with the exception that no task will be bid unless all predecessors
 * are either in the same RFQ or in a previous one.
 * @author John Collins
 */
public class SchedFwdBySteps extends BidScheduler
{
  static Category log = Category.getInstance(SchedFwdBySteps.class.getName());

  // The number of steps is an instance variable
  private int m_nsteps = 1;

  /**
   * Default constructor doesn't need to do much.
   */
  public SchedFwdBySteps ()
  {
    super();
  }

  /**
   * All BidScheduler subclasses must implement an init method that
   * takes a set of Properties.  The default BidScheduler does not use
   * any properties. 
   */
  public void init (Properties props)
  {
    try
    {
      m_nsteps
	  = (int)Math.round((Double.valueOf
			     (props.getProperty("BidScheduler.nsteps", "1")))
			    .doubleValue());
      log.info("prop: Scheduling Steps = " + m_nsteps);
    }
    catch (NumberFormatException nfe)
    {
      log.error("Format error in property file. Using default value for nsteps");
    }
  }

  /**
   * This type of scheduler walks the precedence network from the
   * roots, assembling tasks into "layers".  The layers are then split
   * up into bid sets.
   */
  public List scheduleBids (TaskBidPlan tbPlan)
  {
    LinkedList agenda = new LinkedList();
    Date start = AgentTime.getDate();

    // Compute the length of bids
    TaskPlan taskPlan = tbPlan.getTaskPlan();
    HashMap task2layer = new HashMap();
    int longestPath = findLayers(taskPlan, task2layer);
    int bidLength = (longestPath + 1)/m_nsteps;
    if (bidLength == 0)
      bidLength = 1;
    log.debug("SchedFwdBySteps: bid length = " + bidLength);

    // Now we start allocating tasks to bid cycles
    // The tasks2layers map tells us which layer each task is in.
    for (int bidNumber = 0; bidNumber < m_nsteps; bidNumber ++) {
      HashSet tasksToBid = new HashSet();
      for (int layer = (bidNumber * bidLength);
	   layer < ((bidNumber + 1) * bidLength);
	   layer++) {
	for (Iterator tasks = task2layer.keySet().iterator();
	     tasks.hasNext(); ) {
	  SubtaskRequest task = (SubtaskRequest)tasks.next();
	  if (((Integer)task2layer.get(task)).intValue() == layer) {
	    tasksToBid.add(task);
	    tasks.remove();
	  }
	}
      }
      // Now, tasksToBid contains the actual list of tasks to be
      // included in the next bid.
      if (log.isDebugEnabled()) {
	StringBuffer buf = new StringBuffer();
	buf.append("bid #").append(bidNumber).append(":");
	for (Iterator tasks = tasksToBid.iterator();
	     tasks.hasNext(); ) {
	  buf.append(" ").append(((SubtaskRequest)tasks.next()).getId());
	}
	buf.append(".");
	log.debug(buf.toString());
      }
      // We punt on the time allocation - 10 min each for bidding & eval
      // Eventually this will probably
      // come from a Properties instance or something.
      long chunk = 600000L; // 10 minutes
      Date bidDeadline = new Date(start.getTime() + chunk);
      Date earlyAward = new Date(bidDeadline.getTime());
      Date awardDeadline = new Date(earlyAward.getTime() + chunk);
      Date endOfSlot = new Date(awardDeadline.getTime());
      start = endOfSlot; // start time for next cycle
      
      agenda.add(new BidCommand(tasksToBid,
				bidDeadline, earlyAward, awardDeadline,
				endOfSlot));
    }
    return agenda;
  }

  /**
   * Finds the length of the longest path in the precedence network.
   * This method is not currently used.
   */
  private int findLongestPath (TaskPlan taskPlan)
  {
    LinkedList searchQueue= new LinkedList();
    int maxDepth = 0;

    // First, populate the queue with all the roots
    Iterator roots = taskPlan.getSuccessorRoots();
    while (roots.hasNext()) {
      SubtaskRequest thisTask = (SubtaskRequest)roots.next();
      searchQueue.add(new PathNode(1, thisTask));
    }
    
    // Next, process the queue
    while (!searchQueue.isEmpty())
    {
      PathNode node = (PathNode)searchQueue.remove(0);
      if (node.depth > maxDepth)
	maxDepth = node.depth;
      Iterator successors = taskPlan.getSuccessors(node.task);
      while (successors.hasNext()) {
	SubtaskRequest thisTask = (SubtaskRequest)successors.next();
	searchQueue.add(new PathNode(node.depth + 1, thisTask));
      }
    }
    log.debug("max depth = " + maxDepth);
    return maxDepth;
  }

  /**
   * Walks the precedence network breadth-first and records the layer
   * in which each task is found.
   * @param taskPlan the input TaskPlan instance.
   * @param result a HashMap instance in which to record the results.
   *   Entries associate a SubtaskRequest with the highest layer in
   *   which the task appears.
   * @param result a List instance in which to record the results.
   *   Each entry is a PathNode instance which gives the task and its
   *   layer number.  Entries are sorted by layer, with the first
   *   layer first.  Task that appear in multiple layers will be
   *   listed multiple times.
   */
  private int findLayers (TaskPlan taskPlan, HashMap result)
  {
    LinkedList searchQueue= new LinkedList();
    Integer layer = new Integer(0);

    // First, populate the queue with all the roots
    for (Iterator roots = taskPlan.getSuccessorRoots();
	 roots.hasNext(); ) {
      SubtaskRequest thisTask = (SubtaskRequest)roots.next();
      searchQueue.add(new PathNode(0, thisTask));
      result.put(thisTask, layer);
    }
    
    // Next, process the queue
    while (!searchQueue.isEmpty())
    {
      PathNode node = (PathNode)searchQueue.remove(0);
      if (node.depth == layer.intValue()) {
	// here's where we notice we've moved to the next layer
	layer = new Integer(node.depth + 1);
      }
      for (Iterator successors = taskPlan.getSuccessors(node.task);
	   successors.hasNext(); ) {
	SubtaskRequest thisTask = (SubtaskRequest)successors.next();
	searchQueue.add(new PathNode(node.depth + 1, thisTask));
	result.put(thisTask, layer);
      }
    }
    if (log.isDebugEnabled()) {
      for (Iterator tasks = result.keySet().iterator();
	   tasks.hasNext(); ) {
	SubtaskRequest task = (SubtaskRequest)tasks.next();
	log.debug("Task " + task.getId() + ": layer "
		  + result.get(task));
      }
    }
    return layer.intValue();
  }

  // Local data structure class
  class PathNode
  {
    int depth;
    SubtaskRequest task;
    
    PathNode (int depth, SubtaskRequest task)
    {
      this.depth = depth;
      this.task = task;
    }
  }
}
