/*----------------------------------------------------------------------------
 * Magnet Bid Evaluation components
 * $Id: BidtreeNode.java,v 1.11.2.2 2002/05/05 19:41:47 ashutosh Exp $
 * Copyright (c) 1998 Regents of the University of Minnesota.
 * All Rights Reserved by the Regents of the University of Minnesota.
 * This file contains Original Code and/or Modifications of Original
 * Code as defined in the MAGNET Public Source License, which is
 * available at http://www.cs.umn.edu/magnet/source_license_v1.txt.
 * You may not use this file except in compliance with the License. 
 *----------------------------------------------------------------------------
 * Revision History:
 * 
 * Tue Feb 19 20:06:30 2002  John Collins
 *   Initial Entry
 */

package edu.umn.magnet.client.eval;



import edu.umn.magnet.protocol.cust.Bid;
import edu.umn.magnet.protocol.cust.SubtaskRequest;
import edu.umn.magnet.protocol.cust.TaskPlan;
import edu.umn.magnet.protocol.cust.TaskWithPrecedence;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.NoSuchElementException;
import java.util.Set;
import java.util.Stack;
import java.util.TreeSet;
import java.util.Vector;
import org.apache.log4j.Category;

/**
 * Instances of BidtreeNode serve to wrap TaskBidPlan instances with
 * attributes and behavior to support queue-based search among
 * alternative (possibly partial) solutions represented by the
 * TaskBidPlan.  There is a one-to-one correspondence between a
 * BidtreeNode and a TaskBidPlan.
 * <p>
 * Two constructors are provided. One is for creating a root node with
 * no bids, and the other is for creating a deep clone of an existing
 * node, then modifying it to create a new search node.
 * No default constructor is provided.
 * </p><p>
 * Since this structure is intended to support A* search, we keep
 * track of information necessary to compute the heuristic value for
 * each node. This includes a mapping of un-allocated tasks to
 * minimum-cost bids, and a list of bids that are compatible with the
 * bids currently mapped in the TaskBidPlan.
 * </p>
 * @author John Collins
 */
public class BidtreeNode implements edu.umn.magnet.util.search.SearchNode
{
  // Init the log category
  static Category log = Category.getInstance(BidtreeNode.class.getName());

  // Basic database
  private TaskBidPlan m_tbp = null;
  private BidtreeSearch m_bts = null;
  private int[] m_mask = null;
  private ArrayList m_nextBuckets = null;
  private BtBidWrapper[] m_nextBids = null;
  private int m_nextBidIndex = 0;

  // This is the list of buckets from which bids can be chosen. We
  // don't actually use this to pick bids, but rather to calculate the
  // heuristic evaluation.
  private ArrayList m_openBuckets = null;

  // Heuristic evaluation support
  private double[] m_contributions = null;
  private double m_hvalue = 0.0;

  // Place in the hierarchy
  private int m_nodeNumber = 0;

  // Evaluation results
  private boolean m_feasible = false;
  private boolean m_covered  = false;
  //private HashMap m_evaluations = new HashMap();

  /**
   * Creates a root node by creating one of these with a TaskBidPlan.
   * In the process, all the "empty" TaskBids (those without bids)
   * become uncovered tasks, and all the current Bids get locked.
   */
  public BidtreeNode(TaskBidPlan tbPlan, BidtreeSearch bts)
  {
    super();
    m_tbp = tbPlan;
    m_bts = bts;

    // Create the bidlist for this node
    Bidtree bt = bts.getBidtree();
    TaskDescriptor[] tds = bt.getTaskArray();
    m_mask = new int[tds.length];
    m_mask[0] = Bidtree.IN;
    for (int i = 1; i < m_mask.length; i++) {
      m_mask[i] = Bidtree.BOTH;
    }
    if (bts.isUseSortedBids()) {
      m_nextBids = bt.getBids(m_mask);
    }
    else {
      m_nextBuckets = bt.getBidCollections(m_mask);
    }

    // Initialize the compatibleBids list to all the bids
    m_openBuckets = bt.getBuckets();

    // Create the heuristic map and compute its value - We'll assume
    // there are no excluded bids here.
    m_contributions = new double[tds.length];
    for (int i = 0; i < tds.length; i++) {
      BtBidWrapper theBid = bt.getBids(i)[0];
      double c = theBid.getCostPerTask();
      m_hvalue += c;
      m_contributions[i] = c;
    }
  }

  /**
   * Creates a new BidtreeNode by copying the guts of an old one. 
   * We create a new TaskBidPlan that is a copy of the current one,
   * add the given bid to it, and then retrieve the set of bids to be
   * used to expand this node. This is intended to be called from the
   * public expandNode method.
   */
  protected BidtreeNode (BidtreeNode oldNode,
			 BtBidWrapper newBid,
			 int nodeCount)
  {
    super();
    m_bts = oldNode.m_bts;
    m_nodeNumber = nodeCount;

    // initialize the mask
    m_mask = new int[oldNode.m_mask.length];
    System.arraycopy(oldNode.m_mask, 0, m_mask, 0, m_mask.length);
    Bidtree bt = m_bts.getBidtree();
    int[] tasks = newBid.getTaskIndices();
    for (int i = 0; i < tasks.length; i++) {
      m_mask[tasks[i]] = Bidtree.OUT;
    }

    // find the first uncovered task and make it IN
    int index = 0;
    boolean foundUncovered = false;
    // first, skip past all the OUT tasks
    while (index < m_mask.length && m_mask[index] == Bidtree.OUT) {
      index += 1;
    }
    // if we didn't run off the end of the mask, the next one is
    // IN. Otherwise we've achieve coverage.
    if (index < m_mask.length) {
      m_mask[index] = Bidtree.IN;
      foundUncovered = true;
    }
    
    if (foundUncovered) {
      // use the mask to retrieve the bidlist for this node
      if (m_bts.isUseSortedBids()) {
	m_nextBids = bt.getBids(m_mask);
      }
      else {
	m_nextBuckets = bt.getBidCollections(m_mask);
	unpackNextBucket();
      }

      if (m_nextBids == null || m_nextBids.length == 0) {
	//log.debug("No further bids found: pruning " + nodeCount);
	m_tbp = null;
	return;
      }

      // Now we must update the heuristic value.
      m_hvalue = oldNode.m_hvalue;
      m_contributions = new double[m_mask.length];
      System.arraycopy(oldNode.m_contributions, 0,
		       m_contributions, 0, m_mask.length);
      // first, we remove the contributions from tasks covered by the
      // new bid.
      int[] theseTasks = newBid.getTaskIndices();
      //log.debug("Original h() = " + m_hvalue);
      for (int i = 0; i < theseTasks.length; i++) {
	int taskI = theseTasks[i];
	m_hvalue -= m_contributions[taskI];
	m_contributions[taskI] = 0.0;
      }
      //log.debug("First intermediate h() value = " + m_hvalue);
      // Next, remove the bucket of this bid and all conflicting
      // buckets from the open buckets list, and mark the missing
      // contributions. Note that it's not possible that any bids
      // already mapped in this node can conflict with the new bid.
      m_openBuckets = new ArrayList(oldNode.m_openBuckets);
      m_openBuckets.remove(bt.getBucket(newBid.getBucketIndex()));
      int[] conflicts
	  = bt.getBucket(newBid.getBucketIndex()).getConflicts();

//        if (log.isDebugEnabled()) {
//    	StringBuffer sb = new StringBuffer();
//    	sb.append("Conflict set[");
//    	String sep = "";
//    	for (int k = 0; k < conflicts.length; k++) {
//    	  sb.append(sep).append(bt.getBucket(conflicts[k]).toString());
//    	  sep = ",";
//    	}
//    	sb.append("]");
//    	log.debug(sb.toString());
//        }

      for (int i = 0; i < conflicts.length; i++) {
	BtBucket thisBucket = bt.getBucket(conflicts[i]);
	m_openBuckets.remove(thisBucket);
	int[] thoseTasks = thisBucket.getTaskIndices();
	for (int j = 0; j < thoseTasks.length; j++) {
	  int taskJ = thoseTasks[j];
	  if (m_contributions[taskJ] != -1.0) {
	    m_hvalue -= m_contributions[taskJ];
	    m_contributions[taskJ] = 0.0;
	  } // if m_contr...
	} // for j	
      } // for i

//        if (log.isDebugEnabled()) {
//  	//  	log.debug("Second intermediate h() value = " + m_hvalue);
//    	log.debug("Open buckets:");
//    	for (Iterator obs = m_openBuckets.iterator(); obs.hasNext(); ) {
//    	  log.debug("  " + (BtBucket)obs.next());
//    	}
//        }

      // Now we mark the contributions of the new bid. Note that
      // m_hvalue isn't just the sum of the contributions, just the
      // sum of the uncovered contributions.  All the covered ones are
      // marked in the contributions array with a value of -1.0. When
      // that's done, all the 0.0 entries in contributions will
      // represent tasks that need new estimates.
      for (int i = 0; i < theseTasks.length; i++) {
	m_contributions[theseTasks[i]] = -1.0;
      } // for i
      // Next we walk the contributions array and fill in all the 0.0
      // entries with values from the openBuckets list. Note that this
      // breaks if you allow any bid to have a cost/task of exactly
      // 0.0.
      for (int i = 0; i < m_contributions.length; i++) {
	if (m_contributions[i] == 0.0) {
	  BtBidWrapper[] candidates = bt.getBids(i);

//    	  if (log.isDebugEnabled()) {
//    	    StringBuffer sb = new StringBuffer();
//    	    sb.append("Candidates for task " + i);
//    	    sb.append(" [");
//    	    String sep = "";
//    	    for (int k = 0; k < candidates.length; k++) {
//    	      sb.append(sep).append(candidates[k].getBid().getId());
//    	      sep = ",";
//    	    }
//    	    sb.append("]");
//    	    log.debug(sb.toString());
//    	  }

	  // candidates are sorted by cost/task
	  for (int j = 0;
	       (j < candidates.length
		&& m_contributions[i] == 0.0);
	       j++) {
	    BtBucket candidate
		= bt.getBucket(candidates[j].getBucketIndex());
	    if (m_openBuckets.contains(candidate)) {
	      double cont = candidates[j].getCostPerTask();
	      m_contributions[i] = cont;
	      m_hvalue += cont;
//  	      log.debug("Adding contribution " + cont
//  			+ " for task " + i
//  			+ " from bid " + candidates[j].getBid().getId()
//  			+ "." + candidates[j].getBucketIndex());
	    } // if m_openBuckets.contains
	  } //for j
	  // At this point, if we still have a contribution of 0.0, we
	  // know we can't achieve coverage on this path.
	  if (m_contributions[i] == 0.0) {
	    //log.debug("No coverage on this path");
	    m_tbp = null;
	    return;
	  }
	} // if m_contributions[i]
      } // for i
    }
    else {
      // this node is now covered. No further expansions needed
      //log.debug("Covered node found " + nodeCount);
    } // if foundUncovered

    // Attempt to add the bid to the new TBP
    m_tbp = new TaskBidPlan(oldNode.m_tbp);
    Set bidsToRemove = m_tbp.addBid(newBid.getBid());
    if (bidsToRemove != null && bidsToRemove.size() > 0) {
      // We have a problem here! Just complain and prune
      log.error("Unexpected bid displacement with " + newBid);
      m_tbp = null;
      return;
    }
  }
  
  public String toString ()
  {
    return ("SN" + m_nodeNumber + "(" + m_tbp.toString() + ")");
  }
  
  /**
   * Expands this node by popping bids from the nextBids list and
   * attempting to create new nodes until we either succeed or run out
   * of choices.
   */
  public BidtreeNode expandNode (int nodeCount)
  {
    BidtreeNode result = null;
    while (hasMoreBids() && result == null) {
      BtBidWrapper aBid = popNextBid();
      // have to create a new BidtreeNode & TaskBidPlan, then add the bid
      // to the new node
      //log.debug("expanding node " + m_nodeNumber + " with bid "
      //	+ aBid.getBid().getId() + ", producing node " + nodeCount);
      result = new BidtreeNode(this, aBid, nodeCount);
      if (result.m_tbp == null) {
	// This bid could not be used, ditch the rest of the bucket if
	// we are picking bids from their buckets
	result = null;
	if (m_nextBuckets != null) {
	  unpackNextBucket();
	}
      }
    }
    return result;
  }
  
  // ============== Evaluation results ====================
  //
  //Note that none of these are of any real use until the evaluator has
  //been run.
  
  /**
   * True just in case the instantiated plan is temporally feasible.
   * If the plan is also covered, this means that the set of bids
   * represented by this TaskBidPlan can be accepted to form a
   * feasible set of bids.  If the plan is not covered, it means that
   * the existing subset of bids is temporally consistent with each
   * other and with the original RFQ.
   */
  public boolean isFeasible()
  {
    return m_tbp.isFeasible(); 
  }
  
  /**
   * True just in case all subtasks in this plan are covered by bids.
   */
  public boolean isCovered ()
  {
    return m_tbp.isCovered();
  }

  /**
   * Returns the total cost of the bids that are part of this plan.
   */
  public long getCost ()
  {
    return m_tbp.getCost();
  }

  /**
   * Returns the heuristic evaluation for the uncovered tasks of this
   * node.
   */
  public double getHeuristic ()
  {
    return m_hvalue;
  }

  /**
   * Returns the evaluation of this plan.  This only has meaning after
   * the plan has been evaluated
   */
  public double getEvaluation ()
  {
    return m_tbp.getEvaluation() + m_hvalue;
  }

  /**
   * The value method is a synonym for getEvaluation.  This is the
   * implementation of the BidtreeNode interface.
   */
  public double value ()
  {
    return m_tbp.getEvaluation() + m_hvalue;
  }

  /**
   * Return the TaskPlan underlying this plan.
   */
  public TaskPlan getTaskPlan()
  {
    return m_tbp.getTaskPlan();
  }

  // ================== Model access ==========================

  /**
   * Returns the TaskBidPlan associated with this node.
   */
  public TaskBidPlan getTaskBidPlan ()
  {
    return m_tbp;
  }

  /**
   * Returns the Collection of TaskBid instances from the underlying
   * TaskBidPlan. 
   */
  public Collection getTaskBids ()
  {
    return m_tbp.getTaskBids();
  }

  /**
   * Removes and returns the next bid from the list of candidates for
   * this node. You get a null back if there are no more candidate
   * bids.
   */
  public BtBidWrapper popNextBid ()
  {
    BtBidWrapper result = null;
    if (hasMoreBids()) {
      if (m_nextBids == null
	  || m_nextBidIndex >= m_nextBids.length)
	unpackNextBucket();
      result = m_nextBids[m_nextBidIndex++];
      //if (result == null) {
      //  unpackNextBucket();
      //  result = m_nextBids[m_nextBidIndex++];
      //}
    }
    return result;
  }

  /**
   * True just in case the candidate-bids list is non-empty.
   */
  public boolean hasMoreBids ()
  {
    return ((m_nextBids != null
	     && m_nextBidIndex < m_nextBids.length)
	    || (m_nextBuckets != null
		&& m_nextBuckets.size() > 0));
  }

  /**
   * Unpacks the next bid bucket in case we are not using sorted bids
   */
  private void unpackNextBucket ()
  {
    m_nextBids = null;
    if (m_nextBuckets != null && m_nextBuckets.size() > 0) {
      BtBucket bucket = (BtBucket)m_nextBuckets.remove(0);
      m_nextBids = bucket.getSortedBids();
      m_nextBidIndex = 0;
      //log.debug("Unpacked bucket: " + m_nextBids.length);
    }
  }

  /**
   * Returns the node number for this node
   */
  public int getNodeNumber ()
  {
    return m_nodeNumber;
  }

  /**
   * Sets the search-type and search runtime in the underlying TBP.
   */
  protected void setSearchData (String method, long time)
  {
    m_tbp.setSearchData(method, time);
  }

  protected void setSearchData (String method, long time, int nodeCount)
  {
    m_tbp.setSearchData(method, time, nodeCount);
  }

  // ============== Test access ==============
  ArrayList getNextBidIds ()
  {
    ArrayList result = new ArrayList();
    for (int i = m_nextBidIndex; i < m_nextBids.length; i++) {
      result.add(m_nextBids[i].getBid().getId());
    }
    return result;
  }
}
