/*------------------------------------------------------------------------
 * MAGNET Bid Evaluation component
 * $Id: BtBidWrapper.java,v 1.1.2.2 2002/05/05 19:41:47 ashutosh Exp $
 * Copyright (c) 2002 Regents of the University of Minnesota.
 * All Rights Reserved by the Regents of the University of Minnesota.
 * This file contains Original Code and/or Modifications of Original
 * Code as defined in the MAGNET Public Source License, which is
 * available at http://www.cs.umn.edu/magnet/source_license_v1.txt.
 * You may not use this file except in compliance with the License.
 *------------------------------------------------------------------------
 * Revision History:
 * 
 * Fri Mar  8 15:25:11 2002  John Collins
 *   Initial Entry
 */

package edu.umn.magnet.client.eval;

import edu.umn.magnet.protocol.cust.Bid;
import org.apache.log4j.Category;

import java.util.Set;

/**
 * This class wraps Bid with preprocessed data to speed up Bidtree
 * searches. 
 * @author John Collins
 * @version $Revision: 1.1.2.2 $
 */
public class BtBidWrapper
{
  // Init the log category
  static Category log = Category.getInstance(BtBidWrapper.class.getName());

  /** We don't allow cost/task to be exactly 0.0. This is the
   * substitute value if that happens */
  public static final double MIN_COST_PER_TASK = 1.0e-12;
  // This is the bid we are wrapping
  Bid bid = null;

  // Task coverage is given in terms of indices into the sorted task
  // list generated by the Bidtree abstraction.
  int [] tasks;

  // This is the bucket index, used to find conflict sets
  int bucketIndex = 0;

  // Heuristic evaluation depends on cost/task data
  double costPerTask = 0.0;

  /**
   * We provide a package-accessible constructor to restrict usage of
   * this thing. Note that you can only build these after the sorted
   * task list has been generated in the Bidtree.
   */
  BtBidWrapper (Bid bid, Bidtree bt)
  {
    super();
    //log.debug("Creating bid wrapper for " + bid.getId());
    this.bid = bid;

    // retrieve the ordered list of TaskDescriptors from the BidTree
    TaskDescriptor[] tl = bt.getTaskArray();

    // retrieve the Set of task ID strings from the Bid.
    Set ts = bid.getTaskSet();
    tasks = new int[ts.size()];

    int j = 0;
    for (int i = 0; i < tl.length; i++) {
      if (ts.contains(tl[i].getTaskId()))
	tasks[j++] = i;
    }
    costPerTask = (double)bid.getCost() / (double)ts.size();
    if (costPerTask == 0.0) {
      log.info("Substituting MIN_COST_PER_TASK for bid "
	       + bid.getId());
      costPerTask = MIN_COST_PER_TASK;
    }
  }

  /**
   * Returns the wrapped Bid.
   */
  public Bid getBid ()
  {
    return bid;
  }

  /**
   * Returns the task indices of the tasks covered by this bid as an
   * array of int. The tasks are indexed by the Bidtree.
   */
  public int[] getTaskIndices ()
  {
    return tasks;
  }

  /**
   * Returns the cost per task of the underlying bid.
   */
  public double getCostPerTask ()
  {
    return costPerTask;
  }

  /**
   * Returns the total cost of the underlying bid.
   */
  public double getCost ()
  {
    return (double)bid.getCost();
  }

  /**
   * Sets the bucket index for this bid. This is only meaningful if
   * this bid wrapper has been indexed in a Bidtree.
   */
  public void setBucketIndex (int index)
  {
    bucketIndex = index;
  }

  /**
   * Returns the bucket index for this bid. Using this index, you can
   * find the BtBucket containing this bid from your
   * Bidtree.
   */
  public int getBucketIndex ()
  {
    return bucketIndex;
  }
}
