/*------------------------------------------------------------------------
 * Magnet Bid Evaluation components
 * $Id: CPMeval.java,v 1.2.2.3 2002/05/05 19:41:48 ashutosh Exp $
 * Copyright (c) 1998 Regents of the University of Minnesota.
 * All Rights Reserved by the Regents of the University of Minnesota.
 * This file contains Original Code and/or Modifications of Original
 * Code as defined in the MAGNET Public Source License, which is
 * available at http://www.cs.umn.edu/magnet/source_license_v1.txt.
 * You may not use this file except in compliance with the License.
 *------------------------------------------------------------------------
 * Revision History:
 * 
 * Wed Dec 13 21:56:39 2000  John Collins
 *   Get rid of bid breakdowns, support new bid semantics.
 * 
 * Wed Jan 13 21:22:03 1999  John E. Collins
 *   Don't invoke feasibility penalty twice!
 */

package edu.umn.magnet.client.eval;

import org.apache.log4j.Category;

import java.util.Iterator;
import java.util.Date;

/**
 * Evaluator that runs the Critical Path Method algorithm
 * @author John Collins
 */
public class CPMeval implements Evaluator
{
  // Init the log categories
  static Category log = Category.getInstance(CPMeval.class.getName());

  static private final String s_type = "Temporal";

  TaskBidPlan m_plan = null;
  BidEvaluator m_model;

  public CPMeval () 
  {
    super();
  }

  public CPMeval (BidEvaluator model) 
  {
    super();
    m_model = model;
  }

  /**
     Returns the cost factor due to temporal infeasiblity in the plan.
     As a side effect, the early start and late finish times are
     computed or updated, and the isFeasible flag is set on the plan.
  */
  public double evaluate (TaskBidPlan plan) 
  {
    m_plan = plan;
    updateCPM();
    long negSlack = getNegativeSlack();
    if (negSlack == 0L) {
      plan.setFeasible(true);
      return 0.0;
    }
    else {
      plan.setFeasible(false);
      return (double)negSlack;
    }
  }

  /**
   * Return the evaluation class.
   */
  public String getEvalClass () 
  {
    return s_type;
  }

  /**
   *  Updates the start/finish times in the node graph.
   */
  public void updateCPM () 
  {

    // 1. Clear out the existing start & finish times. Set the start
    // time of each root node to the start time of the plan, and the
    // end time of each leaf node to the end time of the plan.
    for (Iterator nodes = m_plan.getTaskBids().iterator();
	 nodes.hasNext(); ) {
      ((TaskBid)nodes.next()).initCPM();
    }

    // 2. Map through the nodes in any order. For each node,
    // recursively set the early start times of each of the parents, then
    // set the start time of the child. Also, recursively set the late
    // finish time of each child, then the late finish time of the
    // parent.
    for (Iterator nodes = m_plan.getTaskBids().iterator();
	 nodes.hasNext(); ) {
      TaskBid node = (TaskBid)nodes.next();
      computeEarlyStart(node);
      computeLateFinish(node);
    }
  }

  /**
     Checks the feasibility of the plan, returning the total negative
     slack in the plan.  This only works if you run <tt>updateCPM</tt>
     first.
  */
  public long getNegativeSlack () 
  {
    long result = 0;
    for (Iterator tbs = m_plan.getTaskBids().iterator();
	 tbs.hasNext(); ) {
      TaskBid tb = (TaskBid)tbs.next();
      long nodeSlack = tb.getSlackCPM();
      //if (log.isDebugEnabled())
      //log.debug("CPM: slack(task " + tb.getTaskId()
      //	  + ") = " + nodeSlack);
      if (nodeSlack < 0)
	result -= nodeSlack;
    }
    return result;
  }

  private Date computeEarlyStart(TaskBid node) 
  {
    // This actually computes and records early start, and returns
    // the early finish time.  That is what allows the simple
    // recursive operation you see here.
    Date es = node.getEarlyStartCPM();
    if (es == null) 
    {
      // The early start of a node is the later of the latest early
      // finish among the predecessors of node and the early start of
      // the bid being considered for the node.  Only predecessors in
      // different bids are considered here.
      es = node.getEarlyStart(); // this gives the bid value
      for (Iterator predecessors = m_plan.findPredecessors(node.getTask());
	   predecessors.hasNext(); ) {
	TaskBid pred = (m_plan.findTaskBid
			((String)predecessors.next()));
	if (node.getBid() != pred.getBid()) {
	  Date next = computeEarlyStart(pred);
	  if (next.after(es))
	    es = next;
	}
      }
      node.setEarlyStartCPM(es);
      //if (log.isDebugEnabled())
      //log.debug("CPM: set es(task " + node.getTaskId()
      //	  + ") to " + es);
    }
    return new Date(es.getTime() + node.getDuration());      
  }

  private Date computeLateFinish(TaskBid node) 
  {
    // This actually computes and records late finish, and returns
    // the late start time.  That is what allows the simple
    // recursive operation you see here.
    Date lf = node.getLateFinishCPM();
    if (lf == null) 
    {
      // The late finish of a node is the earlier of the earliest late
      // start among the successors of node and the late finish of
      // the bid being considered for the node
      lf = node.getLateFinish(); // this gives the bid value
      for (Iterator successors = m_plan.findSuccessors(node.getTask());
	   successors.hasNext(); ) {
	TaskBid pred = (m_plan.findTaskBid
			((String)successors.next()));
	if (node.getBid() != pred.getBid()) {
	  Date next = computeLateFinish(pred);	    
	  if (next.before(lf))
	    lf = next;
	}
      }
      node.setLateFinishCPM(lf);
      //if (log.isDebugEnabled())
      //log.debug("CPM: set lf(task " + node.getTaskId()
      //	  + ") to " + lf);
    }
    return new Date(lf.getTime() - node.getDuration());      
  }
}
