/*----------------------------------------------------------------------------
 * Magnet Bid Evaluation components
 * $Id: SearchNode.java,v 1.5.2.3 2002/05/05 19:41:49 ashutosh Exp $
 * Copyright (c) 1998 Regents of the University of Minnesota.
 * All Rights Reserved by the Regents of the University of Minnesota.
 * This file contains Original Code and/or Modifications of Original
 * Code as defined in the MAGNET Public Source License, which is
 * available at http://www.cs.umn.edu/magnet/source_license_v1.txt.
 * You may not use this file except in compliance with the License. 
 *----------------------------------------------------------------------------
 * Revision History:
 * 
 * Sat Jan 27 16:18:43 2001  John Collins
 *   Initial Entry, result of refactoring TaskBidPlan
 */

package edu.umn.magnet.client.eval;

import edu.umn.magnet.protocol.cust.Bid;
import edu.umn.magnet.protocol.cust.TaskPlan;
import edu.umn.magnet.protocol.cust.SubtaskRequest;
import edu.umn.magnet.protocol.cust.TaskWithPrecedence;

import org.apache.log4j.Category;

import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Set;
import java.util.Vector;
import java.util.Stack;
import java.util.Iterator;
import java.util.NoSuchElementException;

/**
 * Instances of SearchNode serve to wrap TaskBidPlan instances with
 * attributes and behavior to support queue-based search among
 * alternative (possibly partial) solutions represented by the
 * TaskBidPlan.  There is a one-to-one correspondence between a
 * SearchNode and a TaskBidPlan.
 * <p>
 * Two constructors are provided. One is for creating a root node with
 * no bids, and the other is for creating a deep clone of an existing
 * node in preparation for modifying it to create a new search node.
 * No default constructor is provided.
 * @author John Collins
 */
public class SearchNode implements edu.umn.magnet.util.search.SearchNode
{
  // Init the log category
  static Category log = Category.getInstance(SearchNode.class.getName());

  // Basic database
  private TaskBidPlan m_tbp = null;
  private HashSet m_lockedBids = null;
  private HashSet m_expansionsTried = new HashSet();
  private TabuList m_tabu = null;

  // Data to support bid compatibility constraints
  private BidEvaluator m_bidEval = null;
  private Set m_compatibleBids = null;
  private Set m_usableBids = null;

  // Place in the hierarchy
  private int m_nodeNumber = 0;

  // Evaluation results
  private boolean m_feasible = false;
  private boolean m_covered  = false;
  private HashMap m_evaluations = new HashMap();

  // Selector data
  private Collection m_uncoveredTasks = null;
  private HashMap m_selectorJunk = null; // gets instantiated if a selector
                                   // tries to use it

  /**
   * Creates a root node by creating one of these with a TaskBidPlan.
   * In the process, all the "empty" TaskBids (those without bids)
   * become uncovered tasks, and all the current Bids get locked.
   */
  public SearchNode(TaskBidPlan tbPlan)
  {
    super();
    m_tbp = tbPlan;
    m_lockedBids = new HashSet(10);
    m_tabu = new TabuList();
    m_uncoveredTasks = new HashSet();
    for (Iterator taskBids = tbPlan.getTaskBids().iterator();
	 taskBids.hasNext(); ) {
      TaskBid aTB = (TaskBid)taskBids.next();
      if (aTB.getBid() == null)
	m_uncoveredTasks.add(aTB);
    }
    for (Iterator bids = tbPlan.getBids().iterator();
	 bids.hasNext(); ) {
      m_lockedBids.add(bids.next());
    }
  }

  /**
   * Creates a new SearchNode by copying the guts of an old one, and
   * creating a new TaskBidPlan that is a copy of the current one.
   */
  public SearchNode (SearchNode oldTBP, int nodeCount)
  {
    super();
    m_tbp = new TaskBidPlan(oldTBP.m_tbp);
    m_lockedBids = oldTBP.m_lockedBids;
    m_tabu = (TabuList)oldTBP.m_tabu.clone();
    m_nodeNumber = nodeCount; 
    m_uncoveredTasks = new HashSet(oldTBP.m_uncoveredTasks);

    // Copy over the bid-compatibility info if it exists
    if (oldTBP.m_compatibleBids != null) {
      //log.debug("Init coverage constraints for new node");
      m_bidEval = oldTBP.m_bidEval;
      m_compatibleBids = new HashSet(oldTBP.m_compatibleBids);
    }
  }
  
  public String toString()
  {
    return ("SN" + m_nodeNumber + "(" + m_tbp.toString() + ")");
  }
  
  /**
   * Expands a node by adding a bid.  The new node is returned
   * unevaluated.  It is up to the caller to evaluate it.
   * @return a newly-created TaskBidPlan with the new bid included, or
   *    null if adding the bid would cause displacement of a locked
   *    bid.
   */
  public SearchNode expandNode(Bid aBid, int nodeCount)
  {
    if (aBid == null) {
      log.error("**Strange - TaskBidPlan.expandNode: Null bid");
      (new Exception("TBP: null bid")).printStackTrace();
      return null;
    }
    //first check to see if this bid can be added...
    if (m_tabu.contains(aBid)
	|| m_tbp.getBids().contains(aBid)
	|| m_expansionsTried.contains(aBid)
	|| (m_usableBids != null
	    && ! m_usableBids.contains(aBid)))
      return null;
    // have to create a new SearchNode & TaskBidPlan, then add the bid
    // to the new node
    //log.debug("expanding node " + m_nodeNumber + " with bid "
    //      + aBid.getId() + ", producing node " + nodeCount);
    SearchNode newNode = new SearchNode(this, nodeCount);
    
    m_expansionsTried.add(aBid);
    if (m_usableBids != null)
      m_usableBids.remove(aBid);
    try {
      newNode.addBid(aBid);
    }
    catch (LockedBidException lbe) {
      return null;
    }
    return newNode;
  }
  
  // ============== Evaluation results ====================
  //
  //Note that none of these are of any real use until the evaluator has
  //been run.
  
  /**
   * True just in case the instantiated plan is temporally feasible.
   * If the plan is also covered, this means that the set of bids
   * represented by this TaskBidPlan can be accepted to form a
   * feasible set of bids.  If the plan is not covered, it means that
   * the existing subset of bids is temporally consistent with each
   * other and with the original RFQ.
   */
  public boolean isFeasible()
  {
    return m_tbp.isFeasible(); 
  }
  
  /**
   * True just in case all subtasks in this plan are covered by bids.
   */
  public boolean isCovered ()
  {
    return m_tbp.isCovered();
  }

  /**
   * Returns the total cost of the bids that are part of this plan.
   */
  public long getCost ()
  {
    return m_tbp.getCost();
  }

  /**
   * Returns the evaluation of this plan.  This only has meaning after
   * the plan has been evaluated
   */
  public double getEvaluation ()
  {
    return m_tbp.getEvaluation();
  }

  /**
   * The value method is a synonym for getEvaluation.  This is the
   * implementation of the SearchNode interface.
   */
  public double value ()
  {
    return m_tbp.getEvaluation();
  }

  /**
   * Return the TaskPlan underlying this plan.
   */
  public TaskPlan getTaskPlan()
  {
    return m_tbp.getTaskPlan();
  }
  
  // =============== Selector interface =======================	
  /**
   * Returns the selector junk indexed by the selector's symbol
   */
  public Object getSelectorJunk(String selectorType)
  {
    if (m_selectorJunk == null)
      return null;
    else
      return m_selectorJunk.get(selectorType);
  }

  /**
   * Sets a selector junk object indexed under the selector's symbol
   */
  public void setSelectorJunk(String selectorType, Object junk)
  {
    if (m_selectorJunk == null)
      m_selectorJunk = new HashMap();
    m_selectorJunk.put(selectorType, junk);
  }

  /**
   * Returns a Vector containing TaskBid objects representing the
   * subtasks that are not covered by bids in this node.
   */ 
  public Collection getUncoveredTasks()
  {
    return m_uncoveredTasks;
  }

  /**
   * Sets the uncovered tasks set.  This is done at the beginning of a
   * bid evaluation cycle in a multi-cycle bidding operation.
   */
  public void setUncoveredTasks (Collection newSet)
  {
    m_uncoveredTasks = newSet;
  }

   /**
   * Returns the number of uncovered tasks as an int.
   */
  public int getSizeUncovered()
  {
    return m_uncoveredTasks.size();
  }

  /** 
   * Returns the number of covered tasks as an int
   */

  public int getSizeCovered()
  {
    return (m_tbp.getTaskBids().size() - getSizeUncovered());
  }
  
  /**
   * Returns the coverage factor for this node as a double.  The ratio
   * is the number of covered subtasks divided by the total number of
   * subtasks.
   */
  public double getCoverageFactor()
  {
    return (1.0 - ((double)m_uncoveredTasks.size()
		   /(double)m_tbp.getTaskBids().size()));
  }

  /**
   * Returns the set of bids that are compatible with this node, and
   * are likely to result in a successful expansion.  The set is the
   * compatible bids less the tried bids and the bids on the tabu
   * list.
   */
  public Set getUsableBids ()
  {
    if (m_compatibleBids != null
	&& m_usableBids == null)
      // need to instantiate the usable bids list before returning it
      initUsableBids();
    return m_usableBids;
  }

  /**
   * Returns true if this node has at least one untried compatible
   * bid, or if the compatible bid set for this node is unknown.
   */
  public boolean hasUsableBids ()
  {
    return ((m_usableBids == null)
	    || m_usableBids.size() > 0);
  }

  // ================== Model access ==========================

  /**
   * Returns the TaskBidPlan associated with this node.
   */
  public TaskBidPlan getTaskBidPlan ()
  {
    return m_tbp;
  }

  /**
   * Returns the Collection of TaskBid instances from the underlying
   * TaskBidPlan. 
   */
  public Collection getTaskBids ()
  {
    return m_tbp.getTaskBids();
  }

  /**
   * Locks a bid so that it can't be displaced in future expansions of
   * this node.  If the bid being locked isn't part of this plan, a
   * message is emitted to the log, but no other action is taken.
   */
  public void lockBid (Bid bid)
  {
    // We copy the locked bids set when we do this operation, because
    // we don't create a new copy for every new node.  Failure to copy
    // would therefore smash the locked bids sets in one or more
    // parent nodes.
    log.debug("SN #" + m_nodeNumber
	      + ": Locking bid " + bid.getId());
    m_lockedBids = new HashSet(m_lockedBids);
    m_lockedBids.add(bid);
  }

  /**
   * Returns the set of locked bids.  This is here for testing
   * purposes.
   */
  protected Set getLockedBids ()
  {
    return m_lockedBids;
  }

  /**
   * Adds a Bid into this plan.  It has 
   * to remove old TaskBids which might have been covered by another bid, 
   * and then add the new bid.  This should only be called in the
   * context of expanding a node.
   * @exception LockedBidException thrown if adding the given bid
   *    would cause a locked bid to be displaced.
   */      
  protected void addBid (Bid bid)
    throws LockedBidException
  {
    Set bidsToRemove = m_tbp.addBid(bid);
    for (Iterator bids = bidsToRemove.iterator();
	 bids.hasNext(); ) {
      if (m_lockedBids.contains(bids.next())) {
	throw new LockedBidException();
      }
    }
    // Add the new bid to the tabu list, and possibly create the
    // usable-bids list.
    m_tabu.add(bid);
    updateUsableBids(bid);
  }

  /**
   * Initializes the usable-bids set for this node.  It's the
   * compatible-bids minus the tabu bids, the bids already in the
   * node, and the expansions already tried.
   */
  protected void updateUsableBids (Bid bid)
  {
    // if there's no compatible-bids set, then we don't do anything.
    if (m_compatibleBids == null)
      return;

    // update the compatible-bids set with info from the new bid.
    //log.debug("updateUsableBids " + bid.getId());
    m_compatibleBids.remove(bid);
    Set conflicts = m_bidEval.conflictingBids(bid);
    if (conflicts != null) {
      for (Iterator ibids = conflicts.iterator();
	   ibids.hasNext(); ) {
	m_compatibleBids.remove(ibids.next());
      }
    }
    initUsableBids();
  }

  /**
   * Initializes the usable-bids list
   */
  protected void initUsableBids ()
  {
    // if the usable-bids list hasn't been created yet, create it and
    // then remove all the entries in the tabu list from it.
    if (m_usableBids == null) {
      m_usableBids = new HashSet(m_compatibleBids);
      for (Iterator tbids = m_tabu.iterator();
	   tbids.hasNext(); )
	m_usableBids.remove(tbids.next());
    }
  }

  /**
   * Returns the node number for this node
   */
  public int getNodeNumber ()
  {
    return m_nodeNumber;
  }

  /**
   * Sets the BidEvaluator reference needed to support compatible and
   * usable bids testing
   */
  public void setBidEvaluator (BidEvaluator bidEval)
  {
    m_bidEval = bidEval;
  }

  /**
   * Sets the compatible-bids set to an initial value.  Since we don't
   * want to smash the incoming value, we copy it.
   */
  public void setCompatibleBids (Collection compatibleBids)
  {
    m_compatibleBids = new HashSet(compatibleBids);
  }

  /**
   * Sets the search-type and search runtime in the underlying TBP.
   */
  protected void setSearchData (String method, long time)
  {
    m_tbp.setSearchData(method, time);
  }

  protected void setSearchData (String method, long time, int nodeCount)
  {
    m_tbp.setSearchData(method, time, nodeCount);
  }

  // ================== Static support for uniqueness testing ===========

  // This requires a task ID list of fixed order, and a HashMap of
  // bid ID sequences mapped to SearchNode instances
  static private String[] s_orderedTasks = null;
  static private int s_planSize = 0;
  static private HashSet s_nodeMap = new HashSet(10000);
  static private StringBuffer s_buffer = null;

  static private long s_redundantCount = 0L;
  static private long s_uniqueCount = 0L;

  /**
   * You have to initialize this when you start a new search.  That
   * clears out all the old data and sets up structures of the
   * appropriate size.
   * <p>
   * <i>NOTE:</i> This code assumes an average task name size no
   * larger than 32 characters.  It should be parameterized someday.
   */
  static public void newSearch (BidEvaluator be) {
    //System.out.println("TaskBidPlan.newSearch()");
    TaskPlan tp = be.getTaskPlan();
    s_planSize = tp.size();
    s_orderedTasks = new String[s_planSize];
    s_buffer = new StringBuffer(s_planSize * 32);
    s_nodeMap.clear();
    s_redundantCount = 0L;
    s_uniqueCount = 0L;
    Iterator tasks = tp.iterator();
    for (int i = 0; i < s_planSize; i++) {
      // we'll assume the TaskPlan accurately reports its size, and
      // not check for the end of the Iterator
      s_orderedTasks[i]
	  = ((SubtaskRequest)tasks.next()).getId();
    }
  }

  /**
   * Returns true if the SearchNode given is unique among all the
   * ones tested since the last call to newSearch
   */
  static public boolean isUnique (SearchNode node) {
    // clear out the buffer first, then fill it up
    s_buffer.setLength(0);
    TaskBidPlan tbp = node.getTaskBidPlan();
    for (int i = 0; i < s_planSize; i++) {
      TaskBid tb = tbp.findTaskBid(s_orderedTasks[i]);
      if (tb.getBid() != null)
	s_buffer.append(tb.getBid().getId());
      else
	s_buffer.append("-");
    }
    // check & add
    String tag = s_buffer.toString();
    if (s_nodeMap.contains(tag)) {
      log.debug("TaskBidPlan.isUnique = false " + tag);
      s_redundantCount++;
      if (s_redundantCount % 2000L == 0L) {
        log.info("  redundant count = " + s_redundantCount);
      }
      return false;
    }
    else {
      s_nodeMap.add(tag);
      s_uniqueCount++;
      if (s_uniqueCount % 2000L == 0L) {
        log.info("  unique count = " + s_uniqueCount);
      }
    }
    return true;
  }

  /**
   * This is the exception thrown if an attempt is made to remove a
   * locked bid.
   */
  class LockedBidException
    extends edu.umn.magnet.protocol.MagnetException
  {
    LockedBidException () {
      super();
    }
    
    LockedBidException (String reason) {
      super(reason);
    }
  }
}
