/*----------------------------------------------------------------------------
 * Magnet Bid Evaluation components
 * $Id: SystematicSearch.java,v 1.3.2.3 2002/05/05 19:41:49 ashutosh Exp $
 * Copyright (c) 1999 Regents of the University of Minnesota.
 * All Rights Reserved by the Regents of the University of Minnesota.
 * This file contains Original Code and/or Modifications of Original
 * Code as defined in the MAGNET Public Source License, which is
 * available at http://www.cs.umn.edu/magnet/source_license_v1.txt.
 * You may not use this file except in compliance with the License. 
 *----------------------------------------------------------------------------
 * Revision History:
 * 
 * Fri Apr 20 23:40:59 2001  John Collins
 *   Don't use System properties
 * 
 * Fri Jan 12 09:56:11 2001  John Collins
 *   Add lockAllBids flag.
 * 
 * Thu Dec 14 11:14:43 2000  John Collins
 *   New bid semantics - no breakdowns.
 * 
 * Tue Dec  7 11:39:26 1999  John E. Collins
 *   Update for new data collection scheme in tools.Log
 *
 * Wed March 24 11:52:10 1999 Rashmi Sundareswara
 *   Initial Entry
 */

/* This class is intended for conducting a systematic search among the bids
   received.
    For each bid {
      For each node {
        take a bid and combine with a node, thus creating a new node. Put
        the resulting node back into the queue, if it is feasible;
      }
    }
    If there are any uncovered nodes, 
    for each uncovered node {
       for each Uncovered task {
         for all the  the subtaskbids that cover it {
             evaluate the node with this subtask bid;
              If lower than the expansion done previously, save it, 
             else discard;
         }
       }
    }
*/ 

package edu.umn.magnet.client.eval;

import edu.umn.magnet.protocol.cust.Bid;
import edu.umn.magnet.util.stats.DataCollector;
import edu.umn.magnet.util.search.*;
import edu.umn.magnet.util.MagnetProperties;

import org.apache.log4j.Category;

import java.util.Iterator;
import java.util.Enumeration;
import java.util.NoSuchElementException;
import java.util.LinkedList;


/**
 * Systematic implementation of the Search Engine abstraction
 */
public class SystematicSearch
  extends AbstractSearchEngine
  implements SearchEngine
{
  // Init the log categories
  static Category log = Category.getInstance(SystematicSearch.class.getName());
  static Category result = Category.getInstance("Result");
  
  // Search queue and parameters
  TaskBidPlan m_initialNode = null;
  LinkedList m_searchQueue = new LinkedList();
  private boolean m_lockAllBids = false;
  
  // Statistics
  private int m_nodeCount = 0;  
  
  private DataCollector m_successfulExpansions;
  private SearchNode m_bestPlan = null;  //the best covered and feasible node
  
  // Evaluators and Selectors
  private Eval m_eval;
  
  /**
   * Default constructor.
   */
  public SystematicSearch ()
  {
    super();
  }

  public String getSearchType ()
  {
    return "Sys";
  }

  public void setBidEvaluator (BidEvaluator be)
  {
    super.setBidEvaluator(be);
    readSystemProps();
  }

  /**
   * Run the search.  This is the primary implementation of the
   * SearchEngine interface.  In this case, we just copy some pointers
   * over from the BidEvaluator, set up some randomizers, and take off.
   */
  
  public TaskBidPlan search()
  {
    m_eval = new Eval(bidEval); 
    SearchNode.newSearch(bidEval);
    m_successfulExpansions = new DataCollector();
    SearchNode result = startSearch();
    if (result == null)
      return null;
    else
      return result.getTaskBidPlan();
  }

  /**
   * Sets the initial search node
   */
  public void setInitialNode (TaskBidPlan node)
  {
    m_initialNode = node;
  }

  public SearchNode startSearch()
  {
    log.debug("startSearch()");

    //Create a root node
    TaskBidPlan tbp = m_initialNode;
    SearchNode node = null;
    if (tbp == null) {
      tbp = new TaskBidPlan(bidEval.getTaskPlan());
      node = new SearchNode(tbp);
    }
    else
      node = new SearchNode(tbp);
    //evaluate the root node
    double rootEval = m_eval.evaluate(tbp);
    //add the root node
    m_searchQueue.add(node);
    m_successfulExpansions.newObservation(1.0);
    m_nodeCount++;
    // Check to see if there are Singletons. Singletons are those tasks
    // who only have one bid covering them
    //createSingletons(); // do the easy stuff first, if there is any
    if (! m_searchQueue.isEmpty())
      nodeGenerator();    // this is the heart of the search
    log.info("Complete: Node count:  " + m_nodeCount
	     + ", Feasible and Covered Nodes : " + getPlanCount());
    result.info("Node Count = " + m_nodeCount
		+ ", Feasible and Covered Nodes : " + getPlanCount());
    if (log.isDebugEnabled()) {
      if (m_bestPlan == null)
	log.debug("I have not found a feasible and covered node");
      else
	log.debug(m_bestPlan.toString());
    }
    return m_bestPlan;
  }
    
  protected void nodeGenerator()
  {
    log.debug("nodeGenerator()");
	
    // Now we have to select a bid 
    // We select a bid in a systematic order
    for (Iterator bids = bidEval.getBids().iterator();
	 bids.hasNext(); ){
      Bid bidToConsider = (Bid)bids.next();
      // We select a node in systematic order
      LinkedList newQueue = new LinkedList();
      for (Iterator nodes = m_searchQueue.iterator();
	   nodes.hasNext(); ) {
	SearchNode nodeToExpand = (SearchNode)nodes.next();
	// Here's where we check bid compatibility...
	// Now we have to expand this node with the bid, and evaluate it
	SearchNode nextNode
	    = (SearchNode)nodeToExpand.expandNode(bidToConsider, m_nodeCount);
	if (nextNode != null) {
	  // this was a successful expansion
	  m_nodeCount++;
	  m_successfulExpansions.newObservation(1.0);
	  // evaluate the node and record results
	  TaskBidPlan nextTBP = nextNode.getTaskBidPlan();
	  double eval = m_eval.evaluate(nextTBP);
	  if (log.isDebugEnabled())
	    log.debug("Node " + nextNode.getNodeNumber() + ": -"
		      + (nextTBP.isFeasible() ? "F" : " ")
		      + (nextTBP.isCovered()  ? "C" : " ")
		      + "-, eval = " + eval); 

	  nextNode.lockBid(bidToConsider);

	  if (nextTBP.isFeasible()) {
	    if (nextTBP.isCovered()) {
	      // Track the number of solutions, and don't put good ones
	      // on the queue, since they can't possibly contribute to
	      // other unique solutions.
	      processSolution(nextTBP);
	      
	      if (m_bestPlan == null
		  || (nextTBP.getEvaluation()
		      < m_bestPlan.getEvaluation())) {
		m_bestPlan = nextNode;
	      }
	    }
	    else {
	      // if it's feasible but not covered, we save it for
	      // further expansions
	      newQueue.add(nextNode);
	    }
	  }
	}
	else
	  m_successfulExpansions.newObservation(0.0);
	/* Now the new node has been inserted into the search queue.
	   It will then be expanded again with other bids to try new 
	   possibilities
	*/
      }
      // We've emptied the queue - replace it and go again with the
      // next bid.
      m_searchQueue.addAll(newQueue);
    }
  }

  /**
   * Gets the node count
   */
  public int getNodeCount ()
  {
    return m_nodeCount;
  }

  protected void readSystemProps ()
  {
    String prop;
    // ---- lock all bids flag ----
    prop = MagnetProperties.getProperty("LockAllBids");
    if (prop != null) {
      if (prop.equals("true")) {
	m_lockAllBids = true;
      }
      else {
	m_lockAllBids = false;
      }
    }
    log.info("prop: LockAllBids = " + m_lockAllBids);
  }
}				
