/*------------------------------------------------------------------------
 * Magnet Bid Evaluation components
 * $Id: TaskBid.java,v 1.3.2.3 2002/05/05 19:41:49 ashutosh Exp $
 * Copyright (c) 1998 Regents of the University of Minnesota.
 * All Rights Reserved by the Regents of the University of Minnesota.
 * This file contains Original Code and/or Modifications of Original
 * Code as defined in the MAGNET Public Source License, which is
 * available at http://www.cs.umn.edu/magnet/source_license_v1.txt.
 * You may not use this file except in compliance with the License.
 *------------------------------------------------------------------------
 * Revision History:
 * 
 * Sat Jan 13 11:16:23 2001  John Collins
 *   Remove URD and PUD stuff.
 * 
 * Thu Jan 11 21:07:35 2001  John Collins
 *   Correct computation of late finish for leaf tasks with bids.  It
 *   was coming from the bid itself instead of the task within the bid.
 * 
 * Wed Dec 13 21:57:06 2000  John Collins
 *   Get rid of bid breakdowns, support new bid semantics
 * 
 * Sun Oct  1 15:42:30 2000  John Collins
 *   Use TaskBid for bid planning.  This means it must support the
 *   adjustedDuration property.
 * 
 * Tue Sep 29 20:52:58 1998  John E. Collins
 *   Add unrecoverable decommit probability
 * 
 * Tue Sep 29 13:14:35 1998  John E. Collins
 *   Fix the discount ratio calculation
 * 
 * Mon Sep 28 21:40:05 1998  John E. Collins
 *   Implement bid locking with lockBid.
 * 
 * Mon Sep 28 16:19:33 1998  John E. Collins
 *   More cleanup and compile
 * 
 * Mon Sep 21 21:51:49 1998  John E. Collins
 *   Cleanup
 */

package edu.umn.magnet.client.eval;

import edu.umn.magnet.client.bid.TaskDurationAdjuster;

import edu.umn.magnet.protocol.cust.TaskWithPrecedence;
import edu.umn.magnet.protocol.cust.SubtaskRequest;
import edu.umn.magnet.protocol.cust.Bid;

import org.apache.log4j.Category;

import java.util.Vector;
import java.util.Date;
import java.io.Serializable;

/**
 * This class represents a binding of a task to a bid in the process
 * of evaluating bids.  In order to make predecessor/successor data
 * avaliable, the binding is actually between a TaskWithPrecedence and
 * a GenericBid.
 * <p>
 * A prime use of this structure is the computation of early start and
 * late finish times using the CPM evaluator.  Before running the CPM
 * evaluator, it is necessary to invoke initCPM() on the object.
 * Before the earlyStartCPM and lateFinishCPM attributes are
 * meaningful, the initCPM initialization and the CPM evaluator must
 * have been run across the enclosing TaskBidPlan.
 * @author Rashmi Sundareswara
 */
public class TaskBid implements java.io.Serializable
{
  // init the log category
  static Category log = Category.getInstance(TaskBid.class.getName());

  Bid m_bid  = null;
  TaskWithPrecedence m_twp = null;
  SubtaskRequest m_task = null;

  boolean m_milestone = false;

  // CPM data.  Serves both evaluation and bid prep purposes
  Date m_earlyStartCPM = null;
  Date m_lateFinishCPM = null;
  long m_adjustedDuration = 0L;
  TaskDurationAdjuster m_adjuster = null;

  /** 
      The constructor takes a subtask and a bid. 
  */
  public TaskBid (TaskWithPrecedence twp, Bid bid) 
  {
    super();
    init(twp, bid);
  }

  /**
   * You can also create one without a bid.
   */
  public TaskBid (TaskWithPrecedence twp) 
  {
    m_twp = twp;
    m_task = twp.getTaskData();
  }

  // Initialization is separated from the constructors because of the
  // need to support the GenericBid constructor.  Think about it...
  protected void init (TaskWithPrecedence twp, Bid bid) 
  {
    m_twp  = twp;
    m_task = twp.getTaskData();
    m_bid  = bid;
  }

  /**
     Initializes start and finish time values for the CPM algorithm.  A
     root node gets the early start time set to the later of the
     subtask start time and the bid start time.  A leaf node gets the
     late finish time set to the earlier of the subtask deadline and
     the bid finish time.  All other nodes get start and finish times
     nulled out.  This method must be invoked on all nodes before
     attempting to run the CPM computation.
  */
  public void initCPM ()
  {
    initCPMStart(null);
    initCPMFinish(null);
  }

  public void initCPMStart (Date newStartTime)
  {
    Vector ps = m_twp.getPredecessors();
    if (ps == null || ps.size() == 0) 
    {
      // this is a root -- we only have to initialize it once
      // as long as you never change the plan start date.
      if (m_earlyStartCPM == null) 
      {
	m_earlyStartCPM = m_task.getEarliestTimeToStart();
	if (m_earlyStartCPM == null) {
	  if (newStartTime == null)
	    log.error("TaskBid.initCPMStart: big trouble");
	  m_earlyStartCPM = newStartTime;
	}
	if (m_bid != null) 
	{
	  Date bidEarlyStart = m_bid.getEarlyStart(m_task.getId());
	  if (bidEarlyStart.after(m_earlyStartCPM))
	    m_earlyStartCPM = bidEarlyStart;
	}
	//log.debug("Init root task " + m_task.getId()
	//	  + " to " + m_earlyStartCPM);
      }
    }
    else 
      // Set the early start value to null
      m_earlyStartCPM = null;
  }

  public void initCPMFinish (Date newDeadline)
  {
    Vector ps = m_twp.getSuccessors();
    if (ps == null || ps.size() == 0) 
    {
      // this is a leaf - this works as long as we never change the
      // plan deadline.
      if (m_lateFinishCPM == null) 
      {
	m_lateFinishCPM = m_task.getDeadline();
	if (m_lateFinishCPM == null) {
	  if (newDeadline == null)
	    log.error("TaskBid.initCPMFinish: big trouble");
	  m_lateFinishCPM = newDeadline;
	}

	if (m_bid != null) 
	{
	  Date bidLateFinish = m_bid.getLateFinish(m_task.getId());
	  if (bidLateFinish.before(m_lateFinishCPM))
	    m_lateFinishCPM = bidLateFinish;
	}
	//log.debug("Init leaf task " + m_task.getId()
	//	  + " to " + m_lateFinishCPM);
      }
    }
    else
      // Set the late finish time to null
      m_lateFinishCPM = null;
  }

  /**
     Gets the TaskId from SubtaskRequest.  
  */

  public String getTaskId ()
  {
    return (m_task.getId());
  }

  /**
     Returns the subtask itself.
  */

  public SubtaskRequest getTask ()
  {
    return m_task;
  }

  /** 
      Returns the GenericBid mapped by this node, or null if there is
      no bid.
  */

  public Bid getBid ()
  {
    return m_bid;
  }

  /**
   * Returns the cost of the bid
   */
  public long getCost ()
  {
    return m_bid.getCost();
  }

  /** 
      Returns a boolean value that, if true, indicates that either the
      Customer or the Supplier has specified that the end of this
      subtask will not be tracked. 
      <p>
      If a task is not a milestone, and it
      is accepted as part of a composite bid, then the start time,
      finish time, and duration from the subtask bid have meaning.
      Otherwise they do not.
      <p>
      This property is not modifiable here.  It must be set either in
      the CFB or in the Bid.
  */
  public boolean isMilestone ()
  {
    return m_milestone;
  }

  /** 
      Returns the early start date for the task.  It comes from the
      bid if that is available, otherwise it comes from the subtask
      specification.
  */
  public Date getEarlyStart ()
  {
    if (m_bid != null)
      return m_bid.getEarlyStart(m_task.getId());
    else
      return m_task.getEarliestTimeToStart();
  }

  /**
     Returns the early start time as computed by CPM, or null if the
     CPM evaluator hasn't seen this node yet.
  */
  public Date getEarlyStartCPM () 
  {
    return m_earlyStartCPM;
  }

  /** 
      Sets the CPM-computed early start time.  This is presumably only
      called from the CPM evaluator.
  */
  public void setEarlyStartCPM (Date newES) 
  {
    m_earlyStartCPM = newES;
  }

  /**
     Returns the task duration as specified by the bid, if any.  If
     there is no bid, the value is 0.
  */
  public long getDuration ()
  {
    if (m_bid != null)
      return m_bid.getDuration(m_task.getId());
    else
      return 0l; // estimate needs to be recorded somewhere...
  }

  /**
   * Returns the bid duration if that's available, otherwise the
   * expected duration of this task as recorded in the underlying
   * SubtaskRequest.
   */
  public long getExpectedDuration ()
  {
    // If this task has a bid, we use the bid duration
    if (m_bid != null) {
      return m_bid.getDuration(m_task.getId());
    }
    return Math.round(m_task.getJobInfo().getExpectedDuration());
  }

  /**
   * Returns the adjusted duration for the purpose of setting bid time
   * windows.
   */
  public long getAdjustedDuration (TaskDurationAdjuster adjuster)
  {
    // If this task has a bid, we use the bid duration
    if (m_bid != null) {
      return m_bid.getDuration(m_task.getId());
    }

    // No bid, use the adjuster
    if (adjuster != m_adjuster) {
      // new adjuster - throw out the old info
      m_adjuster = adjuster;
      m_adjustedDuration = 0L;
    }
    
    // If the adjusted duration is non-zero, then we've already dealt
    // with this one.
    if (m_adjustedDuration == 0L) {
      m_adjuster.processTaskBid(this);
    }
    return m_adjustedDuration;
  }

  /**
   * Sets the adjusted duration.  This is needed by the task duration
   * adjuster when setting RFQ time windows.
   */
  public void setAdjustedDuration (long newValue)
  {
    m_adjustedDuration = newValue;
  }

  /**
     Returns the late finish time for this task.  It comes from the
     bid if there is a bid, otherwise it comes from the subtask
     specification.
  */
  public Date getLateFinish ()
  {
    if (m_bid != null)
      return m_bid.getLateFinish(m_task.getId());
    else
      return m_task.getDeadline();
  }

  /**
     Returns the late finish time as computed by CPM, or null if the
     CPM evaluator hasn't seen this node yet.
  */
  public Date getLateFinishCPM () 
  {
    return m_lateFinishCPM;
  }

  /** 
      Sets the CPM-computed late finish time.  This is presumably only
      called from the CPM evaluator.
  */
  public void setLateFinishCPM (Date newLF) 
  {
    m_lateFinishCPM = newLF;
  }

  /** 
      Returns the schedule slack from the bid, if available, otherwise
      from the subtask specification.
  */
  public long getSlack ()
  {
    return (getLateFinish().getTime() - getEarlyStart().getTime())
	- getDuration();
  }

  /**
     Returns the schedule slack from the CPM data.  This is meaningful
     only if the CPM data has been filled in.
  */
  public long getSlackCPM () 
  {
    return (getLateFinishCPM().getTime() - getEarlyStartCPM().getTime())
	- getDuration();
  }
}
