/*------------------------------------------------------------------------
 * MAGNET Customer Agent Plan Management components
 * $Id: PlanStatus.java,v 1.14.2.2 2002/05/05 19:41:50 ashutosh Exp $
 * Copyright (c) 2000 Regents of the University of Minnesota.
 * All Rights Reserved by the Regents of the University of Minnesota.
 * This file contains Original Code and/or Modifications of Original
 * Code as defined in the MAGNET Public Source License, which is
 * available at http://www.cs.umn.edu/magnet/source_license_v1.txt.
 * You may not use this file except in compliance with the License.
 *------------------------------------------------------------------------
 * Revision History:
 * 
 * Tue Jan  8 17:43:51 2002  John Collins  #0000037
 *   Add String accessor for TaskStatus
 * 
 * Tue May 22 14:05:53 2001  John Collins  #0000121
 *   Add getStateName and associated resource handling
 * 
 * Fri Sep 29 19:11:59 2000  John Collins
 *   Initial Entry
 */

package edu.umn.magnet.client.status;

import edu.umn.magnet.client.eval.TaskBidPlan;
import edu.umn.magnet.protocol.InvalidDataException;
import edu.umn.magnet.protocol.cust.SubtaskRequest;
import edu.umn.magnet.protocol.cust.TaskPlan;
import edu.umn.magnet.util.time.AgentTime;
import edu.umn.magnet.util.time.TimerAction;

import org.apache.log4j.Category;

import java.util.Date;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.HashMap;
import java.util.Locale;
import java.util.ResourceBundle;

/**
 * Instances are responsible for tracking the status of a plan through
 * the process of bidding and execution.  The principal elements are
 * the underlying TaskPlan, which represents the external view of a
 * Customer's plan, and the taskStatusData, which includes current
 * status on each task.
 * <p>
 * Each plan maintains its own TimeMap to keep track of plan-related
 * events and the passage of time.  
 * </p><p>
 * PlanStatusEvents will be issued to notify listeners of changes in
 * the following circumstances.  You get notified if you first call
 * <tt>addListener</tt>, giving an instance of a
 * <tt>PropertyChangeListener</tt>.
 * <table>
 * <tr>
 *   <th><u>Property Name</u></th><th><u>Event</u></th>
 * </tr>
 * <tr valign=top>
 *   <td><strong>TaskBidPlan</strong></td>
 *   <td>Issued when a new TaskBidPlan is posted for this plan.  The
 *     old and new TaskBidPlan instances are contained in the event.
 * </tr>
 * <tr valign=top>
 *   <td><strong>State</strong></td>
 *   <td>Issued when the State value changes.  The old and new state
 *     values are contained in the event as String objects.</td>
 * </tr>
 * <tr valign=top>
 *   <td><strong>Start</strong></td>
 *   <td>Issued when the Plan start time is set or changed.  The old
 *     and new start times are contained in the event as Date
 *     instances.</td> 
 * </tr>
 * <tr valign=top>
 *   <td><strong>Deadline</strong></td>
 *   <td>Issued when the Plan deadline time is set or changed.  The
 *     old and new deadline times are contained in the event as Date
 *     instances.</td>
 * </tr>
 * <tr valign=top>
 *   <td><strong>NewSession</strong></td>
 *   <td>Issued when a new Market Session has been created for this
 *     plan.  The event includes the new SessionStatus instance as its
 *     newValue.  You have to go there to interact with the RFQ's in
 *     the session.</td>
 * </tr>
 * </table>
 * </p><p>
 * An active Plan maintains two {@link TMEntry} instances in its
 * {@link TimeMap}, one for the plan start time, and for the plan
 * deadline.  These are used to drive the state changes, and will
 * therefore trigger events when they occur.  You can use these
 * entries to drive progress bars or whatever other time-based
 * functionality you desire.  However, it isn't kosher to change the
 * action or the expiration time of this object.  If you need to
 * change the plan start or deadline info, change it with the {@link
 * #setPlanStart setPlanStart} or {@link #setPlanDeadline
 * setPlanDeadline} methods on this class.
 * </p>
 * @author John Collins
 * @version $Revision $
 */
public class PlanStatus extends AbstractStatus
{
  private static Category log
    = Category.getInstance(PlanStatus.class.getName());

  // Plan state constants:
  public static final String PLANNING = "plan";
  public static final String BIDDING = "bid";
  public static final String EXECUTION = "ex";
  public static final String COMPLETE = "complete";
  public static final String ABORTED = "abort";

  private TaskBidPlan plan = null;
  private HashMap taskStatusData = new HashMap();
  private String planState = PLANNING;
  private Date planStart = null;
  private Date planDeadline = null;
  private TimeMap map = new TimeMap();
  private ArrayList sessions = new ArrayList();
  private TMEntry planStartTME = null;
  private TMEntry planDeadlineTME = null;

  private static ResourceBundle stateStrings = null;

  /**
   * You construct one of these by giving it a TaskBidPlan.  It then
   * builds its task status table.
   */
  protected PlanStatus (TaskBidPlan thePlan)
  {
    super();
    init(thePlan);
  }

  /**
   * You can also construct one of these with pre-set start time and
   * deadline values.
   */
  protected PlanStatus (TaskBidPlan thePlan, Date start, Date end)
  {
    this(thePlan);
    if (start != null) {
      planStart = start;
      setPlanStartTME(start);
    }
    if (end != null) {
      planDeadline = end;
      setPlanDeadlineTME(end);
    }
  }

  /**
   * Initialization, includes loading state string resource bundle
   */
  private void init (TaskBidPlan thePlan)
  {
    plan = thePlan;
    for (Iterator tasks = thePlan.getTaskPlan().iterator();
	 tasks.hasNext(); ) {
      SubtaskRequest task = (SubtaskRequest)tasks.next();
      taskStatusData.put(task, new TaskStatus(task));
    }
    if (stateStrings == null)
      staticInit();
  }

  private static void staticInit ()
  {
    stateStrings = ResourceBundle.getBundle
	("edu.umn.magnet.client.status.PlanStatusNames",
	 Locale.getDefault());
    log.debug("Loaded resource bundle");
  }

  /**
   * Returns the TaskBidPlan
   */
  public TaskBidPlan getTaskBidPlan ()
  {
    return plan;
  }

  /**
   * Sets the current TaskBidPlan.  Note that we assume setting this
   * won't change the underlying TaskPlan instance, and we don't check
   * to see whether it does.
   */
  public void setTaskBidPlan (TaskBidPlan newPlan)
  {
    TaskBidPlan oldPlan = plan;
    plan = newPlan;
    sendStatusEvent(new PlanStatusEvent(this, "TaskBidPlan",
					oldPlan, newPlan));
  }

  /**
   * Returns the underlying TaskPlan.
   */
  public TaskPlan getTaskPlan ()
  {
    return plan.getTaskPlan();
  }

  /**
   * Returns the current plan state.  It's one of PLANNING, BIDDING,
   * EXECUTION, COMPLETE, ABORTED.
   */
  public String getPlanState ()
  {
    return planState;
  }

  /**
   * Synonym for getPlanState.
   */
  public String getState ()
  {
    return getPlanState();
  }

  /**
   * Returns the localized name of the current state.
   */
  public String getStateName ()
  {
    return stateStrings.getString(getPlanState());
  }

  /**
   * Sets the current plan state.  If you set it to COMPLETE or
   * ABORTED, then the underlying TimeMap is stopped.  If you don't
   * stop the TimeMap, its thread will continue to run and may do
   * stuff you don't want, not limited to preventing your program from
   * exiting. 
   */
  public void setPlanState (String newState)
  {
    String oldState = planState;
    planState = newState;
    log.debug("State changed from " + oldState
	      + " to " + planState);
    if (newState.equals(COMPLETE) || newState.equals(ABORTED)) {
      map.halt();
    }
    sendStatusEvent(new PlanStatusEvent(this, "State", oldState, newState));
  }

  /**
   * Synonym for setPlanState.
   */
  public void setState (String newState)
  {
    setPlanState(newState);
  }

  /**
   * Cancels this plan and all it's subsidiary status objects.
   */
  public void cancel ()
  {
    for (Iterator i = sessions.iterator(); i.hasNext(); ) {
      SessionStatus ss = (SessionStatus)i.next();
      ss.cancel();
    }
    setPlanState(ABORTED);
  }

  /**
   * Returns the {@link TimeMap} for this plan.
   */
  public TimeMap getMap ()
  {
    return map;
  }

  // -------- set and get the plan start and end times --------

  /**
   * Sets the start time for the plan. It has to be in the future, and
   * it has to be earlier than the plan deadline, or you will get an
   * <tt>InvalidDataException</tt>.
   * @exception {@link edu.umn.magnet.protocol.InvalidDataException
   * InvalidDataException} thrown if the new start time is 
   *   not in the future, or if it is not earlier than the plan
   *   deadline. 
   * @exception {@link InvalidStateException} thrown if the plan is
   *   already in EXECUTION, COMPLETE, or ABORTED.
   */
  public void setPlanStart (Date newStartTime)
    throws InvalidDataException, InvalidStateException
  {
    if (planState.equals(EXECUTION)
	|| planState.equals(COMPLETE)
	|| planState.equals(ABORTED))
      throw new InvalidStateException("Can't change plan start time in state "
				      + planState);
    Date now = AgentTime.getDate();
    if (newStartTime.getTime() <= now.getTime())
      // not in the future
      throw new InvalidDataException("Specified time " + newStartTime
				     + " not in the future");
    if (planDeadline != null
	&& planDeadline.getTime() <= newStartTime.getTime())
      // not later than plan start
      throw new InvalidDataException("Specified time " + newStartTime
				     + " not earlier than plan deadline "
				     + planDeadline);
    Date oldStart = planStart;
    planStart = newStartTime;
    log.debug("Start time changed from " + oldStart
	      + " to " + planStart);
    sendStatusEvent(new PlanStatusEvent(this, "Start",
					oldStart, newStartTime));
    setPlanStartTME(newStartTime);
  }

  private void setPlanStartTME (Date newStart)
  {
    if (planStartTME != null)
      // get rid of the old one before installing the new one.
      map.deleteEntry(planStartTME);
    planStartTME = new TMEntry(newStart,
			       new TimerAction() {
				   public void action ()
				   {
				     setState(EXECUTION);
				   }
				 });
    map.addEntry(planStartTME);
  }

  /**
   * Gets the start time for the plan.
   */
  public Date getPlanStart ()
  {
    return planStart;
  }

  /**
   * Gets the Plan Start Time Map Entry as a {@link TMEntry} instance.
   */
  public TMEntry getPlanStartTME ()
  {
    return planStartTME;
  }

  /**
   * Sets the deadline for the plan.  It has to be in the future, and
   * it has to be later than the plan start time, or
   * you get an InvalidDataException.
   * @exception {@link edu.umn.magnet.protocol.InvalidDataException
   * InvalidDateException} thrown if the new 
   *   deadline is in the past, or is not later than the plan start time.
   * @exception {@link InvalidStateException} thrown if the current
   *   state is COMPLETE or ABORTED
   */
  public void setPlanDeadline (Date newDeadline)
    throws InvalidDataException, InvalidStateException
  {
    if (planState.equals(COMPLETE) || planState.equals(ABORTED))
      throw new InvalidStateException("Can't change plan start time in state "
				      + planState);
    Date now = AgentTime.getDate();
    if (newDeadline.getTime() <= now.getTime())
      // not in the future
      throw new InvalidDataException("Specified time " + newDeadline
				     + " not in the future");
    if (planStart != null
	&& planStart.getTime() >= newDeadline.getTime())
      // not later than plan start
      throw new InvalidDataException("Specified time " + newDeadline
				     + " not later than plan start time "
				     + planStart);
    Date oldDeadline = planDeadline;
    planDeadline = newDeadline;
    log.debug("Deadline changed from " + oldDeadline
	      + " to " + planDeadline);
    sendStatusEvent(new PlanStatusEvent(this, "Deadline",
					oldDeadline, newDeadline));
    setPlanDeadlineTME(newDeadline);
  }

  private void setPlanDeadlineTME (Date newDeadline)
  {
    if (planDeadlineTME != null)
      // Remove the old one before adding the new one
      map.deleteEntry(planDeadlineTME);
    planDeadlineTME = new TMEntry(newDeadline,
			       new TimerAction() {
				   public void action ()
				   {
				     setState(COMPLETE);
				   }
				 });
    map.addEntry(planDeadlineTME);
  }

  /**
   * Gets the deadline for the plan.
   */
  public Date getPlanDeadline () 
  {
    return planDeadline;
  }

  /**
   * Returns the {@link TMEntry} for the plan deadline
   */
  public TMEntry getPlanDeadlineTME ()
  {
    return planDeadlineTME;
  }

  // -------- Task Status getters --------

  /**
   * Returns the TaskStatus instance for a particular task
   */
  public TaskStatus getTaskStatus (SubtaskRequest task)
  {
    return (TaskStatus)taskStatusData.get(task);
  }

  /**
   * Returns the TaskStatus instance for a particular task identified
   * by its ID string.
   */
  public TaskStatus getTaskStatus (String taskId)
  {
    return (TaskStatus)taskStatusData.get(getTaskPlan().getTaskData(taskId));
  }

  /**
   * Returns an Iterator across the TaskStatus instances
   */
  public Iterator getTaskStatusIterator ()
  {
    return taskStatusData.values().iterator();
  }

  // ----------- Convenience pass-throughs for task status --------------
  
  /**
   * Gets the state of a particular task
   */
  public String getTaskState (SubtaskRequest task)
  {
    return getTaskStatus(task).getState();
  }

  /**
   * Sets the state of a particular task
   */
  public void setTaskState (SubtaskRequest task, String newState)
  {
    getTaskStatus(task).setState(newState);
  }

  /**
   * Returns a list of tasks that are in one of a set of states
   * @param states a Collection of TaskStatus state constants.
   */
  public Collection getTasksInStates (Collection states)
  {
    ArrayList result = new ArrayList(taskStatusData.size());
    for (Iterator stats = getTaskStatusIterator();
	 stats.hasNext(); ) {
      TaskStatus stat = (TaskStatus)stats.next();
      if (stat.isInStates(states))
	result.add(stat.getTask());
    }
    return result;
  }

  // ----------- Session status management -------------

  /**
   * Creates and returns a new SessionStatus instance to track a
   * particular Market Session.
   * Typically, you would create the session itself before creating
   * the SessionStatus instance, so the create method takes a Session
   * reference.
   * @exception {@link InvalidStateException} thrown if the current
   *   state is COMPLETE or ABORTED
   */
  public SessionStatus createSessionStatus (String theSession)
    throws InvalidStateException
  {
    if (planState.equals(COMPLETE) || planState.equals(ABORTED))
      throw new InvalidStateException("Can't create new session in state "
				      + planState);
    SessionStatus newStat = new SessionStatus(this, theSession);
    sessions.add(newStat);
    notifyNewSession(newStat);
    return newStat;
  }

  /**
   * Creates and returns a new SessionStatus instance.  
   * reference yourself later in this case.
   * @exception {@link InvalidStateException} thrown if the current
   *   state is COMPLETE or ABORTED
   */
  public SessionStatus createSessionStatus ()
    throws InvalidStateException
  {
    if (planState.equals(COMPLETE) || planState.equals(ABORTED))
      throw new InvalidStateException("Can't create new session in state "
				      + planState);
    SessionStatus newStat = new SessionStatus(this);
    sessions.add(newStat);
    notifyNewSession(newStat);
    return newStat;
  }

  private void notifyNewSession (SessionStatus newStat)
  {
    sendStatusEvent(new PlanStatusEvent(this, "NewSession", null, newStat));
  }
  
  /**
   * Returns the list of SessionStatus instances associated with this
   * plan. 
   */
  public Collection getSessions ()
  {
    return sessions;
  }
}
