/*------------------------------------------------------------------------
 * MAGNET status component
 * $Id: RFQStatus.java,v 1.14.2.1 2002/05/05 19:41:50 ashutosh Exp $
 * Copyright (c) 2001 Regents of the University of Minnesota.
 * All Rights Reserved by the Regents of the University of Minnesota.
 * This file contains Original Code and/or Modifications of Original
 * Code as defined in the MAGNET Public Source License, which is
 * available at http://www.cs.umn.edu/magnet/source_license_v1.txt.
 * You may not use this file except in compliance with the License.
 *------------------------------------------------------------------------
 * Revision History:
 * 
 * Tue Jan  8 14:34:10 2002  John Collins  #0000037
 *   Manage TaskStatus more carefully, add bids to TaskStatus as they
 *   arrive.
 * 
 * Tue May 22 14:05:53 2001  John Collins  #0000121
 *   Add getStateName and associated resource handling
 *
 * Sun May 20 16:33:45 2001  Paula Greve   #0000102
 *  Add ability to get the RFQ associated with this status
 *
 * Fri May 18 10:43:59 2001  John Collins  #0000109
 *   Add cancel method
 *
 * Tue May  8 11:45:42 2001  John Collins  #0000064
 *   Make the wait for bidding closed be optional, to speed up bid
 *   eval experiments.
 *
 * Wed May  2 13:02:00 2001  John Collins
 *   Initial Entry
 */

package edu.umn.magnet.client.status;

import edu.umn.magnet.protocol.InvalidDataException;
import edu.umn.magnet.protocol.cust.RFQ;
import edu.umn.magnet.protocol.cust.Bid;
import edu.umn.magnet.protocol.cust.SubtaskRequest;

import edu.umn.magnet.client.bid.BidCollector;
import edu.umn.magnet.client.bid.BidCollectorFactory;

import edu.umn.magnet.util.time.AgentTime;
import edu.umn.magnet.util.time.TimerAction;
import edu.umn.magnet.util.MagnetProperties;

import org.apache.log4j.Category;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.Iterator;
import java.util.Locale;
import java.util.ResourceBundle;

/**
 * Instances manage and track the external bid processing for a single
 * RFQ.  The following event notifications are supported, using {@link
 * RFQStatusEvent} instances.
 * <table>
 * <tr>
 *   <th><u>Property Value</u></th><th><u>Event</u></th>
 * </tr>
 * <tr valign=top>
 *   <td><strong>RFQStatus.STATE</strong></td>
 *   <td>Issued when the State value changes.  The old and new state
 *     values are contained in the event as String objects.</td>
 * </tr>
 * <tr valign=top>
 *   <td><strong>RFQStatus.BID_DEADLINE</strong></td>
 *   <td>Issued when the {@link TMEntry} for the Bid Deadline has been
 *     set.  At this point, you can retrieve the TMEntry object with
 *     {@link #getBidDeadlineTME getBidDeadlineTME}.</td>
 * </tr>
 * <tr valign=top>
 *   <td><strong>RFQStatus.EARLIEST_OFFER</strong></td>
 *   <td>Issued when the {@link TMEntry} for the Earliest Offer has
 *     been set.  At this point, you can retrieve the TMEntry object
 *     with {@link #getEarliestOfferTME getEarliestOfferTME}.</td>
 * </tr>
 * <tr valign=top>
 *   <td><strong>RFQStatus.NEW_BID</strong></td>
 *   <td>Issued when a new Bid arrives for this RFQ.</td>
 * </tr>
 * </table>
 * <p>
 * Note that event types must be compared with the
 * <code>equals()</code> method.</p>
 * <p>
 * You use an RFQStatus instance by asking the {@link SessionStatus
 * SessionStatus} object to create
 * it, giving it an RFQ, then calling {@link #acquireBids
 * acquireBids}.  This separation is
 * done to allow for event distribution between initialization and the
 * start of the bid cycle.
 * Here are the state transitions you might want to watch for:
 * <table>
 * <tr>
 *   <th><u>Situation</u></th><th><u>Next State</u></th>
 * </tr>
 * <tr valign=top>
 *   <td>Initialization</td>
 *   <td><strong>RFQStatus.NONE</strong></td>
 * </tr>
 * <tr valign=top>
 *   <td>RFQ submitted</td>
 *   <td><strong>RFQStatus.SUBMITTED</strong></td>
 * </tr>
 * <tr valign=top>
 *   <td>Bid deadline arrives</td>
 *   <td><strong>RFQStatus.BIDDING_CLOSED</strong></td>
 * </tr>
 * <tr valign=top>
 *   <td>All bids acquired and distributed</td>
 *   <td><strong>RFQStatus.BIDS_ACQUIRED</strong></td>
 * </tr>
 * <tr valign=top>
 *   <td>Earliest offer time arrives</td>
 *   <td><strong>RFQStatus.AWARD_READY</strong></td>
 * </tr>
 * <tr valign=top>
 *   <td>All bids awarded (is there a timeout for this?)</td>
 *   <td><strong>RFQStatus.AWARDED</strong></td>
 * </tr>
 * <tr valign=top>
 *   <td>Work complete</td>
 *   <td><strong>RFQStatus.COMPLETE</strong></td>
 * </tr>
 * <tr valign=top>
 *   <td>RFQ Abandonded or aborted</td>
 *   <td><strong>RFQStatus.ABANDONED</strong></td>
 * </tr>
 * </table>
 * </p><p>
 * <strong>Note:</strong>   If you want to run experiments without
 * waiting for the BIDDING_CLOSED timeout, you have to set the
 * RFQStatus.Bids.DontWait property to "true" in your properties file
 * (which sets the property by that name in {@link
 * edu.umn.magnet.util.MagnetProperties MagnetProperties}).  You
 * should also make
 * sure you have left enough time to complete your searches before the
 * plan start date.
 * @author John Collins
 * @version $Revision: 1.14.2.1 $
 */
public class RFQStatus extends AbstractStatus
{
  private static Category log
    = Category.getInstance(RFQStatus.class.getName());

  /** Event type: state change */
  public static final String STATE = "State";
  /** Event type: new bid deadline */
  public static final String BID_DEADLINE = "BidDeadline";
  /** Event type: new earliest offer time */
  public static final String EARLIEST_OFFER = "EarliestOffer";
  /** Event type: new bid arrived */
  public static final String NEW_BID = "NewBid";

  /** State constant: no state, just initialized */
  public static final String NONE = "none";
  /** State constant: RFQ submitted */
  public static final String SUBMITTED = "sub";
  /** State constant: Bid deadline passed */
  public static final String BIDDING_CLOSED = "bc";
  /** State constant: All bids acquired and available */
  public static final String BIDS_ACQUIRED = "ba";
  /** State constant: Awards may be issued */
  public static final String AWARD_READY = "ar";
  /** State constant: Award time is past */
  public static final String AWARDED = "aw";
  /** State constant: Work complete */
  public static final String COMPLETE = "cmp";
  /** State constant: RFQ abandoned */
  public static final String ABANDONED = "ab";

  private RFQ theRFQ = null;
  private SessionStatus parent = null;
  private PlanStatus plan = null;
  private String state = NONE;
  private ArrayList theBids = new ArrayList(100);

  private TMEntry bidDeadlineTME = null;
  private TMEntry earliestOfferTME = null;

  private static ResourceBundle rfqStrings = null;

  protected RFQStatus (SessionStatus parent)
  {
    super();
    try {
      init(parent, null);
    }
    catch (InvalidDataException ide) {
      // Can't happen ?
      log.error("Invalid data exception - shouldn't happen with null RFQ");
    }
  }

  protected RFQStatus (SessionStatus parent, RFQ rfq)
    throws InvalidDataException
  {
    super();
    init(parent, rfq);
  }

  /**
   * Initialization
   */
  private void init (SessionStatus parent, RFQ rfq)
    throws InvalidDataException
  {
    this.parent = parent;
    this.plan = parent.getParent();
    if (rfq != null)
      this.setRFQ(rfq);
    if (rfqStrings == null)
      staticInit();
  }

  private static void staticInit ()
  {
    rfqStrings =
	ResourceBundle.getBundle("edu.umn.magnet.client.status.RFQStatusNames",
				 Locale.getDefault());
    log.debug("Loaded resource bundle");
  }

  /**
   * Returns the current state of the underlying RFQ.
   */
  public synchronized String getState ()
  {
    return state;
  }

  /**
   * Returns the localized name of the current state.
   */
  public String getStateName ()
  {
    return rfqStrings.getString(getState());
  }

  /**
   * Sets the state of the associated RFQ.  An RFQStatusEvent will be
   * issued with the State label.
   */
  private synchronized void setState (String newState)
  {
    String oldState = state;
    state = newState;
    log.info("RFQStatus " + theRFQ.getId() + " state set to " + newState);
    notifyAll();
    sendStatusEvent(new RFQStatusEvent(this, STATE, oldState, newState));
  }

  private synchronized String waitForStateChange (String oldState)
  {
    while (getState().equals(oldState)) {
      try {
	wait();
      }
      catch (InterruptedException ie) {
	log.warn("bid collector thread interrupted");
      }
    }
    return state;
  }

  //todo: setters for COMPLETE and ABANDONED

  /**
   * Cancels the RFQ, aborting any activity and moving to the
   * ABANDONED state.  If a bid collector is still active, it will
   * continue to run, but the bids will be dropped on the floor.  The
   * only event that will be issued by this object is the state change
   * event due to the change to the ABANDONED state.  No further
   * events will be sent.
   * <p>
   * If this status instance is already in the ABANDONED state, then
   * nothing will happen, and no events will be issued.</p>
   * <p>
   * <strong>Question:</strong> Where do we restore the tasks to their
   * pre-rfq states? It's not being done here.</p>
   */
  public void cancel ()
  {
    if (! state.equals(ABANDONED)) {
      // first, get rid of the timeouts so they don't happen while we do
      // the rest of this stuff
      TimeMap theMap = plan.getMap();
      theMap.deleteEntry(bidDeadlineTME);
      bidDeadlineTME = null;
      theMap.deleteEntry(earliestOfferTME);
      earliestOfferTME = null;

      // Set the state to ABANDONED
      setState(ABANDONED);

      // Clean out the instance state
      theRFQ = null;
      parent = null;
      plan = null;
    }
  }

  /**
   * Sets the RFQ.  This works as long as
   * <ul>
   *  <li>You haven't already set the RFQ,</li>
   *  <li>The state is currently NONE,</li>
   *  <li>The time data in the RFQ is currently valid, and</li>
   *  <li>All the tasks specified in the RFQ are in the OFFERED
   *      state.</li>
   * </ul>
   * This starts the timers running.  You then need
   * to call {@link #acquireBids acquireBids} before the clock runs
   * out.
   * @exception InvalidDataException thrown if the bid deadline is not
   *   in the future, or if the earliest offer is not later than the bid
   *   deadline, or if some task is not in the OFFERED state.
   */
  public void setRFQ (RFQ rfq) throws InvalidDataException
  {
    if (theRFQ != null)
      throw new InvalidDataException("You can only set the RFQ once.");
    if (!state.equals(NONE))
      throw new InvalidDataException("Incorrect state in RFQStatus");
    for (Iterator tasks = rfq.getTaskPlan().getTasks().iterator();
	 tasks.hasNext(); ) {
      SubtaskRequest task = (SubtaskRequest)tasks.next();
      String state = plan.getTaskState(task);
      if (!state.equals(TaskStatus.OFFERED)) {
	log.error("setRFQ: bad task state " + state + " for task "
		  + task.getId());
	throw new InvalidDataException("Incorrect task state for task "
				       + task.getId());
      }
    }

    this.theRFQ = rfq;

    // set up timers based on RFQ data.  Once you've done this, the
    // clock is running.
    setBidDeadline(theRFQ.getBidDeadline());
    setEarliestOffer(theRFQ.getEarliestOffer());
  }

 /**
   * Gets the RFQ.  This allows verification of the RFQ that this status is using.
   */
  public RFQ getRFQ ()
  {
    return this.theRFQ;
  }

  /**
   * Sets the bid deadline timer.  The state will change to
   * BIDDING_CLOSED when the timer runs out.
   * @exception InvalidDataException thrown if the bid deadline is not
   *   in the future.
   */
  private void setBidDeadline(Date deadline)
    throws InvalidDataException
  {
    Date now = AgentTime.getDate();
    if (deadline.getTime() <= now.getTime()) {
      log.error("Bid deadline not in the future");
      throw new InvalidDataException("Specified time " + deadline
				     + " not in the future");
    }
    // we'll assume this only happens once, so the old deadline is
    // null.
    log.debug("Bid Deadline set to " + deadline);
    sendStatusEvent(new RFQStatusEvent(this, BID_DEADLINE, null, deadline));
    setBidDeadlineTME(deadline);
  }

  /**
   * Sets the BidDeadline TMEntry and emits the appropriate event
   */
  private void setBidDeadlineTME (Date deadline)
  {
    TimeMap theMap = plan.getMap();
    if (bidDeadlineTME != null)
      theMap.deleteEntry(bidDeadlineTME);
    bidDeadlineTME = new TMEntry(deadline,
				 new TimerAction() {
				     public void action () {
				       setState(BIDDING_CLOSED);
				     }
				   });
    theMap.addEntry(bidDeadlineTME);

  }

  /**
   * Returns the TMEntry instance that controls the bid deadline.
   */
  public TMEntry getBidDeadlineTME ()
  {
    return bidDeadlineTME;
  }

  /**
   * Cancels the bid deadline TME.  Presumably you intend to either
   * abort the plan or change the state to BIDDING_CLOSED manually.
   */
  private void cancelBidDeadlineTME ()
  {
    TimeMap theMap = plan.getMap();
    theMap.deleteEntry(bidDeadlineTME);
    bidDeadlineTME = null;
  }

  /**
   * Sets the earliest offer timer.  The state will change to
   * AWARD_READY when the timer runs out.
   * @exception InvalidDataException thrown if the bid deadline is not
   *   set, or is not earlier than the earliest offer time.
   */
  private void setEarliestOffer (Date earliestOffer)
    throws InvalidDataException
  {
    Date now = AgentTime.getDate();
    if (bidDeadlineTME == null
	|| bidDeadlineTME.getExpirationTime() >= earliestOffer.getTime()) {
      log.error("Invalid earliest offer time");
      throw new InvalidDataException("Specified time " + earliestOffer
				     + " not later than bid deadline");
    }
    // we'll assume this only happens once, so the old deadline is
    // null.
    log.debug("Earliest Offer set to " + earliestOffer);
    sendStatusEvent(new RFQStatusEvent(this, EARLIEST_OFFER,
				       null, earliestOffer));
    setEarliestOfferTME(earliestOffer);
  }

  /**
   * Sets the TMEntry for the earliest offer time.
   */
  private void setEarliestOfferTME (Date earliestOffer)
  {
    TimeMap theMap = plan.getMap();
    if (earliestOfferTME != null)
      theMap.deleteEntry(earliestOfferTME);
    earliestOfferTME = new TMEntry(earliestOffer,
				   new TimerAction() {
				       public void action () {
					 setState(AWARD_READY);
				       }
				     });
    theMap.addEntry(earliestOfferTME);
  }

  /**
   * Returns the {@link TMEntry} instance representing the Earliest
   * Offer time in the RFQ.
   */
  public TMEntry getEarliestOfferTME ()
  {
    return earliestOfferTME;
  }

  /**
   * Submits RFQ, collects bids, distributes them as RFQStatusEvents with a
   * label of "NewBid".  State changes are made to happen as
   * expected.  When the bid deadline specified in the RFQ passes,
   * state is changed to BIDDING_CLOSED, and when all bids are
   * received, the state is changed to BIDS_ACQUIRED.
   * @exception InvalidStateException thrown if the RFQ is null, or if
   *   this status instance is not in the NONE state.
   */
  public void acquireBids ()
    throws InvalidStateException
  {
    if (theRFQ == null)
      throw new InvalidStateException("RFQStatus: the RFQ is null");
    if (!state.equals(NONE))
      throw new InvalidStateException("RFQStatus: incorrect state "
				      + state + ", should be NONE");

    // next, instantiate a bid collector and wrap it in a thread to do
    // the actual bid acquisition.
    new Processor(BidCollectorFactory.getInstance(), theRFQ);
  }

  /**
   * Returns the set of bids received so far for this RFQ.  If the
   * state is SUBMITTED, this list may not be complete.
   */
  public Collection getBids ()
  {
    return theBids;
  }

  /**
   * Adds a bid to the bid collection and issues an RFQStatusEvent
   * with the NewBid label.
   */
  public void addBid (Bid newBid)
  {
    if (state.equals(SUBMITTED)
	|| state.equals(BIDDING_CLOSED)
	|| state.equals(BIDS_ACQUIRED)) {
      this.theBids.add(newBid);
      for (Iterator taskIds = newBid.getTaskSet().iterator();
	   taskIds.hasNext(); ) {
	TaskStatus ts = plan.getTaskStatus((String)taskIds.next());
	ts.addBid(newBid);
      }
      sendStatusEvent(new RFQStatusEvent(this, NEW_BID, null, newBid));
    }
    else
      log.error("Attempt to add bid in state " + state);
  }

  /**
   * Removes a bid from the bid collection.
   * TODO:  Does this need to issue an RFQStatusEvent?  I don't think so...
   * <p>
   * (JEC) This method seems bogus.  Why is it here?  I don't agree
   *   that we should be deleting bids at this level.  See
   *   #0000110. An alternative approach might be to mark the bids
   *   themselves somehow. If you do that, there are several places
   *   where we iterate through bids that would have to be updated to
   *   look at the marks. Examples include looking at the bids for an
   *   RFQ, and at the bids for a particluar task. Another alternative
   *   would be to use a more 
   *   general constraint representation, but you would still have to
   *   be able to find and evaluate the constraints from the bids to
   *   be able to process the bid list in the several places where
   *   that's done.
   */
  public void deleteBid (Bid delBid)
  {
    int index;

    if (this.theBids != null) {
      index = this.theBids.indexOf(delBid);
      if ((index >= 0) && (index < this.theBids.size())) {
	this.theBids.remove(index);
	for (Iterator taskIds = delBid.getTaskSet().iterator();
	     taskIds.hasNext(); ) {
	  TaskStatus ts = plan.getTaskStatus((String)taskIds.next());
	  ts.removeBid(delBid);
      }
      }
    } // end of if - then else for valid removal
    return;
  }

  /**
   * Checks for bid coverage. Returns true just in case all tasks in
   * this RFQ are covered by bids.
   */
  public boolean isBidCoverage ()
  {
    if (theRFQ == null)
      return false;
    for (Iterator tasks = theRFQ.getTaskPlan().getTasks().iterator();
	 tasks.hasNext(); ) {
      TaskStatus ts = plan.getTaskStatus((SubtaskRequest)tasks.next());
      if (ts.getBids().size() == 0)
	return false;
    }
    return true;
  }

  /**
   * We do the actual bid collecting with an inner class, so we can
   * return immediately
   */
  private class Processor
    extends Thread
  {
    private BidCollector theCollector = null;
    private RFQ theRFQ = null;

    Processor (BidCollector collector, RFQ rfq)
    {
      super();
      theCollector = collector;
      theRFQ = rfq;
      this.start();
    }

    public void run ()
    {
      log.debug("Submitting RFQ");
      setState(SUBMITTED);
      Collection theBids = theCollector.collectBids(theRFQ);
      // wait for state change before distributing bids.  If you want
      // to run experiments without waiting for the BIDDING_CLOSED
      // timeout, you have to set the RFQStatus.Bids.DontWait property
      // to "true".
      Boolean dontWait
	  = new Boolean(MagnetProperties.getProperty("RFQStatus.Bids.DontWait"));
      if (dontWait.booleanValue()) {
	cancelBidDeadlineTME();
	setState(BIDDING_CLOSED);
      }
      else {
	waitForStateChange(SUBMITTED);
      }
      // distribute bids, then set state to BIDS_ACQUIRED
      for (Iterator bids = theBids.iterator(); bids.hasNext(); ) {
	Bid bid = (Bid)bids.next();
	log.debug("Adding bid " + bid.getId());
	addBid(bid);
      }
      setState(BIDS_ACQUIRED);
    }
  }
}
