/*------------------------------------------------------------------------
 * MAGNET xx component
 * $Id: TMEntry.java,v 1.5.2.1 2002/05/05 19:41:51 ashutosh Exp $
 * Copyright (c) 2001 Regents of the University of Minnesota.
 * All Rights Reserved by the Regents of the University of Minnesota.
 * This file contains Original Code and/or Modifications of Original
 * Code as defined in the MAGNET Public Source License, which is
 * available at http://www.cs.umn.edu/magnet/source_license_v1.txt.
 * You may not use this file except in compliance with the License.
 *------------------------------------------------------------------------
 * Revision History:
 * 
 * Tue May  8 12:45:42 2001  John Collins  #0000064
 *   Add removePredecessor method.
 * 
 * Sun Apr 15 13:11:41 2001  John Collins
 *   Initial Entry
 */

package edu.umn.magnet.client.status;

import edu.umn.magnet.util.time.AgentTime;
import edu.umn.magnet.util.time.TimerAction;

import org.apache.log4j.Category;

import java.util.Date;
import java.util.Collection;
import java.util.ArrayList;
import java.util.Iterator;

/**
 * These are entries in the TimeMap.  Each has a set of predecessors
 * with associated delays, and a set of successors.  The "expiration"
 * time for an entry is the latest of the expiration times of its
 * predecessors, plus the associated delays.
 * <p>
 * An entry may be in one of three states:
 * <table>
 * <tr><td><strong>PENDING</strong></td>
 *     <td>timeout period has not yet started.</td></tr>
 * <tr><td><strong>CURRENT</strong></td>
 *     <td>this entry's time is currently running.</td></tr>
 * <tr><td><strong>COMPLETE</strong></td>
 *     <td>this entry has expired.  The <tt>completionTime</tt> property
 *         gives the actual completion time.</td></tr>
 * </table>
 * @author John Collins
 * @version $Revision: 1.5.2.1 $
 */
public class TMEntry
{
  // State values
  public static final int PENDING = 0;
  public static final int CURRENT = 1;
  public static final int COMPLETE = 2;

  private static Category log
    = Category.getInstance(TMEntry.class.getName());

  // ----------------------- instance data --------------------------
  private int state = PENDING;

  // each entry in the predecessors list is an instance of the inner
  // class Predecessor, containing a TMEntry and a delay interval.
  private ArrayList predecessors = new ArrayList();

  // here's the action object
  private TimerAction action = null;

  // expected and actual expiration times, and validity stamp.  These
  // are actually dates on the AgentTime scale - you can turn one into
  // a Date with new Date (value);  If this entry has predecessors and
  // is not COMPLETE, then the expectedExpiration is valid just in
  // case the validity stamp is at least as great as the highest
  // validity stamp on any of its predecessors.
  private long actualExpiration = -1L;
  private long expectedExpiration = -1L;
  private long validityStamp = -1L;

  // -------------------- constructors --------------------
  /**
   * Default constructor.  If you use this, you need to set the
   * predecessors and the action separately.
   */
  public TMEntry ()
  {
    super();
  }

  /**
   * Constructor for an entry with no predecessors.  The expected
   * expiration time is set directly.
   */
  public TMEntry (Date expiration, TimerAction action)
  {
    super();
    this.expectedExpiration = expiration.getTime();
    setAction(action);
  }

  /**
   * Complete constructor for an entry with a single predecessor.  If
   * you need multiple predecessors, use addPredecessor for the rest.
   */
  public TMEntry (TMEntry pred, long interval, TimerAction action)
  {
    super();
    addPredecessor(pred, interval);
    setAction(action);
  }

  // ------------------ instance query and manipulation -----------------
  /**
   * Adds a predecessor to this entry.  Doing so cannot change a
   * COMPLETE entry to a PENDING one, since its action will have
   * already taken place.  This method can be called in two different
   * contexts:
   * <ol>
   * <li>From the context where the TMEntry is being created, before
   * it is made part of a TimeMap.  In this case, no recomputation of
   * expiration time is possible.
   * <li>From a TimeMap, (see {@link
   * TimeMap#addPredecessor(TMEntry,TMEntry,long) 
   * TimeMap.addPredecessor}) which is responsible for re-computing
   * the expiration time of this entry, and for determining whether a
   * state change is required.
   * </ol>
   * If this method is called from outside the TimeMap after this
   * entry has been added to the TimeMap, the new precedence relation
   * will likely have no effect.
   */
  public void addPredecessor (TMEntry pred, long interval)
  {
    predecessors.add(new Predecessor(pred, interval));
    validityStamp = -1;
  }

  /**
   * Returns the predecessor list for this entry.
   * @return a {@link java.util.Collection Collection} of
   *   {@link TMEntry.Predecessor TMEntry.Predecessor} instances.
   */
  public Collection getPredecessors ()
  {
    return this.predecessors;
  }

  /**
   * Removes a TMEntry from the predecessor list of this entry.
   * Silently does nothing if oldPredecessor is not actually a member
   * of the predecessor list.
   */
  public void removePredecessor (TMEntry oldPredecessor)
  {
    Predecessor theOne = null;
    Iterator preds = predecessors.iterator();
    while (preds.hasNext() && theOne == null) {
      Predecessor pre = (Predecessor)preds.next();
      if (pre.pred == oldPredecessor) {
	theOne = pre;
      }
    }
    if (theOne != null) {
      predecessors.remove(theOne);
    }
    validityStamp = -1;
  }
  
  /**
   * Sets the action for this entry.  This will of course have no
   * impact if this entry is already complete.
   */
  public void setAction (TimerAction newAction)
  {
    this.action = newAction;
  }

  /**
   * Returns the state of this entry.
   */
  public int getState ()
  {
    return state;
  }

  /**
   * True just in case this entry is COMPLETE.
   */
  public boolean isComplete ()
  {
    return (state == COMPLETE);
  }

  /**
   * Makes this entry complete, by running its action and updating the
   * state.  TBD: decide whether to make the actual expiration time be
   * the current AgentTime, or the expected expiration time.
   */
  protected void makeComplete ()
  {
    log.debug("makeComplete " + this);
    actualExpiration = getExpirationTime();
    state = COMPLETE;
    if (action != null)
      action.action();
  }

  /**
   * Makes this entry CURRENT
   */
  protected void makeCurrent ()
  {
    log.debug("makeCurrent " + this);
    state = CURRENT;
  }

  /**
   * Makes this entry PENDING
   */
  protected void makePending ()
  {
    log.debug("makePending " + this);
    state = PENDING;
  }

  /**
   * An entry is valid if it is COMPLETE, has no predecessors, or if
   * its stamp is equal to the current stamp.
   */
  private boolean isValid ()
  {
    if (isComplete()
	|| predecessors.size() == 0
	|| getStamp() == TMEntry.currentStamp())
      return true;
    else
      return false;
  }

  /**
   * Returns the stamp on this entry.
   */
  private long getStamp ()
  {
    return this.validityStamp;
  }

  /**
   * Returns the expiration time for this entry.  If this entry is
   * COMPLETE, then returns the actual completion time, otherwise
   * (potentially) computes and returns the expected completion time.
   */
  public long getExpirationTime ()
  {
    if (isComplete())
      return actualExpiration;
    else if (!isValid()) {
      // here we compute the expected expiration time, and update the
      // stamp so we can possibly avoid computation next time.
      for (Iterator preds = predecessors.iterator(); preds.hasNext(); ) {
	Predecessor pred = (Predecessor)preds.next();
	long exp = (pred.getEntry().getExpirationTime() 
		    + pred.getInterval());
	if (exp > expectedExpiration) {
	  expectedExpiration = exp;
	}
      }
      // we know it's valid now...
      validityStamp = TMEntry.currentStamp();
    }
    return expectedExpiration;
  }

  // -------------------- static timestamp generator -----------------
  private static long s_stamp = 0;

  /**
   * Increments the value of the stamp, thereby invalidating all
   * cached expiration times.  Returns the value of the updated
   * stamp.
   */
  protected static long updateStamp ()
  {
    return ++s_stamp;
  }

  /**
   * Returns the value of the latest stamp issued.  This is not a very
   * strong validity test, especially if there are multiple active
   * time maps in the running system.
   */
  protected static long currentStamp ()
  {
    return s_stamp;
  }

  // -------------------- predecessor data holder ------------------
  /**
   * Instances of this class are used as the entries in the
   * predecessor list of a TMEntry.  Each contains a TMEntry and an
   * interval.  The meaning is that the "host" TMEntry will expire at
   * least as late as the expiration time of the Predecessor entry
   * plus the associated interval.
   * <p>
   * <strong>NOTE:</strong> These are intended to be read-only outside
   * the scope of the TimeMap itself.
   */
  public class Predecessor
  {
    TMEntry pred = null;
    long interval = -1L;

    Predecessor (TMEntry predecessor, long interval)
    {
      this.pred = predecessor;
      this.interval = interval;
    }

    /**
     * Returns the TMEntry instance.
     */
    public TMEntry getEntry ()
    {
      return this.pred;
    }

    /**
     * Returns the interval.
     */
    public long getInterval ()
    {
      return this.interval;
    }

    /**
     * Sets the interval.  After this entry is in the TimeMap, this
     * should only be done by the TimeMap itself, 
     * which is responsible for computing the impact of an interval
     * change.  In any case, changing the interval cannot unCOMPLETE an
     * entry.
     */
    void setInterval (long newValue)
    {
      this.interval = newValue;
      TMEntry.updateStamp();
    }
  }
}
