/*------------------------------------------------------------------------
 * MAGNET agent status component
 * $Id: TimeMap.java,v 1.5.2.1 2002/05/05 19:41:51 ashutosh Exp $
 * Copyright (c) 2001 Regents of the University of Minnesota.
 * All Rights Reserved by the Regents of the University of Minnesota.
 * This file contains Original Code and/or Modifications of Original
 * Code as defined in the MAGNET Public Source License, which is
 * available at http://www.cs.umn.edu/magnet/source_license_v1.txt.
 * You may not use this file except in compliance with the License.
 *------------------------------------------------------------------------
 * Revision History:
 * 
 * Tue May  8 12:18:26 2001  John Collins  #0000064
 *   Add deleteEntry method
 * 
 * Sat Mar 31 22:21:49 2001  John Collins
 *   Initial Entry
 */

package edu.umn.magnet.client.status;

import edu.umn.magnet.util.time.AgentTime;

import org.apache.log4j.Category;

import java.util.Collection;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.Iterator;

/**
 * This is the basic Time Map structure.  A Time Map contains a set of
 * TMEntry instances.  The set of these that have no future
 * predecessors are "heads".  The one with the closest expiration time
 * is the "next" entry.  When a TimeMap is "active" and has a
 * non-empty set of entries, then there is always a thread active,
 * waiting for the next event time.  When that time arrives, the
 * entry's action is invoked, the earliest of the heads is chosen as
 * next, and the timer is started for it.  Time is derived from
 * AgentTime, which may be wall-clock time or some other time, and may
 * pass in real time, or more slowly or quickly than real time.
 * <p>
 * An obvious question is, Is this a singleton class, or is there one
 * per plan, or do we use some other instantiation scheme?  This implementation
 * intends the one-per-plan approach.  If you trash a plan, you can
 * just dump the whole time map and not have to clean out the items
 * for that plan from the singleton map.  You can generate a report on
 * the status of a plan without having to filter info.  On the other
 * hand, if you want to know what events are expected in the next 10
 * minutes, you have to query all the maps.  </p>
 * @author John Collins
 * @version $Revision: 1.5.2.1 $ */
public class TimeMap extends Thread
{
  private static Category log
    = Category.getInstance(TimeMap.class.getName());

  // Here's the actual map - it's possible that we may decide to keep
  // it in sorted order by time, but that would require action every
  // time some event occurred that could potentially change the time
  // for some entry.
  private LinkedList heads = new LinkedList();

  // This is the TMEntry that is currently waiting for its time.  This
  // is the one you need to cancel if you put a new item on the front
  // of the map list.
  private TMEntry currentEntry = null;

  // The TimeMap keeps track of successor relations among TMEntries.
  // Each entry in successors is a TMEntry with its (unsorted) set of
  // successors.
  private HashMap successors = new HashMap();

  // True just in case the thread is running
  private boolean running = false;

  /**
   * All we need is a default constructor.  This is expected to be
   * called by the PlanStatus constructor, and you access it through
   * the PlanStatus instance.
   */
  public TimeMap ()
  {
    super();
    this.start();
  }

  // -------- population and maintenance of the TimeMap ----------

  /**
   * Adds a new head entry to the map.  We assume that any necessary
   * constraints are already contained in the new entry.  If any of
   * those constraints refer to TMEntry objects that are not already
   * in the map, they will be added before the specified entry is
   * added.
   * <p>
   * If the new entry expires sooner than the current <i>next</i>
   * entry, then the timeout for <i>next</i> is canceled, <i>next</i>
   * is moved to the head list, and the new entry becomes <i>next</i>.
   * </p><p>
   * If the new entry is already registered with this map, then no
   * action is taken.
   */
  public synchronized void addEntry (TMEntry newEntry)
  {
    log.debug("addEntry " + newEntry);

    // First, check to see whether we already know about this one.
    if (successors.get(newEntry) != null)
      return;

    // Second, we make a first walk through the predecessors, and make
    // sure they are all registered.  Then add the new successor
    // relation for each of them.  As we walk the list, we also
    // determine whether there are any non-COMPLETE predecessors
    boolean allComplete = true;
    for (Iterator preds = newEntry.getPredecessors().iterator();
	 preds.hasNext(); ) {
      TMEntry pred = ((TMEntry.Predecessor)preds.next()).getEntry();
      addEntry(pred);
      if (!pred.isComplete())
	allComplete = false;
      Collection succ = (Collection)successors.get(pred);
      if (! succ.contains(newEntry))
	succ.add(newEntry);
    }

    // Create a new empty successors list for the new entry.
    successors.put(newEntry, new HashSet());

    // Finally, if this entry is not COMPLETE, check whether new entry
    // should be a head or current.
    if (!newEntry.isComplete()) {
      // it's a head if it has no predecessors, or if all its
      // predecessors are COMPLETE
      if (newEntry.getPredecessors().isEmpty()
	  || allComplete) {
	heads.add(newEntry);
	// if it's a head, it could also precede the current entry.
	checkCurrent(newEntry);
	notify();
      }
    }
  }

  /**
   * Removes an entry from the TimeMap.  If it's already COMPLETE, we
   * don't need to do much, other than make it not be a predecessor
   * for any other entry.  If it's current, we have to update the
   * current entry.
   */
  public synchronized void deleteEntry (TMEntry oldEntry)
  {
    // if it has successors, remove it from the predecessor list of
    // each successor
    for (Iterator succs = ((HashSet)successors.get(oldEntry)).iterator();
	 succs.hasNext(); ) {
      TMEntry succ = (TMEntry)succs.next();
      succ.removePredecessor(oldEntry);
    }

    // if it's COMPLETE, that's all we have to do.
    if (!oldEntry.isComplete()) {
      // otherwise, if it has predecessors, delete it from each of their
      // successor lists.
      Collection predecessors = oldEntry.getPredecessors();
      if (predecessors != null) {
	for (Iterator preds = predecessors.iterator();
	     preds.hasNext(); ) {
	  TMEntry pred = ((TMEntry.Predecessor)preds.next()).getEntry();
	  HashSet succs = (HashSet)successors.get(pred);
	  succs.remove(oldEntry);
	}
      }
      // remove from list of heads
      heads.remove(oldEntry);

      // if it's current, abort and update
      if (currentEntry == oldEntry) {
	cancelCurrent();
      }
    }
  }

  /**
   * Adds a predecessor relation between two TMEntry objects.  If
   * either of the two entries is not already in the map, they will be
   * added.
   */
  public synchronized void addPredecessor (TMEntry predecessor,
					   TMEntry successor,
					   long interval)
  {
    addEntry(predecessor);
    successor.addPredecessor(predecessor, interval);
    // if successor is not already in this map, then addEntry will do
    // the right thing
    if (successors.get(successor) == null)
      addEntry(successor);
    // otherwise we have to check head and current status here - but
    // only if successor is not already COMPLETE
    else if (!successor.isComplete()) {
      // keep track of successor relationships
      HashSet succs = (HashSet)successors.get(predecessor);
      succs.add(successor);

      // if the new predecessor is COMPLETE and is the only
      // predecessor, then successor could possibly be moved up from
      // head to current. 
      if (predecessor.isComplete()
	  && successor.getPredecessors().size() == 1) {
	checkCurrent(successor);
      }
      // otherwise, if the new predecessor is a head, then the
      // successor cannot be current or head.
      else if (heads.contains(predecessor)) {
	if (currentEntry == successor)
	  cancelCurrent();
	else
	  heads.remove(successor);
      }
    }
  }

  /**
   * Returns <tt>true</tt> just in case there are no future entries in
   * the map.
   */
  public synchronized boolean isEmpty ()
  {
    return (currentEntry == null && heads.isEmpty());
  }

  /**
   * Checks to see if a particular head entry should be current.  This
   * method needs to be called from a synchronized method.
   * @return true just in case the existing current entry was
   *   canceled.
   */
  private boolean checkCurrent (TMEntry entry)
  {
    if ((currentEntry != null)
	&& (entry.getExpirationTime()
	    < currentEntry.getExpirationTime())) {
      heads.add(currentEntry);
      cancelCurrent();
      return true;
    }
    else
      return false;
  }

  /**
   * Cancels the timeout on the current entry.  The map thread will
   * then be forced to recompute current and resume.  We kill the
   * currentEntry, but make no attempt here to add the currentEntry
   * back into the heads list.  This method must be called from a
   * synchronized method.
   */
  private void cancelCurrent ()
  {
    if (currentEntry != null) {
      currentEntry.makePending();
      currentEntry = null;
      interrupt();
    }
  }

  // -------- Run the time forward and process entries ----------

  /**
   * Returns true if the TimeMap's thread is running.
   */
  public synchronized boolean isRunning()
  {
    return running;
  }

  /**
   * Stops the thread safely.  Currently there is no way to re-start
   * it. 
   */
  public synchronized void halt ()
  {
    running = false;
    interrupt();
  }

  /**
   * Sets up the next entry, waiting for it if necessary.
   */
  private synchronized void getNextEntry ()
    throws InterruptedException
  {
    while (heads.isEmpty()) {
	wait();
    }

    TMEntry earliest = null;
    for (Iterator hds = heads.iterator(); hds.hasNext(); ) {
      TMEntry hd = (TMEntry)hds.next();
      if ((earliest == null)
	  || (hd.getExpirationTime() < earliest.getExpirationTime()))
	earliest = hd;
    }
    heads.remove(earliest);
    currentEntry = earliest;
    log.debug("next entry: " + earliest);
    earliest.makeCurrent();
  }

  /**
   * Checks each of the candidates for head status.
   */
  private synchronized void promoteHeads (Collection candidates)
  {
    for (Iterator cds = candidates.iterator(); cds.hasNext(); ) {
      TMEntry candidate = (TMEntry)cds.next();
      if (! heads.contains(candidate)) {
	boolean allComplete = true;
	for (Iterator preds = candidate.getPredecessors().iterator();
	     preds.hasNext() && allComplete; ) {
	  if (!((TMEntry.Predecessor)preds.next()).getEntry().isComplete())
	    allComplete = false;
	}
	if (allComplete) {
	  log.debug("add head " + candidate);
	  heads.add(candidate);
	}
      }
    }
  }

  /**
   * Here's the thread.  It waits for the time of the next entry, then
   * runs its action and moves on to the next.  If there are no
   * active entries, it waits for one to be added.
   */
  public void run ()
  {
    running = true;
    while(isRunning()) {
      try {
	getNextEntry();
	// we are guaranteed at this point that currentEntry is now the
	// earliest non-COMPLETE entry in the map.
	log.debug("start timeout for " + currentEntry);
	AgentTime.waitUntil(currentEntry.getExpirationTime());
	// here's where we have actually completed a timeout.  Run the
	// action and start up the next one.
	currentEntry.makeComplete();
	promoteHeads((HashSet)successors.get(currentEntry));
	currentEntry = null;
      }
      catch (InterruptedException ie) {
	// get here when a timeout is canceled - there's nothing that
	// needs doing with the possible exception of saying so
	log.debug("interrupted");
      }
    }
  }
}
