/*------------------------------------------------------------------------
 * MAGNET unit test
 * $Id: TimeMapTest.java,v 1.3.2.1 2002/05/05 19:41:53 ashutosh Exp $
 * Copyright (c) 2000 Regents of the University of Minnesota.
 * All Rights Reserved by the Regents of the University of Minnesota.
 * This file contains Original Code and/or Modifications of Original
 * Code as defined in the MAGNET Public Source License, which is
 * available at http://www.cs.umn.edu/magnet/source_license_v1.txt.
 * You may not use this file except in compliance with the License.
 *------------------------------------------------------------------------
 * Revision History:
 * 
 * Tue May  8 13:06:40 2001  John Collins  #0000064
 *   Test entry removal
 * 
 * Mon Apr 16 21:43:51 2001  John Collins
 *   Initial Entry
 */

package edu.umn.magnet.client.status;

import edu.umn.magnet.util.time.AgentTime;
import edu.umn.magnet.util.time.TimerAction;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestResult;
import junit.framework.TestSuite;

import java.util.Date;
import java.util.LinkedList;
import java.util.Iterator;

/**
 * This is a tester for edu.umn.magnet.client.status.TimeMap
 * @author John Collins
 */
public class TimeMapTest extends TestCase
{
  
  public TimeMapTest (String testName)
  {
    super(testName);
  }

  /**
   * You run the main method to exercise this thing.  There are no
   * command-line args.
   */
  public static void main (String[] args)
  {
    junit.textui.TestRunner.run(suite());
  }

  public static Test suite ()
  {
    return new TestSuite(TimeMapTest.class);
  }
  
  // ----------- test data ------------

  public static final long BASE = new Date().getTime();
  public static final long SECOND = 1000L; // 1 second
  public static final long SLOP = 40; // time slop = 40 msec

  private TimeMap theMap = new TimeMap();
  private LinkedList completions = new LinkedList(); // completion values
  private final TimeMapTest tmt = this;

  // -------------- setup methods ---------------
  private void setup1 ()
  {
    AgentTime.setDate(new Date(BASE));
    completions.clear();
  }

  // ------------- test methods ------------
  // single entry
  public void testSingleEntry ()
  {
    setup1();
    theMap.addEntry(new TMEntry(new Date(BASE + SECOND),
				makeNotifyAction("A")));
    try {
      Thread.sleep(SECOND * 3L);
    }
    catch (InterruptedException ie) {
      // ignore
    }
    assertEquals("one completion", 1, completions.size());
    CompletionData item = getNextCompletion();
    assertEquals("correct value", "A", item.content);
    assert("TimeMap now empty", theMap.isEmpty());
    //System.out.println("Slop = " + (item.when.getTime() - BASE - SECOND));
    assert("correct completion time", 
	   (Math.abs(item.when.getTime() - BASE - SECOND)
	    < SLOP));
    assert("no more completions", completions.isEmpty());
  }

  // two independent entries
  public void testTwoEntries ()
  {
    setup1();
    TMEntry te1 = new TMEntry(new Date(BASE + SECOND/2L),
			      makeNotifyAction("A"));
    theMap.addEntry(te1);
    TMEntry te2 = new TMEntry(new Date(BASE + (SECOND/2L)*3L),
			      makeNotifyAction("B"));
    theMap.addEntry(te2);
    try {
      Thread.sleep(SECOND);
    }
    catch (InterruptedException ie) {
      // ignore
    }
    assertEquals("one completion", 1, completions.size());
    CompletionData item = getNextCompletion();
    assertEquals("correct value", "A", item.content);
    assert("TimeMap not empty", !theMap.isEmpty());

    try {
      Thread.sleep(SECOND);
    }
    catch (InterruptedException ie) {
      // ignore
    }
    assertEquals("one completion", 1, completions.size());
    item = getNextCompletion();
    assertEquals("correct value", "B", item.content);
    assert("TimeMap empty", theMap.isEmpty());
  }

  // Two entries, b->a
  public void testPredecessor ()
  {
    setup1();
    TMEntry te1 = new TMEntry(new Date(BASE + SECOND/2L),
			      makeNotifyAction("A"));
    theMap.addEntry(te1);
    TMEntry te2 = new TMEntry(te1, SECOND,
			      makeNotifyAction("B"));
    theMap.addEntry(te2);
    try {
      Thread.sleep(SECOND);
    }
    catch (InterruptedException ie) {
      // ignore
    }
    assertEquals("one completion", 1, completions.size());
    CompletionData item = getNextCompletion();
    assertEquals("correct value", "A", item.content);
    assert("TimeMap not empty", !theMap.isEmpty());

    try {
      Thread.sleep(SECOND);
    }
    catch (InterruptedException ie) {
      // ignore
    }
    assertEquals("one completion", 1, completions.size());
    item = getNextCompletion();
    assertEquals("correct value", "B", item.content);
    assert("TimeMap empty", theMap.isEmpty());
  }

  public void testRemove1 ()
  {
    setup1();
    TMEntry te1 = new TMEntry(new Date(BASE + SECOND/2L),
			      makeNotifyAction("A"));
    theMap.addEntry(te1);
    TMEntry te2 = new TMEntry(new Date(BASE + (SECOND/2L)*3L),
			      makeNotifyAction("B"));
    theMap.addEntry(te2);
    theMap.deleteEntry(te1); // here's the removal
    try {
      Thread.sleep(SECOND);
    }
    catch (InterruptedException ie) {
      // ignore
    }
    assertEquals("no completions", 0, completions.size());
    assert("TimeMap not empty", !theMap.isEmpty());

    try {
      Thread.sleep(SECOND);
    }
    catch (InterruptedException ie) {
      // ignore
    }
    assertEquals("one completion", 1, completions.size());
    CompletionData item = getNextCompletion();
    assertEquals("correct value", "B", item.content);
    assert("TimeMap empty", theMap.isEmpty());
  }

  // ------------- helper methods ---------------

  private TimerAction makeNotifyAction (Object result)
  {
    final Object resultHolder = result;
    return new TimerAction()
      {
	public void action () {
	  tmt.notifyCompletion(AgentTime.getDate(), resultHolder);
	}
      };
  }

  private synchronized CompletionData getNextCompletion ()
  {
    return (CompletionData)completions.removeFirst();
  }

  private synchronized void notifyCompletion (Date when, Object result)
  {
    completions.add(new CompletionData(when, result));
    //System.out.println("Completion at " + AgentTime.getDate());
  }

  // ---------- data holder for completion notifications ----------
  private class CompletionData
  {
    Date when = null;
    Object content = null;

    CompletionData (Date when, Object content)
    {
      super();
      this.when = when;
      this.content = content;
    }
  }
}
