/*------------------------------------------------------------------------
 * Magnet client components
 * $Id: RandomPlanFactory.java,v 1.8.2.1 2002/05/05 19:41:53 ashutosh Exp $
 * Copyright (c) 1998 Regents of the University of Minnesota.
 * All Rights Reserved by the Regents of the University of Minnesota.
 * This file contains Original Code and/or Modifications of Original
 * Code as defined in the MAGNET Public Source License, which is
 * available at http://www.cs.umn.edu/magnet/source_license_v1.txt.
 * You may not use this file except in compliance with the License.
 *------------------------------------------------------------------------
 * Revision History:
 *
 * Fri Jan 25 12:44:00 2002 Anne Schoolcraft
 *   Fixing bug #171 -- make random plans start with the current AgentTime, not
 *    the beginning of Unix time (ie 1970)
 *
 * 25May2001 DFK - Bug #0000100: option to gen random tasks w/ non-null fields.
 *
 * Fri Apr 20 23:46:56 2001  John Collins
 *   Eliminate use of system properties
 *
 * Sun Feb 25 13:09:19 2001  John Collins
 *   Update for new component structure.
 *
 * Wed Sep 27 12:35:10 2000  John Collins
 *   PlanFactory should just create a plan, not set the RFQ time
 *   windows.
 *
 * Tue Jan 12 14:48:05 1999  Tim Lee
 *   Changed the use of Math.random to use RVFManager instead.
 *
 * Sun Jan 10 14:08:56 1999  John E. Collins
 *   Replace interconnection algorithm with one that creates uniform
 *   interconnection density.
 *
 * Wed Jan  6 09:55:36 1999  John E. Collins
 *   Make the makePlan method static, extend PlanFactory
 *
 * Sun Jan  3 14:59:49 1999  John E. Collins
 *   Use system property to specify branching factor
 *
 * Sat Jan  2 13:54:30 1999  John E. Collins
 *   Move to plan package
 *
 * Sep 29, 1998 - Max Tsvetovat
 *   Initial Entry
 */

package edu.umn.magnet.client.plan;

import java.util.Date;
import java.util.Iterator;
import java.util.Vector;

import edu.umn.magnet.protocol.cust.TaskPlan;
import edu.umn.magnet.protocol.cust.SubtaskRequest;
import edu.umn.magnet.protocol.cust.TaskWithPrecedence;
import edu.umn.magnet.protocol.InvalidDataException;
import edu.umn.magnet.protocol.JobInfo;

import edu.umn.magnet.util.rvars.UniformVariable;
import edu.umn.magnet.util.rvars.RVFManager;
import edu.umn.magnet.util.rvars.RandomVariableFactory;
import edu.umn.magnet.util.MagnetProperties;
import edu.umn.magnet.util.time.AgentTime;

import org.apache.log4j.Category;

/*
 * RandomPlanFactory creates task plans that have randomized precedence
 * relations with parametrized mean branching factor
 *
 * The plan is constructed as an Acyclic Directed Graph, and then fed into
 * PlanCPMEval algorithm to put start and finish dates in using CPM algo.
 *
 * @author Max Tsvetovat
 * @version $Revision: 1.8.2.1 $
 */
public class RandomPlanFactory extends PlanFactory
{
  // init the log category
  static Category log =
    Category.getInstance(RandomPlanFactory.class.getName());

  /**
   * <tt>makePlan</tt> constructs a Randomized TaskPlan
   * arguments: number of tasks (int), StartDate (java.util.Date),
   * planSlack(double, > 0) - percentage of total work time to add to slack
   * cpmSlack (double, 0 < cpmSlack < 1.0) - percentage of task size
   * contraction for increasing task overlap
   * Branching factor defaults to nTasks/10
   */
  public static TaskPlan makePlan(int nTasks) {
    return makePlan(nTasks, false);
  }


  /**
   * <tt>makePlan</tt> constructs a Randomized TaskPlan
   * arguments: number of tasks (int), StartDate (java.util.Date),
   * planSlack(double, > 0) - percentage of total work time to add to slack
   * cpmSlack (double, 0 < cpmSlack < 1.0) - percentage of task size
   * contraction for increasing task overlap
   * Branching factor defaults to nTasks/10
   */
  public static TaskPlan makePlan(int nTasks, boolean fullySpecify)
  { // First, extract the branch factor from system properties
    log.debug("RandomPlanFactory.makePlan(" + nTasks + ")");
    String bf = MagnetProperties.getProperty("PlanBranchFactor");
    double branchFactor = (double)(nTasks / 10.0);
    if (bf != null) {
      try {
	branchFactor = Double.valueOf(bf).doubleValue();
	log.debug("Random plan branch factor = " + bf);
      }
      catch (NumberFormatException nfe) {
	log.error("Error in specification of branch factor: " + bf
		  + ", using " + branchFactor);
      }
    }
    double linkProbability = branchFactor / (nTasks - 1);

    // Then proceed to construct the plan
    TaskPlan plan = new TaskPlan();
    if (fullySpecify) {
      plan.setPlanName("A new plan");
    }
    SubtaskRequest[] tasks = new SubtaskRequest[nTasks];

    //Create an ordered list of tasks and add the tasks to the plan
    for (int i = 0; i < nTasks; i++) {
      JobInfo type
	  = ((JobInfoFactory)JobInfoFactory.getInstance()).find();
      SubtaskRequest task = new SubtaskRequest(type.getTaskType()
					       + String.valueOf(i));
      task.setJobInfo(type);
      // Bug #100: Allow all member fields to be non-null..
      if (fullySpecify) {
        try {
          //set start time to 24 hrs after current time
          task.setEarliestTimeToStart(new Date(AgentTime.getDate().getTime()+86400000));
          log.debug("Set random plan start time to "+ task.getEarliestTimeToStart());
          //add the duration to the start time to get the deadline
          task.setDeadline(new Date(task.getEarliestTimeToStart().getTime() +
            (long)type.getExpectedDuration()));
        } catch (InvalidDataException ide) {
        }
      }
      tasks[i] = task;
      try {
	plan.addTask(task);
      }
      catch (Exception e) {
	e.printStackTrace();
	return null;
      }
    }
    // Create precedence relations inside the list.
    // Each node has a probability of linkProbability of being linked
    // to each succeeding node.

    // Create a UniformVariable called uVariable to produce random
    // values between 0 and 1 using the "PlanSeed" RVF.

    RandomVariableFactory rvFactory=
        RVFManager.getRVF("PlanSeed");
    UniformVariable uVariable=rvFactory.createUniformVariable(0,1);

    for (int linkFrom = 0; linkFrom < (nTasks - 1); linkFrom++) {
      for (int linkTo = linkFrom + 1; linkTo < nTasks; linkTo++) {
	if (uVariable.nextDouble() < linkProbability) {

	  // We're going to make linkTo a successor of linkFrom.
	  // Before we do that, we'll remove common predecessors of
	  // linkTo and linkFrom from the precedessor list of linkTo
	  TaskWithPrecedence linkToTWP = plan.getTWP(tasks[linkTo]);
	  Vector existing = linkToTWP.getPredecessors();
	  // We have to collect the list to be removed before
	  // operating on them to avoid smashing the Iterator.
	  Vector remove = new Vector();
	  for (Iterator preds = plan.getPredecessors(tasks[linkFrom]);
	       preds.hasNext(); ) {
	    SubtaskRequest pred = (SubtaskRequest)preds.next();
	    TaskWithPrecedence predTWP = plan.getTWP(pred);
	    if (existing.indexOf(predTWP) != -1)
	      remove.add(predTWP);
	  }
	  // When we remove a link, we have to do both ends.  The TWP
	  // abstraction wasn't really intended to support this
	  // operation.
	  for (Iterator outs = remove.iterator();
	       outs.hasNext(); ) {
	    TaskWithPrecedence out =
		(TaskWithPrecedence)outs.next();
	    log.debug("removing link: " + out.getTaskId()
		      + " to " + linkToTWP.getTaskId());
	    out.removeSuccessor(linkToTWP);
	    linkToTWP.removePredecessor(out);
	  }

	  // Now we actually create the new link
	  try {
	    log.debug("adding link: " + tasks[linkFrom].getId()
		      + " to " + tasks[linkTo].getId());
	    plan.addSuccessor(tasks[linkFrom], tasks[linkTo]);
	  }
	  catch (Exception e) {
	    e.printStackTrace();
	    return null;
	  }
	}
      }
    }
    return plan;
  }
}
