/*------------------------------------------------------------------------
 * MAGNET supplier simulation components
 * $Id: ContiguousBidder.java,v 1.3.2.1 2002/05/05 19:41:54 ashutosh Exp $
 * Copyright (c) 2000 Regents of the University of Minnesota.
 * All Rights Reserved by the Regents of the University of Minnesota.
 * This file contains Original Code and/or Modifications of Original
 * Code as defined in the MAGNET Public Source License, which is
 * available at http://www.cs.umn.edu/magnet/source_license_v1.txt.
 * You may not use this file except in compliance with the License.
 *------------------------------------------------------------------------
 * Revision History:
 * 
 * Thu Oct  5 22:32:28 2000  John Collins
 *   Initial Entry - refactored out of the old XSupplyMonster
 */

package edu.umn.magnet.client.sup;

import edu.umn.magnet.util.rvars.NormalVariable;
import edu.umn.magnet.util.rvars.UniformVariable;

import edu.umn.magnet.protocol.cust.Bid;
import edu.umn.magnet.protocol.cust.RFQ;
import edu.umn.magnet.protocol.cust.SubtaskRequest;
import edu.umn.magnet.protocol.cust.TaskPlan;
import edu.umn.magnet.protocol.InvalidDataException;

import org.apache.log4j.Category;

import java.util.Collection;
import java.util.Properties;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;

/**
 * This thing generates bids on contiguous sets of tasks, starting
 * with either a seed task or a randomly-chosen seed.
 * @author John Collins
 */
public class ContiguousBidder extends BidGenerator
{
  static Category log = Category.getInstance(ContiguousBidder.class.getName());

  private RFQ m_cfb = null;

  private UniformVariable m_uVariable = null;
  private NormalVariable m_normalVariable = null;
  
  // Configuration properties
  private double m_followLinkProbability;
  private double m_followLinkVariability;

  public ContiguousBidder ()
  {
    super();
  }

  protected void init (Properties props, SupplyMonster sm)
  {
    super.init(props, sm);
    m_uVariable = sm.getRVF().createUniformVariable(0,1);
    m_normalVariable = sm.getRVF().createNormalVariable(0,1);
    m_followLinkVariability
	= sm.getPropertyDouble("Sup.FollowLinkVar", 0.5);
    m_followLinkProbability
	= sm.getPropertyDouble("Sup.FollowLinkProb", 0.5);
  }

  protected Collection generateSTBids(SubtaskRequest seedTask) 
  {
    TaskPlan plan = getTaskPlan();
    int numberOfTasks = plan.size();
    HashMap task2spec = new HashMap();  // maps tasks to TaskSpec instances
    ArrayList openList    = new ArrayList();  // task to possibly bid on
    HashSet visitedList = new HashSet();  // tasks we have touched
    HashSet idList = new HashSet(); // IDs of tasks we have touched
    
    if (numberOfTasks == 0) 
    {
      log.error("SupplyMonster.generateTasks(): No tasks in plan");
      return null; // nothing to bid on
    }

    // We generate a number normally distributed about 1.0 as a
    // multiplier for the link-following probability.  We don't let it
    // be less than 0.0.  This is intended to increase the variability
    // of bid sizes.
    double flv = (m_followLinkVariability
		  * m_normalVariable.nextDouble()) + 1.0;
    if (flv < 0.0) flv = 0.0;
    double followLinkProb = m_followLinkProbability * flv;
    if (m_followLinkProbability > 1.0)
      m_followLinkProbability = 1.0;

    // If we weren't given a seed, then find a random task to start
    // bid consideration
    if (seedTask == null) {
      Iterator ste = getTaskPlan().iterator(); // plan's tasks
      int seedIndex   = (int)(m_uVariable.nextDouble()
			      * (numberOfTasks - 1));
      for (int i = 0; i <= seedIndex && ste.hasNext(); i++) {
	seedTask = (SubtaskRequest)ste.next();
      }
    }
    openList.add(seedTask);

    // find and fill tasks to bid on, starting from seedTask
    SubtaskRequest task = null;
    TaskSpec spec = null;
    while (!openList.isEmpty()) {
      task = (SubtaskRequest) openList.remove(0);
      
      // if we haven't touched this task already, process it;
      // else skip it.
      if (! visitedList.contains(task)) {
	visitedList.add(task); // prevent future cycles
	if (idList.contains(task.getId())) {
	  String id = task.getId();
	  log.error("Duplicate taskId " + id);
	  SubtaskRequest planT = getTaskPlan().getTaskData(id);
	  if (planT != seedTask) {
	    log.error("Plan task " + planT.getId()
		      + " not same as seed task " + seedTask.getId());
	  }
	}
	else
	  idList.add(task.getId());

	// If after generating a task's value it becomes infeasable,
	// we don't include this task or any of it's ancestors and
	// simply continue processing the rest of the tasks on the
	// openList.
	
	spec = getSTBidGenerator().generateBid(task,
					       earlyStart(task,
							  task2spec),
					       lateFinish(task,
							  task2spec));
	if (spec != null) {
	  // add this taskbid to the bid vector
	  task2spec.put(task, spec);
	  
	  // choose or don't choose ancestors and children
	  Iterator family = getTaskPlan().getSuccessors(task);
	  SubtaskRequest familytask = null;
	  while (family.hasNext()) {
	    familytask = (SubtaskRequest) family.next();
	    if (m_uVariable.nextDouble() <= followLinkProb) {
	      openList.add(familytask);
	    }
	  }
	  family = getTaskPlan().getPredecessors(task);
	  while (family.hasNext()) {
	    familytask = (SubtaskRequest) family.next();
	    if (m_uVariable.nextDouble() <= followLinkProb) {
	      openList.add(familytask);
	    }
	  }
	}
      } 
    }
    log.debug("ContigBidGen: bid size = " + task2spec.size());
    return task2spec.values();
  }

  /**
   * Sets the probability used in accepting ancestors and 
   * predecessors during taskrq selection.  
   * @param probability new probability to use
   */
  public void setFollowLinkProbability(double probability) 
  {
    if (probability <= 0) {
      log.error("SupplyMonster::setFollowLinkProbability()" +
		": illegal probability ( " + probability + " )");
    }
    else
      m_followLinkProbability = probability;
  }

  /**
   * Get the probability used in accepting ancestors and 
   * predecessors during taskrq selection.
   * @return current probability of acception family
   */
  public double getFollowLinkProbability() 
  {
    return m_followLinkProbability;
  }
}
