/*------------------------------------------------------------------------
 * MAGNET client support components
 * $Id: BidExaminer.java,v 1.1.2.1 2002/05/05 19:41:55 ashutosh Exp $
 * Copyright (c) 2000 Regents of the University of Minnesota.
 * All Rights Reserved by the Regents of the University of Minnesota.
 * This file contains Original Code and/or Modifications of Original
 * Code as defined in the MAGNET Public Source License, which is
 * available at http://www.cs.umn.edu/magnet/source_license_v1.txt.
 * You may not use this file except in compliance with the License.
 *------------------------------------------------------------------------
 * Revision History:
 * 
 * Thu Jan  4 14:43:01 2001  John Collins
 *   Initial Entry
 */

package edu.umn.magnet.protocol.cust;

import edu.umn.magnet.protocol.InvalidDataException;

import java.util.Set;
import java.util.Date;

/**
 * This interface supports "what if" manipulations on bids.  You can
 * take a bid, set an externally-imposed early start time on some
 * tasks, and see how that affects early-finish times on other tasks
 * that may be related through opaque relationships in the bid.
 * <p>
 * It is expected that this interface will be implemented by some
 * hidden class inside Bid, which will give it access to non-public
 * information in the Bid. </p>
 * <p>
 * This version only supports setting and propagation of early start
 * times, which means your bid analysis must propagate forward through
 * the task graph.  It would be a straightforward extension to support
 * late finish times and backward propagation.
 * @author John Collins
 */
public interface BidExaminer
{
  /**
   * Returns a reference to the Bid upon which this thing is built.
   */
  public Bid getBid ();

  /**
   * Returns the early start date for the bid as a whole
   */
  public Date getEarlyStart ();

  /**
   * Returns the early start date for a particular task.  This may be
   * different from the early start date specified by the bid in case
   * this or some other task has had its early start date changed.
   */
  public Date getEarlyStart (String taskId);

  /**
   * Sets the early start date for some task.  This may affect dates
   * for other tasks in the bid.  
   * @exception InvalidDataException is thrown if
   * this change would cause an internal infeasibility in the bid.
   * This could happen, for example, if there were a subsequent task
   * (as determined by the TaskPlan) with a late start time that
   * couldn't be met.
   */
  public void setEarlyStart (String taskId, Date newValue)
    throws InvalidDataException;

  /**
   * Returns the early finish date for the bid as a whole.
   */
  public Date getEarlyFinish ();

  /**
   * Returns the early finish date for a particular task, possibly
   * modified by setting the early start for some task.
   */
  public Date getEarlyFinish (String taskId);
}
