/*-----------------------------------------------------------------------
 * Magnet Protocol Elements
 * $Id: SubtaskRequest.java,v 1.8.2.1 2002/05/05 19:41:56 ashutosh Exp $
 * Copyright (c) 1998 Regents of the University of Minnesota.
 * All Rights Reserved by the Regents of the University of Minnesota.
 * This file contains Original Code and/or Modifications of Original
 * Code as defined in the MAGNET Public Source License, which is
 * available at http://www.cs.umn.edu/magnet/source_license_v1.txt.
 * You may not use this file except in compliance with the License.
 *-----------------------------------------------------------------------
 * Revision History:
 * 
 * Tue May 22 21:55:20 2001  John Collins  #0000123
 *   Add updated serialVersionUID.  From now on, if you add or remove
 *   or change the types of non-transient member fields, you risk
 *   lossage when it comes to loading old plan files.
 *
 * 23Apr2001 DFK - Bug #0000029: constructor(SubtaskRequest) now
 *                  also copies taskName,comments.
 *
 * Sun Apr 22 11:31:03 2001  John E. Collins
 *   #0000026 Provide access to the name of the JobInfo
 *
 * Thu Feb 22 17:21:23 2001  John Collins
 *   Move to customer protocol component
 *
 * 11Mar2001 DFK - Added get/set for TaskName and Comments strings.
 *
 * Sat Dec 30 09:07:32 2000  John Collins
 *   Remove decommitment penalty stuff
 *
 * Mon Sep 25 11:43:24 2000  John Collins
 *   - Update names (this used to be called SubtaskRqClass)
 *   - Process visitor requests
 *
 * Wed Nov  3 14:28:53 1999  John Collins
 *   Replace JobInfo with Properties
 *
 * Wed Jan  6 12:07:07 1999  John E. Collins
 *   Get rid of the deprecated "name" attribute
 *
 * Thu Oct 29 17:04:23 1998  John E. Collins
 *   Add serialization signature.
 *
 * Wed Sep 16 11:43:41 1998  John E. Collins
 *   Deprecate the name attribute. It doesn't seem to be used
 *   anywhere.  The code for it here is very strange...
 */

package edu.umn.magnet.protocol.cust;

import edu.umn.magnet.protocol.JobInfo;
import edu.umn.magnet.protocol.AbstractJobInfoFactory;
import edu.umn.magnet.protocol.InvalidDataException;

import java.util.Date;

public class SubtaskRequest
  implements SubtaskRqSchedule, java.io.Serializable
{
  // Computed for version 1.7
  static final long serialVersionUID = 7986273244610037190L;

  /** A String denoting this task's ID; unique within the system. */
  protected String m_id = "";
  /** The earliest start time for this task. */
  protected Date m_earlyStart = null;
  /** The latest possible finish time for this task. */
  protected Date m_deadline = null;
  /** General info about task type, market statistics, etc for this
   *  task. */
  protected transient JobInfo m_jobinfo = null;
  protected String m_jobinfoName = null;
  /** A user-provided String name for this task. */
  private String _taskName = new String("");
  /** User-provided freeform comments about this task. */
  private String _taskComments = new String("");

  /**
   * Default constructor
   */
  public SubtaskRequest () {
    super();
  }

  public SubtaskRequest (SubtaskRequest task)
  {
    m_id=task.getId();
    m_earlyStart=task.getEarliestTimeToStart();
    m_deadline=task.getDeadline();
    m_jobinfo=task.getJobInfo();
    _taskName=task.getTaskName();
    _taskComments=task.getTaskComments();
  }

  /**
   * Construct a new task with the given ID string.
   */
  public SubtaskRequest (String id) {
    super();
    m_id = id;
  }

  /**
   * Returns the ID of the subtask.
   */
  public String getId() {
    return m_id;
  }

  /**
   * Sets that task ID. This is only valid if the ID has not yet been
   * set.
   * @exception InvalidDataException thrown if the ID is already set
   *     to a nonempty string, or if the new ID is blank or null.
   */
  public void setId(String newName) throws InvalidDataException {
    if (!m_id.equals(""))
      throw new InvalidDataException("Task ID already set");
    if (newName == null	|| newName.equals(""))
      throw new InvalidDataException("New ID is blank or null");
    m_id = newName;
  }

  /**
   * Returns the completion deadline for this subtask
   */
  public Date getDeadline () {
    return m_deadline;
  }

  /**
   * Sets the completion deadline for this subtask
   * @param deadline the date by which the work must be completed
   * @exception InvalidDataException thrown if the given deadline is earlier
   *     than the earliest start time.
   */
  public void setDeadline (Date deadline)
    throws InvalidDataException
    {
      if ((deadline == null)
	  || (m_earlyStart == null)
	  || (deadline.after(m_earlyStart))) {
	m_deadline = deadline;
      }
      else {
	throw new InvalidDataException("Deadline " + deadline
				       + " not later than early start "
				       + m_earlyStart);
      }
    }

  /**
   * Returns the earliest start time for this subtask.
   */
  public Date getEarliestTimeToStart () {
    return m_earlyStart;
  }

  /**
   * Sets the time before which this subtask cannot start
   * @param earlyStart the Date before which this subtask cannot start
   * @exception InvalidDataException thrown if the early start is not earlier
   *     than the late finish.
   */
  public void setEarliestTimeToStart (Date earlyStart)
    throws InvalidDataException
    {
      if ((earlyStart == null)
	  ||(m_deadline == null)
	  || (earlyStart.before(m_deadline))) {
	m_earlyStart = earlyStart;
      }
      else {
	throw new InvalidDataException("Early_start_not");
      }
    }

  /**
   * Returns the name of the job info for this subtask.
   */
  public String getJobInfoName ()
  {
    return m_jobinfoName;
  }

  /**
   * Returns the job information for this subtask.
   */
  public JobInfo getJobInfo ()
  {
    if (m_jobinfo == null && m_jobinfoName != null) {
      m_jobinfo =
	  AbstractJobInfoFactory.getInstance().getJobInfo(m_jobinfoName);
    }
    return m_jobinfo;
  }

  /**
   * Sets the job information for this subtask
   */
  public void setJobInfo (JobInfo ji)
  {
    m_jobinfo = ji;
    m_jobinfoName = ji.getTaskType();
  }

  /**
   * Sets the name of the job info for this task.
   */
  public void setJobInfoName (String newJI)
  {
    m_jobinfoName = newJI;
    m_jobinfo = null;
  }

  /**
   * Handle visitor requests
   */
  public void acceptVisitor (TaskVisitor tv)
  {
    tv.processSubtaskRequest(this);
  }

  /**
   * Get the user-defined task Name.
   * If not set, returns an empty String.
   */
  public String getTaskName() {
    return _taskName;
  }

  /** Set the task name.
   *  This is the nice user-chosen/readable Title.
   *  If arg is <tt>null</tt>, name is set to an empty String.
   */
  public void setTaskName(String aName) {
    if (null==aName) {
      _taskName = new String("");
    } else {
      _taskName = new String(aName);
    }
  }
  /**
   * Get the user-defined task comments.
   * If not set, returns an empty String.
   */
  public String getTaskComments() {
    return _taskComments;
  }

  /** Set the task comments.
   *  This is the user-entered freeform text.
   *  If arg is <tt>null</tt>, name is set to an empty String.
   */
  public void setTaskComments(String theComments) {
    if (null==theComments) {
      _taskComments = new String("");
    } else {
      _taskComments = new String(theComments);
    }
  }
}
