/*----------------------------------------------------------------------
 * Magnet Protocol elements
 * $Id: TaskPlan.java,v 1.11.2.1 2002/05/05 19:41:57 ashutosh Exp $
 * Copyright (c) 1998 Regents of the University of Minnesota.
 * All Rights Reserved by the Regents of the University of Minnesota.
 * This file contains Original Code and/or Modifications of Original
 * Code as defined in the MAGNET Public Source License, which is
 * available at http://www.cs.umn.edu/magnet/source_license_v1.txt.
 * You may not use this file except in compliance with the License.
 *----------------------------------------------------------------------
 * Revision History:
 * 
 * Tue May 22 21:55:20 2001  John Collins  #0000123
 *   Add updated serialVersionUID.  From now on, if you add or remove
 *   or change the types of non-transient member fields, you risk
 *   lossage when it comes to loading old plan files.
 *
 * 15May2001 DFK - Bug #0000077: Added detail strings to thrown exceptions.
 * 02May2001 DFK - Specify TaskID in TPEvent obj when firing planChanged.
 * 30Apr2001 DFK - Bug #0000032: added get/set for Plan Start Date and Plan Deadline.
 *
 * 11Mar2001 2pm DFK
 *   Added removeSuccessor() methods.
 *
 * 11Mar2001 11am DFK
 *   Added support for PlanListeners.  Added get/set for planName string.
 *
 * Thu Jan 11 14:19:13 2001  John Collins
 *   Add logging of plan statistics
 *
 * Sat Dec 30 17:08:20 2000  John Collins
 *   Get rid of Enumerations finally
 *
 * Thu Sep 28 16:19:17 2000  John Collins
 *   Add the getTasks method
 *   Add constructor TaskPlan(oldPlan, tasks)
 *
 * Sun Sep 24 21:30:15 2000  John Collins
 *   Change references SubtaskRqClass -> SubtaskRequest
 *
 * Thu Oct 29 17:06:42 1998  John E. Collins
 *   Add serialization signature.
 *
 * Wed Sep 16 14:21:59 1998  John E. Collins
 *   Throw InvalidTaskException on attempt to create successor link
 *   with a task that is not a member of the plan.
 *
 * Wed Sep 16 08:57:21 1998  John E. Collins
 *   Change name CircularDependencyEx -> CircularDependencyException,
 *   Change name InvalidTaskEx -> InvalidTaskException
 *   Change subtask accessor getID -> getId to match remainder of system
 *
 * Tue Sep 15 19:10:31 1998  John E. Collins
 *   Change references to SubtaskRequest to SubtaskRqClass to provide
 *   access to time data.
 *
 * Sat Jun 27 16:27:20 1998  John E. Collins
 *   Move to protocol package.
 */

package edu.umn.magnet.protocol.cust;

import org.apache.log4j.Category;

import java.io.Serializable;
import java.util.Date;
import java.util.HashMap;
import java.util.Vector;
import java.util.Stack;
import java.util.Iterator;
import java.util.NoSuchElementException;
import java.util.Collection;
import java.util.ArrayList;
import java.util.LinkedList;

import javax.swing.event.EventListenerList;

import edu.umn.magnet.protocol.InvalidDataException;  // for setPlan{StartDate,Deadline}().

/**
 * A TaskPlan object is a set of <tt>SubtaskRequest</tt> objects
 * embedded in a precedence network. Subtasks and precedence relations
 * can be manipulated with add/remove accessors, and can be queried
 * with get accessors. Of particular interest are the methods
 * <tt>isInSuccessors</tt> and <tt>getAllSuccessors</tt>. The first
 * tests whether a subtask is in the transitive closure of the successors
 * of another subtask, and the second returns the transitive closure of
 * successors of a subtask, sorted by a preorder depth-first walk of the
 * successor graph, with successive duplicates removed.
 * @version $Id: TaskPlan.java,v 1.11.2.1 2002/05/05 19:41:57 ashutosh Exp $
 */
public class TaskPlan implements java.io.Serializable
{
  static Category log = Category.getInstance(TaskPlan.class.getName());

  // computed for version 1.9.  If you have to change this, existing
  // plan files will be orphaned.
  static final long serialVersionUID = -8763256086910684392L;

  /** Holds a list of all tasks in this TaskPlan, mapping TaskIDs
   *  to {@link TaskWithPrecedence} objs.
   */
  private HashMap m_taskHash = new HashMap();
  /** List of listeners */
  protected transient EventListenerList _listenerList = null;
  /** User-assigned name of this plan. */
  private String _planName = new String("");

  /**
   * Default constructor
   */
  public TaskPlan() {
    super();
    _listenerList = new EventListenerList();
  }

  /**
   * Create a new TaskPlan from an old one, using only the specified
   * set of tasks.  Any precedence relations in the old one between
   * tasks in the new one will be preserved.  Any new tasks introduced
   * will be unconnected.
   */
  public TaskPlan(TaskPlan oldPlan, Collection tasks)
  {
    super();
    _listenerList = new EventListenerList();

    // First, add all the tasks.
    for (Iterator taski = tasks.iterator();
	 taski.hasNext(); ) {
      try {
	this.addTask((SubtaskRequest)taski.next());
      }
      catch (InvalidTaskException ite) {
	log.error("Error adding task: Unexpected Invalid Task Exception");
      }
    }

    // Then, add the relevant precedence relations.  We use identity
    // of the underlying SubtaskRequest instances to match the
    // contents of the old plan with the specified task list.
    HashMap oldContents = oldPlan.getPlanContents();
    for (Iterator twpi = oldContents.values().iterator();
	 twpi.hasNext(); ) {
      // If this one is in the new list, then try to set its
      // successors
      TaskWithPrecedence thisTWP = (TaskWithPrecedence)twpi.next();
      if (tasks.contains(thisTWP.getTaskData())) {
	for (Iterator succ = oldPlan.getSuccessors(thisTWP);
	     succ.hasNext(); ) {
	  SubtaskRequest successor = (SubtaskRequest)succ.next();
	  if (tasks.contains(successor)) {
	    try {
	      this.addSuccessor(thisTWP.getTaskData(), successor);
	    }
	    catch (InvalidTaskException ite) {
	      log.error("adding successor: Unexpected Invalid Task Exception");
	    }
	    catch (CircularDependencyException cde) {
	      log.error("adding successor: Unexpected Circular Dependency Exception");
	    }
	  }
	}
      }
    }
  }

  /**
   * Create a new TaskPlan from an old one.  This probably doesn't
   * really do what you want...
   * @deprecated Not well thought-out.
   */
  public TaskPlan(TaskPlan plan) {
    super();
    _listenerList = new EventListenerList();
    HashMap otherPlan = plan.getPlanContents();
    for (Iterator tsks = otherPlan.keySet().iterator();
	 tsks.hasNext(); ) {
      String key = (String)tsks.next();
      m_taskHash.put(key, new TaskWithPrecedence
		     ((TaskWithPrecedence)otherPlan.get(key)));
    }
  }

  /** Initializes the listner list - this is needed due to the serializability
   *  of this object.
   */
  public void initListenerList () {
    _listenerList = new EventListenerList();
  }  // end of initListenerList

  //---------------- Add and Remove Tasks -----------------------------------
  /**
   * Adds a task to the plan. It is assumed that the taskID is unique within
   * the plan; an error is thrown if this is not true. Successors may be
   * added later with addSuccessor. This method silently does nothing if
   * <tt>newTask</tt> is already in the plan's task set.
   * @param newTask a <tt>SubtaskRequest</tt> object to be added to the
   *     task set of this plan.
   * @exception InvalidTaskException if the taskID of <tt>newTask</tt>
   *     already belongs to a different task in the plan.
   */
  public void addTask (SubtaskRequest newTask) throws InvalidTaskException {
    TaskWithPrecedence twp =
      (TaskWithPrecedence)m_taskHash.get(newTask.getId());
    // silently do nothing if newTask is already in the set
    if (twp != null) {
      // complain if there's another task by the same name in the set
      if (!twp.getTaskData().equals(newTask)) {
	throw new InvalidTaskException("TaskID "+newTask.getId()+" is not unique");
      }
    }
    else {
      m_taskHash.put(newTask.getId(), new TaskWithPrecedence(newTask));
      firePlanDataChanged("added task (ID="+newTask.getId()+") to plan");
    }
  }

  /**
   * Removes a task from the plan, and cleans up any precedence relationships
   * associated with the task. Silently does nothing if the task is not in
   * the plan's task set. All precedence relationships involving the task are
   * removed from the plan.
   * @param oldTask a <tt>SubtaskRequest</tt> instance that is to be
   *     removed from the plan's task set.
   * @exception InvalidTaskException if a different task by the same
   *     name is in the plan's task set.
   */
  public void removeTask (SubtaskRequest oldTask) throws InvalidTaskException {
    TaskWithPrecedence twp =
      (TaskWithPrecedence)m_taskHash.get(oldTask.getId());
    // silently do nothing if oldTask isn't in the set
    if (twp != null) {
      // if there's a different task by the same name, we complain
      if (!twp.getTaskData().equals(oldTask)) {
	throw new InvalidTaskException("error: different task by the same name");
      }
      else {
	for (Iterator e = twp.getPredecessors().iterator();
	     e.hasNext(); ) {
	  ((TaskWithPrecedence)e.next()).removeSuccessor(twp);
	}
	for (Iterator e = twp.getSuccessors().iterator();
	     e.hasNext(); ) {
	  ((TaskWithPrecedence)e.next()).removePredecessor(twp);
	}
	m_taskHash.remove(oldTask.getId());
        firePlanDataChanged("removed task");
      }
    }
  }

  // ------------------- Operations on predecessors and successors ---------
  /**
   * Makes task2 be a successor of task1 in this plan.
   * @param task1 the <tt>SubtaskRequest</tt> object that is to be the
   *     predecessor.
   * @param task2 the <tt>SubtaskRequest</tt> object that is to be
   *     the successor.
   * @exception CircularDependencyException if task1 is already in the
   *     successor set of task2.
   * @exception InvalidTaskException if either task1 or task2 is not
   *     in the task set of plan.
   */
  public void addSuccessor (SubtaskRequest task1, SubtaskRequest task2)
    throws CircularDependencyException, InvalidTaskException
    {
      addSuccessor(getTWP(task1), getTWP(task2));
    }

  /**
   * Makes task2 be a successor of task1 in this plan.
   * @param taskID1 the name of the <tt>SubtaskRequest</tt> object that is
   *     to be the predecessor.
   * @param taskID2 the name of the <tt>SubtaskRequest</tt> object that is
   *     to be the successor.
   * @exception CircularDependencyException if the task named by
   *     <tt>taskID1</tt> is already in the successor set of the task
   *     named by <tt>taskID2</tt>.
   * @exception NullPointerException if either of the tasks named by
   *     <tt>task1ID</tt> or <tt>task2ID</tt> is not in the task set of plan.
   *     -- should this be <tt>InvalidTaskException</tt>? --
   */
  public void addSuccessor (String task1ID, String task2ID)
    throws CircularDependencyException, InvalidTaskException
    {
      addSuccessor(getTWP(task1ID), getTWP(task2ID));
    }

  public void addSuccessor (TaskWithPrecedence twp1, TaskWithPrecedence twp2)
    throws CircularDependencyException, InvalidTaskException
    {
      // crash if either twp1 or twp2 is null
      if (twp1 == null || twp2 == null)
	throw new InvalidTaskException("Null task");
      // crash if twp1 is already a successor of twp2
      if (twp2.isInSuccessors(twp1))
	throw new CircularDependencyException(twp1.getTaskId()+" is already a successor of "+twp2.getTaskId());
      twp1.addSuccessor(twp2);
      twp2.addPredecessor(twp1);
      /** @todo Should TaskPlan be a listener to Tasks that change (like this)? */
      firePlanDataChanged("added successor");
    }

  /** Remove successor relationship of one task to another.
   *  If twp2 is not a successor to twp1, it fails silently.
   *  @param twp1 The task that precedes twp2
   *  @param twp2 The successor to twp1
   *  @exception InvalidTaskException If either task is null.
   */
  public void removeSuccessor(TaskWithPrecedence twp1, TaskWithPrecedence twp2)
    throws InvalidTaskException
  {
    // crash if either twp1 or twp2 is null
    if (twp1 == null || twp2 == null)
      throw new InvalidTaskException("Null task");
    twp1.removeSuccessor(twp2);
    twp2.removePredecessor(twp1);
    /** @todo Should TaskPlan be a listener to Tasks that change (like this)? */
    firePlanDataChanged("removed successor");
  }

  /** Remove successor relationship of one task to another.
   *  If task2 is not a successor to task1, it fails silently.
   *  @param task1ID The taskID of the task that precedes task2
   *  @param task2ID The taskID of the task that succeeds task1
   *  @exception InvalidTaskException If either task is null.
   */
  public void removeSuccessor(String task1ID, String task2ID)
    throws InvalidTaskException
  {
    removeSuccessor(getTWP(task1ID), getTWP(task2ID));
  }

  /** Remove successor relationship of one task to another.
   *  If task2 is not a successor to task1, it fails silently.
   *  @param task1 The task that precedes task2
   *  @param task2 The successor to task1
   *  @exception InvalidTaskException If either task is null.
   */
  public void removeSuccessor(SubtaskRequest task1, SubtaskRequest task2)
    throws InvalidTaskException
  {
    removeSuccessor(getTWP(task1), getTWP(task2));
  }


  /**
   * Returns an Iterator of the immediate successors of task.
   */
  public Iterator getSuccessors (SubtaskRequest task) {
    return getSuccessors(getTWP(task));
  }

  /**
   * Returns an Iterator of the immediate successors of named task.
   * Values are SubtaskRequest instances.
   */
  public Iterator getSuccessors (String taskID) {
    return getSuccessors(getTWP(taskID));
  }

  private Iterator getSuccessors (final TaskWithPrecedence task) {
    return new Iterator() {
	Iterator m_successors = task.getSuccessors().iterator();
	public boolean hasNext () {
	  return m_successors.hasNext();
	}
	public Object next () {
	  return ((TaskWithPrecedence)m_successors.next()).getTaskData();
	}
	public void remove () {
	  throw new UnsupportedOperationException();
	}
    };
  }

  /**
   * Returns true if task2 is among the tasks in the transitive closure of
   * the successors of task1.
   */
  public boolean isInSuccessors (SubtaskRequest task1,
                                 SubtaskRequest task2) {
    TaskWithPrecedence twp1 = getTWP(task1);
    TaskWithPrecedence twp2 = getTWP(task2);

    return ((twp1 != null) &&
	    (twp2 != null) &&
	    twp1.isInSuccessors(twp2));
  }

  /**
   * Returns true if the task named by <tt>taskID2</tt> is among the tasks
   * in the transitive closure of the successors of the task named by
   * <tt>taskID1</tt>.
   */
  public boolean isInSuccessors (String taskID1, String taskID2) {
    TaskWithPrecedence twp1 = getTWP(taskID1);
    TaskWithPrecedence twp2 = getTWP(taskID2);

    return ((twp1 != null) &&
	    (twp2 != null) &&
	    twp1.isInSuccessors(twp2));
  }

  /**
   * Returns an Iterator of the transitive closure of the successors of
   * task. The elements of the Iterator are <tt>SubtaskRequest</tt>
   * instances.  Each item appears at most once in the Iterator.
   */
  public Iterator getAllSuccessors (SubtaskRequest task) {
    return getAllSuccessors (getTWP(task));
  }

  /**
   * Returns an Iterator of the transitive closure of the successors of
   * the task named by <tt>taskID</tt>. The elements of the Iterator are
   * <tt>SubtaskRequest</tt> instances. Each item appears at most once in
   * the result.
   */
  public Iterator getAllSuccessors (String taskID) {
    return getAllSuccessors (getTWP(taskID));
  }

  private Iterator getAllSuccessors (TaskWithPrecedence twp) {
    // we do this with a preorder depth-first walk of the precedence tree
    // -- it could also be done breadth-first.
    // Either way, we have to maintain a stack of partially-processed
    // Iterators.
    Vector results;
    Object task;
    Iterator rawData = new SuccessorWalker (twp);

    // Before returning, we process the raw data to eliminate duplicates
    // and extract the SubtaskRequest
    results = new Vector(m_taskHash.size());
    while (rawData.hasNext()) {
      task = ((TaskWithPrecedence)rawData.next()).getTaskData();
      if (results.indexOf(task) == -1)
	results.addElement(task);
    }
    return results.iterator();
  }

  /**
   * Returns an Iterator of the roots of the Successor DAG. Those are
   * tasks that have no Predecessors.
   */
  public Iterator getSuccessorRoots () {
    ArrayList result = new ArrayList();
    TaskWithPrecedence task;

    for (Iterator e = m_taskHash.values().iterator(); e.hasNext(); ) {
      task = (TaskWithPrecedence)e.next();
      if (task.getPredecessors().size() == 0)
	result.add(task.getTaskData());
    }
    return result.iterator();
  }

  /**
   * Returns an Iterator of the immediate predecessors of task.
   */
  public Iterator getPredecessors (SubtaskRequest task) {
    return getPredecessors(getTWP(task));
  }

  /**
   * Returns an Iterator of the immediate predecessors of named task.
   */
  public Iterator getPredecessors (String taskID) {
    return getPredecessors(getTWP(taskID));
  }

  private Iterator getPredecessors (final TaskWithPrecedence task) {
    return new Iterator () {
	Iterator m_predecessors = task.getPredecessors().iterator();
	public boolean hasNext () {
	  return m_predecessors.hasNext();
	}
	public Object next () {
	  return
	      ((TaskWithPrecedence)m_predecessors.next()).getTaskData();
	}
	public void remove () {
	  throw new UnsupportedOperationException();
	}
    };
  }

  // ------------------- Access to individual tasks ------------------------
  /**
   * Returns the SubtaskRequest associated with a task ID.
   * @param taskID a String naming a SubtaskRequest item that is a member
   *     of the task set for this plan.
   * @return the SubtaskRequest object named by <tt>taskID</tt> or Nil if
   *     the named task is not in the task set of this plan.
   */
  public SubtaskRequest getTaskData (String taskID) {
    TaskWithPrecedence twp = (TaskWithPrecedence)m_taskHash.get(taskID);
    if (twp == null)
      return null;
    else
      return twp.getTaskData();
  }

  public TaskWithPrecedence getTWP (String taskID) {
    return (TaskWithPrecedence)m_taskHash.get(taskID);
  }

  public TaskWithPrecedence getTWP (SubtaskRequest task) {
    return (TaskWithPrecedence)m_taskHash.get(task.getId());
  }

  // ------------------- Access to task list -------------------------------
  /**
   * Returns the number of tasks in this plan
   */
  public int size ()
  {
    return m_taskHash.size();
  }

  /**
   * Returns an Iterator of all the tasks in this plan, in no particular
   * order.
   */
  public Iterator iterator ()
  {
    return new Iterator() {
	Iterator m_elements = m_taskHash.values().iterator();
	public boolean hasNext () {
	  return m_elements.hasNext();
	}
	public Object next () {
	  return ((TaskWithPrecedence)m_elements.next()).getTaskData();
	}
	public void remove () {
	  throw new UnsupportedOperationException();
	}
    };
  }

  /**
   * Returns a Collection of all the tasks in this plan, in the form
   * of an ArrayList of SubtaskRequest instances.
   */
  public Collection getTasks ()
  {
    ArrayList result = new ArrayList(this.size());
    for (Iterator tasks = this.iterator();
	 tasks.hasNext(); )
      result.add(tasks.next());
    return result;
  }

  public HashMap getPlanContents()
  {
    return m_taskHash;
  }

  /**
   * Generates plan statistics to a log file
   */
  public void logPlanStats (Category log)
  {
    // compute task count and average fan-in
    int taskCount = 0;
    int fanIn = 0;
    for (Iterator tasks = getTasks().iterator();
	 tasks.hasNext(); ) {
      SubtaskRequest task = (SubtaskRequest)tasks.next();
      taskCount++;
      for (Iterator preds = getPredecessors(task);
	   preds.hasNext(); ) {
	Object pred = preds.next();
	fanIn++;
      }
    }
    double avgFanIn = (double)fanIn / (double)taskCount;

    // compute max depth
    int maxDepth = 0;
    for (Iterator tsks = getSuccessorRoots();
	 tsks.hasNext(); ) {
      int depth = countDepth((SubtaskRequest)tsks.next(), 1);
      if (depth > maxDepth)
	maxDepth = depth;
    }
    log.info("TaskCount: " + taskCount
	     + ", FanIn: " + avgFanIn
	     + ", Depth: " + maxDepth);
  }

  public void logPlanStats ()
  {
    logPlanStats(log);
  }

  private int countDepth (SubtaskRequest root, int count)
  {
    int max = count;
    for (Iterator succs = getSuccessors(root);
	 succs.hasNext(); ) {
      int newCount = countDepth((SubtaskRequest)succs.next(),
				count + 1);
      if (newCount > max)
	max = newCount;
    }
    return max;
  }

  //--------------------------------------------------------------------
  /**
   * We define a local class that provides a tree-walking Enumerator for
   * walking the DAG of successor relationships. This does a preorder
   * depth-first walk of the DAG. Duplicates in the output are caused by
   * "joins" in the DAG.
   */
  private class SuccessorWalker implements Iterator
  {
    Stack stack = new Stack();

    protected SuccessorWalker (TaskWithPrecedence root) {
      super();
      stack.push(root.getSuccessors().iterator());
    }

    public boolean hasNext() {
      Object junk;
      while (!stack.empty()) {
	if (((Iterator)stack.peek()).hasNext())
	  return true;
	else
	  junk = stack.pop();
      }
      return false;
    }

    public Object next() {
      Object result;
      // run down the stack to find a non-empty Iterator
      while (!stack.empty() &&
	     !((Iterator)stack.peek()).hasNext()) {
	result = stack.pop();
      }
      // if we haven't hit bottom, use the next element and dive
      // down again.
      if (!stack.empty()) {
	result = ((Iterator)stack.peek()).next();
	stack.push(((TaskWithPrecedence)result).getSuccessors().iterator());
      }
      else {
	throw new NoSuchElementException();
      }
      return result;
    }

    public void remove ()
    {
      throw new UnsupportedOperationException();
    }
  }

  // DFK - Methods to support/feed PlanListener instances.
  /** Adds a listener to the list that's notified each time a change to the data model occurs. */
  public void addPlanListener(PlanListener l) {
    _listenerList.add(PlanListener.class, l);
  }

  /** Remove a listener from the notify-list. */
  public void removePlanListener(PlanListener l) {
    _listenerList.remove(PlanListener.class, l);
  }

  /** Indicate (in a vague sense) that the plan has somehow been changed. */
  public void firePlanDataChanged() {
    TPEvent myTPEvent = new TPEvent(this, "TaskPlan has been modified");  // generic for now
    firePlanDataChanged(myTPEvent);
  }

  /** Indicate that the plan has somehow been changed.
   *  @param aMessage Some arbitrary message (currently not used for anything useful)
   */
  public void firePlanDataChanged(String aMessage) {
    TPEvent myTPEvent = new TPEvent(this, aMessage);  // generic for now
    firePlanDataChanged(myTPEvent);
  }

  /** Indicate that the plan has somehow been changed.
   *  @param evt The {@link TPEvent} to broadcast to all listeners.
   */
  public void firePlanDataChanged(TPEvent evt) {
    // Guaranteed to return a non-null array
    Object[] listeners = _listenerList.getListenerList();
    // Process the listeners last to first, notifying
    // those that are interested in this event
    for (int i = listeners.length-2; i>=0; i-=2) {
//      System.err.println("dfk: firePlanDataChanged(): listeners["+i+"]="+listeners[i]);
//       if (listeners[i] instanceof edu.umn.magnet.userif.PlanListener) {
       if (listeners[i] == PlanListener.class) {
//           System.err.println("dfk: firePlanDataChanged(): listeners["+i+"] is a PlanListener");
           ((PlanListener)listeners[i+1]).planChanged(evt);
       }
    }
  }
  // -------------------------------

  /** Get the plan name.
   *  If unassigned, the name is an empty String.
   */
  public String getPlanName() {
    return _planName;
  }

  /** Set the plan name.
   *  This is the nice user-chosen/readable Title.
   *  If arg is <tt>null</tt>, name is set to an empty String.
   */
  public void setPlanName(String aName) {
    if (null==aName) {
      _planName = new String("");
    } else {
      _planName = new String(aName);
    }
    firePlanDataChanged("changed plan name");
  }

  /** Get the Plan Start Date.
   *  This is the earliest EarlyStartDate of all tasks in this TaskPlan.
   *  @return the earlyStartDate, or <tt>null</tt> if no tasks present.
   */
  public Date getPlanStartDate() {
    if (this.size() <= 0) {
      return null;
    } else {
      Date eStart=null;
      boolean firsttime = true;
      for (Iterator i = this.iterator(); i.hasNext(); ) {
        SubtaskRequest req = (SubtaskRequest)i.next();
        if (firsttime) {
          eStart = req.getEarliestTimeToStart();
          firsttime = false;
        } else {
          if (eStart.after(req.getEarliestTimeToStart())) {
            eStart = req.getEarliestTimeToStart();
          }
        }
      }//for
      return eStart;
    }
  }

  /** Set the Plan Start Date.
   *  This is the earliest EarlyStartDate of all tasks in this TaskPlan.
   *  All tasks' dates will be offset so that the earliest task's Start date is this date.
   *  @todo Maybe allow exceptions to be thrown upwds instead of silently failing??
   */
  public void setPlanStartDate(Date dEStart) {
    Date currentStart = this.getPlanStartDate();
    if (null==currentStart) {  return;  }
    /** @todo Is subtracting millisecond offsets the "right way" to do this? */
    long offset = dEStart.getTime() - currentStart.getTime();
    if (0==offset) { return; }  // do nothing
    for (Iterator i = this.iterator(); i.hasNext(); ) {
      SubtaskRequest req = (SubtaskRequest)i.next();
      Date tempEStart = req.getEarliestTimeToStart();
      Date tempDeadline = req.getDeadline();
//      try {
//        req.setEarliestTimeToStart(new Date(tempEStart.getTime() + offset));
//        req.setDeadline(new Date(tempDeadline.getTime() + offset));
        /**
         * If you change dates programatically via SubtaskRequest.set*(Date),
         *  you fall victim to the temporal police;  in the middle of shoving
         *  all dates forward/backward, you might temporarily be in violation
         *  of temporal predecences.   So instead, side-step the checking by
         *  directly manipulating each Task's date objects.
         *  Warning: no "checking" is done afterwards to verify you're still self-consistent..
         */
        req.m_earlyStart=(new Date(tempEStart.getTime() + offset));
        req.m_deadline=(new Date(tempDeadline.getTime() + offset));
//      } catch (InvalidDataException ide) {
//        return;
//      }
    }//for
    firePlanDataChanged("changed plan start date");
  }

  /** Get the Plan Deadline.
   *  This is the latest Deadline of all tasks in this TaskPlan.
   *  @return the deadline, or <tt>null</tt> if no tasks present.
   */
  public Date getPlanDeadline() {
    if (this.size() <= 0) {
      return null;
    } else {
      Date deadline=null;
      boolean firsttime = true;
      for (Iterator i = this.iterator(); i.hasNext(); ) {
        SubtaskRequest req = (SubtaskRequest)i.next();
        if (firsttime) {
          deadline = req.getDeadline();
          firsttime = false;
        } else {
          if (deadline.before(req.getDeadline())) {
            deadline = req.getDeadline();
          }
        }
      }//for
      return deadline;
    }
  }

  /** Set the Plan Deadline.
   *  This is the latest Deadline of all tasks in this TaskPlan.
   *  All tasks' dates will be offset so that the latest task's Deadline date is this date.
   *  @todo Maybe allow exceptions to be thrown upwds instead of silently failing??
   */
  public void setPlanDeadline(Date dDeadline) {
    Date currentDeadline = this.getPlanDeadline();
    if (null==currentDeadline) {  return;  }
    /** @todo Is subtracting millisecond offsets the "right way" to do this? */
    long offset = dDeadline.getTime() - currentDeadline.getTime();
    if (0==offset) { return; }  // do nothing
    for (Iterator i = this.iterator(); i.hasNext(); ) {
      SubtaskRequest req = (SubtaskRequest)i.next();
      Date tempEStart = req.getEarliestTimeToStart();
      Date tempDeadline = req.getDeadline();
//      try {
//        req.setEarliestTimeToStart(new Date(tempEStart.getTime() + offset));
//        req.setDeadline(new Date(tempDeadline.getTime() + offset));
        /**
         * If you change dates programatically via SubtaskRequest.set*(Date),
         *  you fall victim to the temporal police;  in the middle of shoving
         *  all dates forward/backward, you might temporarily be in violation
         *  of temporal predecences.   So instead, side-step the checking by
         *  directly manipulating each Task's date objects.
         *  Warning: no "checking" is done afterwards to verify you're still self-consistent..
         */
        req.m_earlyStart=(new Date(tempEStart.getTime() + offset));
        req.m_deadline=(new Date(tempDeadline.getTime() + offset));
//      } catch (InvalidDataException ide) {
//        return;
//      }
    }//for
    firePlanDataChanged("changed plan deadline");
  }
}
