/*------------------------------------------------------------------------
 * MAGNET protocol component
 * $Id: AbstractJobInfoFactory.java,v 1.2.2.2 2002/05/05 19:41:58 ashutosh Exp $
 * Copyright (c) 2001 Regents of the University of Minnesota.
 * All Rights Reserved by the Regents of the University of Minnesota.
 * This file contains Original Code and/or Modifications of Original
 * Code as defined in the MAGNET Public Source License, which is
 * available at http://www.cs.umn.edu/magnet/source_license_v1.txt.
 * You may not use this file except in compliance with the License.
 *------------------------------------------------------------------------
 * Revision History:
 * 
 * Sun Apr 22 11:57:13 2001  John E. Collins
 *   Initial Entry, part of bug ID 0000026
 */

package edu.umn.magnet.protocol;

import org.apache.log4j.Category;

import java.util.HashMap;

/**
 * This class implements the root of a JobInfoFactory singleton
 * hierarchy.  At startup, you must call the
 * static <tt>instantiate()</tt> method on a subclass that knows how to
 * serve up JobInfo instances in the current environment  The result
 * will be that the an instance of that subclass will become the
 * singleton.
 * @author John Collins
 * @version $Revision: 1.2.2.2 $
 */
public abstract class AbstractJobInfoFactory
{
  // This is the collection of task types
  private static HashMap jobTypes = new HashMap();

  // This is the actual factory
  private static AbstractJobInfoFactory instance = null;

  // --------- manage the singleton ------------
  /**
   * Sets the instance that will be returned as the singleton factory
   * object.  This should be done at startup.
   */
  public static void setInstance (AbstractJobInfoFactory factory)
  {
    instance = factory;
  }

  /**
   * Gets the current factory instance.
   */
  public static AbstractJobInfoFactory getInstance ()
  {
    return instance;
  }

  /**
   * Instantiates the singleton.  You need to call this on a subclass,
   * to avoid the error you'll get if you call it on this abstract
   * superclass.  Note that if you call this a second time, the
   * original singleton will get replaced with a new one.
   * <p>
   * This method must be implemented appropriately in each subclass.
   * Otherwise, the inherited version does nothing.
   * </p>
   */
  public static void instantiate ()
  {
  }

  // --------- manage the JobInfo name -> instance dictionary --------
  /**
   * Returns the JobInfo matching the given task type string, or null
   * if no such JobInfo can be found.
   * @deprecated  Use getJobInfo instead.
   */
  public JobInfo find (String taskType)
  {
    return (JobInfo)jobTypes.get(taskType);
  }

  /**
   * Returns the JobInfo matching the given task type string, or null
   * if no such JobInfo can be found.
   */
  public JobInfo getJobInfo (String taskType)
  {
    return (JobInfo)jobTypes.get(taskType);
  }

  /**
   * Adds a JobInfo instance to the local dictionary.
   */
  public void addJobInfo (JobInfo newJI)
  {
    jobTypes.put(newJI.getTaskType(), newJI);
  }

  /**
   * Clears out the JobInfo dictionary
   */
  protected void clear ()
  {
    jobTypes.clear();
  }

  /**
   * Returns the dictionary itself for specialized processing by
   * subclasses.
   */
  public HashMap list ()
  {
    return jobTypes;
  }
}
