/* ----------------------------------------------------------------------
 * Magnet Protocol Elements
 * $Id: ProtocolElement.java,v 1.2.2.2 2002/05/05 19:41:58 ashutosh Exp $
 * Copyright (c) 1998 Regents of the University of Minnesota.
 * All Rights Reserved by the Regents of the University of Minnesota.
 * This file contains Original Code and/or Modifications of Original
 * Code as defined in the MAGNET Public Source License, which is
 * available at http://www.cs.umn.edu/magnet/source_license_v1.txt.
 * You may not use this file except in compliance with the License.
 * ----------------------------------------------------------------------
 * Revision History:
 * 
 * Tue Mar  6 14:44:37 2001  John Collins
 *   Back to protocol class.  This really is common to all
 *   components.  Also, remove the Data thingie.
 * 
 * Thu Feb 22 17:20:34 2001  John Collins
 *   Move to customer protocol component
 * 
 * Thu Oct 29 16:48:14 1998  John E. Collins
 *   Add serialization signature.
 * 
 * Sat Jun 27 16:11:40 1998  John E. Collins
 *   Move to protocol package.
 * 
 * Wed Jun  3 09:28:16 1998  John E. Collins
 *   Add isVisible predicate to determine whether a particular
 *   protocol element is visible to a particular participant.
 */

package edu.umn.magnet.protocol;

/**
 * This is the root class of all the Magnet protocol elements. The only thing
 * they all have in common is an ID and the ID of the Session, so we put that
 * here.
 * @author John Collins
 */
public abstract class ProtocolElement implements java.io.Serializable
{
  /**
   * Every protocol element has an ID that is at least unique within
   * the set of elements of a particular type, in a particular
   * session, and related to a particular agent.
   */
  private String m_id = null;

  /**
   * This is the session ID of the session to which this message
   * was sent.
   */
  private String m_sessionId = null;

  /**
   * This is the ID of the agent that issued this message.
   */
  private String m_agentId = null;
  
  /**
   * Default constructor
   */
  public ProtocolElement ()
  {
    super();
  }

  /**
   * Constructs one with all data filled in.
   */
  public ProtocolElement (String id, String sessionId, String agentId)
  {
    super();
    m_id = id;
    m_sessionId = sessionId;
    m_agentId = agentId;
  }

  /**
   * Returns the ID of a protocol element
   */
  public String getId () {
    return m_id;
  }

  /**
   * Sets the ID of a protocol element. Note that you can only set this
   * once;  after that it complains. You are probably better off making
   * the constructor set the ID, or letting the server set it. You can
   * get a unique ID from the Exchange.
   * @exception InvalidDataException thrown if an attempt is made to change
   *     the ID of a protocol element.
   */
  public void setId (String newId) throws InvalidDataException {
    if (m_id == null && newId != null && newId.length() > 0)
      m_id = newId;
    else
      throw new InvalidDataException ("Attempt to change object ID from "
				      + m_id + " to " + newId);
  }
  
  /**
   * Returns the ID of the Session to which this piece of protocol belongs.
   */
  public String getSessionId () {
    return m_sessionId;
  }
  
  /**
   * Sets the Session ID for this element. This should probably only be set
   * by a constructor
   */
  public void setSessionId (String sessionId)
  {
    m_sessionId = sessionId;
  }

  /**
   * Returns the Agent ID for this item.
   */
  public String getAgentId ()
  {
    return m_agentId;
  }

  /**
   * Sets the Agent ID for this item. Note that it is not necessary
   * for a client to set this value; the server will do that.
   */
  public void setAgentId (String newId) {
    m_agentId = newId;
  }
}

