/*------------------------------------------------------------------------
 * MAGNET User Interface components
 * $Id: GenericGanttObject.java,v 1.4.2.1 2002/05/05 19:42:01 ashutosh Exp $
 * Copyright (c) 2000 Regents of the University of Minnesota.
 * All Rights Reserved by the Regents of the University of Minnesota.
 * This file contains Original Code and/or Modifications of Original
 * Code as defined in the MAGNET Public Source License, which is
 * available at http://www.cs.umn.edu/magnet/source_license_v1.txt.
 * You may not use this file except in compliance with the License.
 *------------------------------------------------------------------------
 * Revision History:
 *
 * 11Apr2001 DFK - Initial Entry
 * 07May2001 DFK - Bug #0000022: Fixed affine xform (positioning) errors;
 *                  Added support for startShape upPentagon.
 * 13May2001 DFK - Bug #0000053: Removed upPentagon default; added thinLine (for bids).
 * 21May2001 DFK - Bug #0000099: Added support for slack tails
 */

package edu.umn.magnet.client.userif;

import java.util.*;
import java.text.*;
import java.awt.*;
import javax.swing.*;
import java.awt.event.*;
import java.awt.geom.*;
import java.awt.font.*;


/** Inner class to represent the common stuff for Gantt objects.
 *  DAVE: I wonder if these feature members should be chart-wide,
 *        not per-obj-instance variables..
 */
public abstract class GenericGanttObject {
  /** Solid stroke style for start/end shapes. (Default) */
  public static final int TYPE_SOLID    = 0;
  /** Dashed stroke style for start/end shapes. */
  public static final int TYPE_DASHED   = 2;
  /** Framed stroke style for start/end shapes. */
  public static final int TYPE_FRAMED   = 1;

  /** An empty shape (eg. none) for start/end shapes. */
  public static final int SHAPE_NONE = 0;
  /** An upright pentagon shape for start/end shapes. */
  public static final int SHAPE_UP_PENTAGON = 1;
  /** A standard wide line shape for middle shapes. */
  public static final int SHAPE_LINE_WIDE = 2;
  /** A standard thin line shape for middle shapes (suitable for overlaid bids). */
  public static final int SHAPE_LINE_THIN = 3;
  /** The tail associated w/ SHAPE_LINE_WIDE. */
  public static final int SHAPE_TAIL_WIDE = 4;
  /** The tail associated w/ SHAPE_LINE_THIN. */
  public static final int SHAPE_TAIL_THIN = 5;
  /** The small vertical endtail associated w/ SHAPE_LINE_THIN. */
  public static final int SHAPE_ENDTAIL_WIDE = 6;
  /** The small vertical endtail associated w/ SHAPE_LINE_THIN. */
  public static final int SHAPE_ENDTAIL_THIN = 7;
  /** Always update this one to keep an accurate count.. */
  protected static final int SHAPE_MAX = 8;
  /** @todo Add more shapes! */

  /** An array to hold Gantt shapes. */
  protected static Shape[] Shapes;

  /** The bounding box in which the Gantt object's middle can be drawn.
   *  This is after any RowHeader/IntercellSpacing is subtracted off.
   *  The Start and End shapes are assumed to be centered on the spots that're
   *  middle-height of the left and right ends, respectively.
   */
  Rectangle boundingBox = null;
  /** The shape of the start figure.
   *  See the SHAPE_* enums..
   */
  int startShape = SHAPE_NONE;
//  int startShape = SHAPE_UP_PENTAGON;
  /** The type of the start figure.
   *  Examples are {@link #TYPE_SOLID} (default), {@link #TYPE_DASHED},
   *  {@link TYPE_FRAMED}.
   */
  int startType = TYPE_SOLID;
  /** The color of the start figure.
   *  (Default is {@link Color#blue}.)
   */
  Color startColor = Color.blue;

  /** The shape of the middle bar.
   *  See the SHAPE_* enums..
   */
  int middleShape = SHAPE_LINE_WIDE;
  /** The pattern of the middle bar */
  int middlePattern = 0;
  /** The color of the middle bar
   *  (Default is {@link Color#blue}.)
   */
  Color middleColor = Color.blue;

  /** The shape of the end figure.
   *  See the SHAPE_* enums..
   */
  int endShape = SHAPE_NONE;
  /** The type of the end figure.
   *  Examples are {@link #TYPE_SOLID} (default), {@link #TYPE_DASHED},
   *  {@link TYPE_FRAMED}.
   */
  int endType = TYPE_SOLID;
  /** The color of the end figure.
   *  (Default is {@link Color#blue}.)
   */
  Color endColor = Color.blue;

  /** The default stroke */
  Stroke defaultStroke = new BasicStroke();
  /** @todo Add getters/setters for all those pesky class members.. */

  /** Startup processing: create the various shapes.
   *  Create them within a square (maybe w[0..16] x h[0..16] ?) range,
   *  Center them around (0,0),    and let the renderer's code
   *  AffineTransform-scale them to the desired size/shape/width/etc.
   */
  static {
    // Size the Shapes array.
    Shapes = new Shape[SHAPE_MAX];
    /**
     *  Create the up-pentagon shape.
     */
    Shapes[SHAPE_UP_PENTAGON] = new Polygon();
    // I know there's some clever mathy way to calculate the height of a pentagon,
    //   but I'm feeling too math-impaired to figure it out.. :)
    Point2D pentPoint = new Point2D.Double(0d, -3.75d);    // (Approximately.. :)  )
    AffineTransform af = AffineTransform.getRotateInstance(
                                  360d / 5d * Math.PI / 180d);
    for (int i=0; i < 5; i++) {
      ((Polygon)Shapes[SHAPE_UP_PENTAGON]).addPoint(
                            (int)pentPoint.getX(), (int)pentPoint.getY());
      pentPoint = af.deltaTransform(pentPoint, null);
    }

    /**
     * Create the lines.
     */
    //dfk05May2001; affine errors??
    Shapes[SHAPE_LINE_WIDE] = new Rectangle(-5, -3, 10, 6);
//    Shapes[SHAPE_LINE_WIDE] = new Rectangle(0, 0, 10, 6);
    Shapes[SHAPE_LINE_THIN] = new Rectangle(-5, -2, 10, 4);

    /**
     * Create the tails.
     */
    Shapes[SHAPE_TAIL_WIDE] = new GeneralPath();  //Polygon();
    Shapes[SHAPE_TAIL_THIN] = new GeneralPath();
    // For each shape, make a sideways "T", like this:  ---|
//    ((Polygon)Shapes[SHAPE_TAIL_WIDE]).addPoint(-5, -1);
//    ((Polygon)Shapes[SHAPE_TAIL_WIDE]).addPoint(4, -1);
//    ((Polygon)Shapes[SHAPE_TAIL_WIDE]).addPoint(4, -3);
//    ((Polygon)Shapes[SHAPE_TAIL_WIDE]).addPoint(5, -3);
//    ((Polygon)Shapes[SHAPE_TAIL_WIDE]).addPoint(5, 3);
//    ((Polygon)Shapes[SHAPE_TAIL_WIDE]).addPoint(5, 1);
//    ((Polygon)Shapes[SHAPE_TAIL_WIDE]).addPoint(-5, 1);
//    ((Polygon)Shapes[SHAPE_TAIL_WIDE]).addPoint(-5, -1);
//    // dfk: triangle, for debugging
//    ((Polygon)Shapes[SHAPE_TAIL_WIDE]).addPoint(-5, 0);
//    ((Polygon)Shapes[SHAPE_TAIL_WIDE]).addPoint(5, -3);
//    ((Polygon)Shapes[SHAPE_TAIL_WIDE]).addPoint(5, 3);
//    ((Polygon)Shapes[SHAPE_TAIL_WIDE]).addPoint(-5, 0);
    ((GeneralPath)Shapes[SHAPE_TAIL_WIDE]).append(
          new Rectangle2D.Float(-5f, -0.5f, 10f, 1f), true);
//    ((GeneralPath)Shapes[SHAPE_TAIL_WIDE]).moveTo(-5f, 0f);
//    ((GeneralPath)Shapes[SHAPE_TAIL_WIDE]).lineTo(5f, 0f);
//    ((GeneralPath)Shapes[SHAPE_TAIL_WIDE]).moveTo(5f, 3f);
//    ((GeneralPath)Shapes[SHAPE_TAIL_WIDE]).lineTo(5f, -3f);

//    ((Polygon)Shapes[SHAPE_TAIL_THIN]).addPoint(-5, 0);
//    ((Polygon)Shapes[SHAPE_TAIL_THIN]).addPoint(5, 0);
//    ((Polygon)Shapes[SHAPE_TAIL_THIN]).addPoint(5, -2);
//    ((Polygon)Shapes[SHAPE_TAIL_THIN]).addPoint(5, 2);
//    ((Polygon)Shapes[SHAPE_TAIL_THIN]).addPoint(5, 0);
//    ((Polygon)Shapes[SHAPE_TAIL_THIN]).addPoint(-5, 0);
    ((GeneralPath)Shapes[SHAPE_TAIL_THIN]).append(
          new Rectangle2D.Float(-5f, -0.5f, 10f, 1f), true);

    // Endtails
    Shapes[SHAPE_ENDTAIL_WIDE] = new GeneralPath();  //Polygon();
    Shapes[SHAPE_ENDTAIL_THIN] = new GeneralPath();
    ((GeneralPath)Shapes[SHAPE_ENDTAIL_WIDE]).append(
          new Rectangle2D.Float(-1f, -3f, 2f, 6f), true);
    ((GeneralPath)Shapes[SHAPE_ENDTAIL_THIN]).append(
          new Rectangle2D.Float(-1f, -1.0f, 2f, 2f), true);
  }

  /** "paint" the object into the provided graphics context. */
  public void paint(Graphics2D g2) {
    if (null==boundingBox  ||
          (boundingBox.getHeight()==0d  &&  boundingBox.getWidth()==0d) )
    {  return;  }  // nowhere to draw the shape
    Dimension destD = boundingBox.getSize();
    int destHeight = destD.height;

    // Draw middle region first.
    // Scale the object into the destination height size.
    Shape s=null;
    switch (middleShape) {
      case SHAPE_LINE_WIDE:
      case SHAPE_LINE_THIN:
      case SHAPE_TAIL_WIDE:
      case SHAPE_TAIL_THIN:
      default:
          {
//            g2.setStroke(new BasicStroke(5f));
//            s = Shapes[SHAPE_LINE_WIDE];
            s = Shapes[middleShape];
            // Position it to be centered @ the origin.
            AffineTransform at0 = AffineTransform.getTranslateInstance(
                                      s.getBounds().getWidth()/2.0d, boundingBox.getHeight()/2.0d);
            s = at0.createTransformedShape(s);
            // Scale it to the correct size
            AffineTransform at1 = AffineTransform.getScaleInstance(
                                destD.getWidth() / 10d ,
                                destD.getHeight() / 10d );  // assuming 10x10
            s = at1.createTransformedShape(s);
            // Position it
            AffineTransform at2 = AffineTransform.getTranslateInstance(
                                      boundingBox.getX(), boundingBox.getY());
            s = at2.createTransformedShape(s);
            g2.setColor(middleColor);
            g2.setStroke(defaultStroke);
            if (0==middlePattern) {
              g2.fill(s);
            } else {
              /** @todo call fillByImage(); need ImageIcons for fill patterns..... */
            }
          }
          break;
    }//switch middleShape

    /** @todo Draw the beginning shape */
    switch (startShape) {
      case SHAPE_UP_PENTAGON:
          {
            s = Shapes[SHAPE_UP_PENTAGON];
            double originalHalfWidth = s.getBounds().getWidth()/2.0d;
            // Position it to be centered @ the origin.
            AffineTransform at0 = AffineTransform.getTranslateInstance(
                                      s.getBounds().getWidth()/2.0d, boundingBox.getHeight()/2.0d);
            s = at0.createTransformedShape(s);
            // Scale it to the correct size
            AffineTransform at1 = AffineTransform.getScaleInstance(
                                destD.getHeight()/*Yes, height, not width!!*/ / 10d ,
                                destD.getHeight() / 10d );  // assuming 10x10
            s = at1.createTransformedShape(s);
            // Position it
            AffineTransform at2 = AffineTransform.getTranslateInstance(
                                      boundingBox.getX()-originalHalfWidth/*center it*/,
                                      boundingBox.getY());
            s = at2.createTransformedShape(s);
            g2.setColor(startColor);
            g2.setStroke(defaultStroke);
            /** @todo Make patterns for start/end too */
            if (0==middlePattern) {
              g2.fill(s);
            } else {
              /** @todo call fillByImage(); need ImageIcons for fill patterns..... */
            }
          }
          break;

      default:
         break;

    }//switch startShape

    /** @todo Draw the ending shape */
//    ((GeneralPath)Shapes[SHAPE_TAIL_WIDE]).append(
//          new Rectangle2D.Float(5f, -3f, 0.01f, 3f), true);
    switch (endShape) {
      case SHAPE_UP_PENTAGON:
          {
            s = Shapes[SHAPE_UP_PENTAGON];
            double originalHalfWidth = s.getBounds().getWidth()/2.0d;
            // Position it to be centered @ the origin.
            AffineTransform at0 = AffineTransform.getTranslateInstance(
                                      s.getBounds().getWidth()/2.0d, boundingBox.getHeight()/2.0d);
            s = at0.createTransformedShape(s);
            // Scale it to the correct size
            AffineTransform at1 = AffineTransform.getScaleInstance(
                                destD.getHeight()/*Yes, height, not width!!*/ / 10d ,
                                destD.getHeight() / 10d );  // assuming 10x10
            s = at1.createTransformedShape(s);
            // Position it
            AffineTransform at2 = AffineTransform.getTranslateInstance(
                                      boundingBox.getX()+boundingBox.getWidth()-originalHalfWidth/*center it*/,
                                      boundingBox.getY());
            s = at2.createTransformedShape(s);
            g2.setColor(endColor);
            g2.setStroke(defaultStroke);
            /** @todo Make patterns for start/end too */
//            if (0==middlePattern) {
              g2.fill(s);
//            } else {
//              /** @todo call fillByImage(); need ImageIcons for fill patterns..... */
//            }
          }//SHAPE_UP_PENTAGON
          break;

      case SHAPE_ENDTAIL_WIDE:
      case SHAPE_ENDTAIL_THIN:
          {
            s = Shapes[endShape];
            double originalHalfWidth = s.getBounds().getWidth()/2.0d;
            // Position it to be centered @ the origin.
            AffineTransform at0 = AffineTransform.getTranslateInstance(
                                      s.getBounds().getWidth()/2.0d, boundingBox.getHeight()/2.0d);
            s = at0.createTransformedShape(s);
            // Scale it to the correct size
            AffineTransform at1 = AffineTransform.getScaleInstance(
                                destD.getHeight()/*Yes, height, not width!!*/ / 10d ,
                                destD.getHeight() / 10d );  // assuming 10x10
            s = at1.createTransformedShape(s);
            // Position it
            AffineTransform at2 = AffineTransform.getTranslateInstance(
                                      boundingBox.getX()+boundingBox.getWidth()-2*originalHalfWidth,
                                      boundingBox.getY());
            s = at2.createTransformedShape(s);
            g2.setColor(endColor);
            g2.setStroke(defaultStroke);
//            if (0==endPattern) {
              g2.fill(s);
//            } else {
//              /** @todo call fillByImage(); need ImageIcons for fill patterns..... */
//            }
          }
          break;

      default:
          break;

    }//switch endShape

  }

//  protected void fillByImage(Graphics2D g2,
//   Shape shape, int xOffset, Image img) {
//    if (img == null)
//      return;
//    int wImg = img.getWidth(this);
//    int hImg = img.getHeight(this);
//    if (wImg <=0 || hImg <= 0)
//      return;
//    g2.setClip(shape);
//    Rectangle bounds = shape.getBounds();
//    for (int xx = bounds.x+xOffset;
//     xx < bounds.x+bounds.width; xx += wImg)
//      for (int yy = bounds.y; yy < bounds.y+bounds.height;
//       yy += hImg)
//        g2.drawImage(img, xx, yy, this);  //dfk; "this" needs to be an ImageObserver; any Component, incl Frame or Window..
//  }

}

