/*------------------------------------------------------------------------
 * MAGNET Manager User Interface component
 * $Id: GuiEvalUtil.java,v 1.7.2.1 2002/05/05 19:42:01 ashutosh Exp $
 * Copyright (c) 2001 Regents of the University of Minnesota.
 * All Rights Reserved by the Regents of the University of Minnesota.
 * This file contains Original Code and/or Modifications of Original
 * Code as defined in the MAGNET Public Source License, which is
 * available at http://www.cs.umn.edu/magnet/source_license_v1.txt.
 * You may not use this file except in compliance with the License.
 *------------------------------------------------------------------------
 * Revision History:
 *
 * 06May2001 PBG Initial Entry
 * 13May2001 PBG Bug #0000056, #0000063: Added the PrintBidDetails Stuff
 * 23May2001 PBG Bug #0000118: Bid IDs need to print with Bid Information
 */


package edu.umn.magnet.client.userif;
import edu.umn.magnet.util.MagnetProperties;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;
import java.util.*;                     // for iterator
import edu.umn.magnet.client.status.*;  //for status objects
import edu.umn.magnet.protocol.cust.*;  // for TaskPlan
import edu.umn.magnet.protocol.cust.Bid;     // for Bid data structure
import edu.umn.magnet.protocol.cust.SubtaskRequest; // for Subtask Structure
import edu.umn.magnet.client.eval.TaskBidPlan;  // for the TaskBidPlan
import edu.umn.magnet.client.eval.BidEvaluator; // for Bid Evaluation

/** This is utility class to do the GUI Bid Evaluation Activities
 *
 *  This class will not contain any states, rather it will be available to
 *  be called by the various GUI components in order to do some bid processing.
 *
 *  This is being created in order to keep the code more modular and for re-use.
 *
 */
public class GuiEvalUtil {

  /** Any constants */

  /** Methods that are all-purpose and should belong in a "utility" class */

  /** Create Bid Evaluator
   *  This methods creates a bid evaluator to be used for doing the evaluations
   *  within the User Interface.  In order to do this one must not only create
   *  the bid evaluator - but initialize it as well.  Since this is a set of
   *  fairly static tasks - it has been put inside a Utility method in order
   *  to keep the code centralized.
   *  The steps to create a bid evaluator:
   *  <ul>
   *  <li> Construct the BidEvaluator
   *  <li> Set the Properties
   *  <li> Set the Session ID
   *  <li> Set the Plan Status
   *  <li> Set the RFQ Status
   *  </ul>
   */

   public static BidEvaluator createBidEvaluator(PlanStatus ps, RFQStatus rs,
                                  TaskPlan tp) {

    String sessionId = null;
    SessionStatus ss = null;
    BidEvaluator be = null;
    Properties props = null;

    ss = GuiUtil.getSessionStatus(AgentStatus.getInstance(), tp);
    if ((ss == null) || (ps == null) || (rs == null)) return null;

    // Props are:  "Eval.UseCoverageConstraints" defaults to true - set to "false"
    // "Eval.UseFeasibilityConstraints" defaults to true - set to "false"
    // "Eval.UseOptimizeCompletion" defaults to false - set to "true"
    // "Search" - this indicates a space seperated list of search engine types
    //            "IP", "SYSTEMATIC" - deprecated
    //            "IPCON" - IP with constraints
    //            "SA" - Simulated Annealing
    // current implementation right now goes in and hard codes the eval to use
    //  SA or IPCON

    // data is OK - create and init the BidEvaluator
    be = new BidEvaluator();
    be.setProperties(MagnetProperties.getProperties());
    be.setSessionId(ss.getSessionId());
    be.setPlanStatus(ps);
    be.setRFQStatus(rs);
    return be;
  } // end of createBidEvaluator

 /**
   * This method returns a formatted string laying out all of the contents
   * in the bid structure for a particular bid
   */
  public static String printBid(Bid _bid) {
     String ts = new String("");
     java.util.Set taskSet = null;

     if (_bid == null) {
        /** @todo: Need to dispaly an error here to the user*/
        return new String("Bid Passed in was null");
     }

     ts+="Bid ID:\t"+_bid.getId()+"\n";
     ts+="Bid is for RFQ:\t"+_bid.getRFQId()+"\n";
     ts+="Bid Cost:\t"+_bid.getCost()+"\n";
     /** @todo: Need to put Duration into meaningful format */
     ts+="Duration for Entire Bid:\t"+_bid.getDuration()+"\n";
     ts+="Bid Could be Started as Early As:\t"+_bid.getEarlyStart()+"\n";
     ts+="Bid Must be Started no Later Than:\t"+_bid.getLateStart()+"\n";
     ts+="Bid Could be Finished as Early As:\t"+_bid.getEarlyFinish()+"\n";
     ts+="Bid Would be Finished no Later Than:\t"+_bid.getLateFinish()+"\n";
     ts+="Bid Needs to be Accepted by:\t"+_bid.getLatestAccept()+"\n";
     ts+="\nTasks Included in this Bid:\n\n";

     taskSet = _bid.getTaskSet();
     if (taskSet.isEmpty()) {
      ts+="\tThere are no tasks for this bid.\n";
     } else {
       for (Iterator tsi = taskSet.iterator(); tsi.hasNext();) {
            String taskId = (String) tsi.next();
            ts+="\tTask ID:\t"+taskId+"\n";
            ts+="\t\tDuration:\t"+_bid.getDuration(taskId)+"\n";
            ts+="\t\tEarliest Start Date:\t"+_bid.getEarlyStart(taskId)+"\n";
            ts+="\t\tLatest Start Date:\t"+_bid.getLateStart(taskId)+"\n";
            ts+="\t\tEarliest Finish Date:\t"+_bid.getEarlyFinish(taskId)+"\n";
            ts+="\t\tLatest Finish Date:\t"+_bid.getLateFinish(taskId)+"\n";
            ts+="\t\tAssociated Milestone Task:\t"+_bid.getMilestone(taskId)+"\n";
            ts+="\n";
      } // end of for loop
    } // end of if-else loop
    return ts;
  } // end of PrintBid


  /** The following method is to allow the award of bids to be in one place
   *  Not sure if this is right - but this is where it is
   *
   * Before this is called the data should be validated -
   * No validation of the TaskBidPlan is done here
   */
  public static void doAward(TaskBidPlan tbp, PlanStatus ps, Frame frame,
                              BidEditorPanel bpe) {

    // set the PlanStatus data
    ps.setTaskBidPlan(tbp);
    ps.setState(ps.EXECUTION);

    // Update the BidManager
    MagnetClientMain.getBidManager().setCurrentPlan(tbp);
    //removes all the other Evaluation Windows that are open
    bpe.removeEvaluationWindows();
    new AwardDialog(frame, tbp, false, bpe);
    bpe.setEditable(false);
    return;
  } // end of doAward

} // end of GuiEvalUtil
