/**
 * ------------------------------------------------------------------------
 * MAGNET User Interface components
 * $Id: HighlightedTableCellRenderer.java,v 1.4 2002/01/14 15:02:40 jcollins Exp $
 * Copyright (c) 2000,2001 University of Minnesota
 * ------------------------------------------------------------------------
 * Revision History:
 *
 * 13May2001 DFK - Bug #0000053: initial submission
 */

package edu.umn.magnet.client.userif;

import java.awt.*;
import javax.swing.JTable;
import javax.swing.table.DefaultTableCellRenderer;

import org.apache.log4j.Category;

/**
 * A table cell renderer that supports 5 different preset styles of
 * row-highlighting.  You can use the colored defaults (see
 * constructor), or create your own via <code>setColor()</code>.  A
 * code example (a selected bidtable row will highlight tasktable
 * tasks) can be seen in <code>BidEditorPanel.loadBidDataFromTable</code>.
 */
public class HighlightedTableCellRenderer extends DefaultTableCellRenderer {
  /**
   * Enums for color selection.
   * They are combined like this:
   *   <code>getColor(C_NOT_SELECTED | C_HIGHLIGHT1 | C_BACKGROUND); </code>
   * The "NOT" items (not selected, not highlighted, background) are defaults
   * (they're zeroes) so they don't contribute to the ORed bitmask,
   * so you don't need to use 'em..
   * HIGHLIGHTs 1 thru 5 should be considered mutually exclusive,
   * but I'm not gonna rule-check 'em.
   */
  public static final int C_BACKGROUND      = 0<<0;
  public static final int C_FOREGROUND      = 1<<0;
  public static final int C_NOT_SELECTED    = 0<<1;
  public static final int C_SELECTED        = 1<<1;
  public static final int C_NOT_HIGHLIGHT0  = 0<<2;
  public static final int C_HIGHLIGHT0      = 1<<2;
  public static final int C_NOT_HIGHLIGHT1  = 0<<3;
  public static final int C_HIGHLIGHT1      = 1<<3;
  public static final int C_NOT_HIGHLIGHT2  = 0<<4;
  public static final int C_HIGHLIGHT2      = 1<<4;
  public static final int C_NOT_HIGHLIGHT3  = 0<<5;
  public static final int C_HIGHLIGHT3      = 1<<5;
  public static final int C_NOT_HIGHLIGHT4  = 0<<6;
  public static final int C_HIGHLIGHT4      = 1<<6;
  /** Keep updating these to monitor the first/last position.. */
  private static final int C_MIN            = 0<<0;
  private static final int C_MAX            = 1<<7;
  /** The private array of colors.. */
  private Color[] color;
  /** An array of the highlight enums. */
  private int[] highlights = new int[] { C_HIGHLIGHT0,
					 C_HIGHLIGHT1,
					 C_HIGHLIGHT2,
					 C_HIGHLIGHT3,
					 C_HIGHLIGHT4 };

  /** log category */
  private static Category log =
    Category.getInstance(HighlightedTableCellRenderer.class.getName());

  /* Constructor, which sets up default colors.
   *  Currently, highlight 0 is this yucky red color, and highlight 1 is a
   *  more soothing grey color.
   *  You can override these with {@link HighlightedTableCellRenderer#setColor(Color)}
   *   to anything you want, or edit this constructor so everyone can enjoy your overrides.
   */
  public HighlightedTableCellRenderer()
  {
    super();
    color = new Color[C_MAX];
    // For now, assign each highlight the SAME color scheme.
    // People can override them (or change this code) as they wish
    // using setColor().. 
    for (int i=0; i < highlights.length; i++) {
      color[C_FOREGROUND | C_SELECTED | highlights[i]] = Color.white;
      color[C_FOREGROUND | C_SELECTED | 0/*C_NOT_HIGHLIGHT1*/] = Color.white;
      color[C_BACKGROUND | C_SELECTED | highlights[i]] = Color.red;
      color[C_BACKGROUND | C_SELECTED | 0/*C_NOT_HIGHLIGHT1*/] = Color.blue;

      color[C_FOREGROUND | C_NOT_SELECTED | highlights[i]] = Color.green;
      color[C_FOREGROUND | C_NOT_SELECTED | 0/*C_NOT_HIGHLIGHT1*/] = Color.black;
      color[C_BACKGROUND | C_NOT_SELECTED | highlights[i]] = Color.red;
      color[C_BACKGROUND | C_NOT_SELECTED | 0/*C_NOT_HIGHLIGHT1*/] = Color.white;
    }

    // DFK:  I like this "light-gray" scheme;  set it up as highlight #1.
    color[C_HIGHLIGHT1 | C_FOREGROUND | C_SELECTED] = Color.yellow;
    color[C_HIGHLIGHT1 | C_BACKGROUND | C_SELECTED] = Color.green;
    color[C_HIGHLIGHT1 | C_FOREGROUND | C_NOT_SELECTED] = Color.black;
    color[C_HIGHLIGHT1 | C_BACKGROUND | C_NOT_SELECTED] = Color.lightGray;
  };

  /** A quickie debugging tool to identify which color is being referenced. */
  private String identify(int which) {
    String s = new String("Pattern "+which+": ");
    s += (0!=(which&C_FOREGROUND)) ? "Foreground, " : "Background, ";
    s += (0!=(which&C_SELECTED)) ? "Selected, " : "Not Selected, ";
    boolean found=false;
    for (int i=0; i<highlights.length; i++) {
      if (0!=(which&highlights[i]))  {
        found=true;
        s+= "Highlight "+i+", ";
      }
    }
    if (!found)  s += "No highlights";
    return s;
  }

  /** Get the requested color.
   *  Example:
   *  <code>
   *    HighlightedTableCellRenderer r = new HighlightedTableCellRenderer();
   *    Color c = r.getColor(r.C_FOREGROUND | C_SELECTED | C_NOT_HIGHLIGHT1);
   *  </code>
   *  @param whichType Specifies which color to return, like foreground and selected and not highlight 1.
   *  @return The requested color, or Color.black if not found.
   */
  public Color getColor(int whichType) {
    if (whichType<C_MIN  ||  whichType>C_MAX) {  return Color.black;  }
    return color[whichType];
  }

  /** Override a particular color.
   *  Example:
   *  <code>
   *    HighlightedTableCellRenderer r = new HighlightedTableCellRenderer();
   *    r.setColor(r.C_FOREGROUND | C_SELECTED | C_NOT_HIGHLIGHT1, Color.blue);
   *  </code>
   *  @param whichType Specifies which color to return, like foreground and selected and not highlight 1.
   *  @param aColor The specified color, such as <tt>Color.blue</tt>
   */
  public void setColor(int whichType, Color aColor) {
//    log.debug("setColor("+identify(whichType)+","+aColor+"): entering");
    if (whichType<C_MIN  ||  whichType>C_MAX  ||  null==aColor) {  return;  }
    color[whichType]=aColor;
  }

//  /** The table cell foreground color when highlighted and selected. (Default: white) */
//  private Color cSelectedHighlightedForeground = Color.white;
//  /** The table cell foreground color when not highlighted and selected. (Default: white) */
//  private Color cSelectedNotHighlightedForeground = Color.white;
//
//  /** The table cell background color when highlighted and selected. (Default: red) */
//  private Color cSelectedHighlightedBackground = Color.red;
//  /** The table cell background color when not highlighted and selected. (Default: blue) */
//  private Color cSelectedNotHighlightedBackground = Color.blue;
//
//  /** The table cell foreground color when highlighted and not selected. (Default: green) */
//  private Color cNotSelectedHighlightedForeground = Color.green;
//  /** The table cell foreground color when not highlighted and not selected. (Default: black) */
//  private Color cNotSelectedNotHighlightedForeground = Color.black;
//
//  /** The table cell background color when highlighted and not selected. (Default: red) */
//  private Color cNotSelectedHighlightedBackground = Color.red;
//  /** The table cell background color when not highlighted and not selected. (Default: white) */
//  private Color cNotSelectedNotHighlightedBackground = Color.white;


  public Component getTableCellRendererComponent(JTable table, Object value,
                      boolean isSelected, boolean hasFocus, int row, int column)
  {
    Component comp =
        super.getTableCellRendererComponent(table,value,isSelected,hasFocus,row,column);
    // Put logic here that sets the background/foreground/whatever....
    TPTableModel m=null;
    if (null!=table  &&
        null!=(m=((TPTableModel)table.getModel())) )
    {
//      int[] highlights = new int[] {
//              C_HIGHLIGHT0, C_HIGHLIGHT1, C_HIGHLIGHT2, C_HIGHLIGHT3, C_HIGHLIGHT4 };
      // For each highlight type, do the specified rendering hints.
      for (int i=0/*H_TYPE0*/; i <= highlights.length/*4*//*H_TYPE4*/; i++) {
        boolean isHighlighted = m.rowIsHighlighted(row, i);
        if (isHighlighted) {
//        if (isSelected) {
          Color fgColor = isSelected
                                ? color[C_SELECTED|highlights[i]|C_FOREGROUND]
                                : color[C_NOT_SELECTED|highlights[i]|C_FOREGROUND] ;
          Color bgColor = isSelected
                                ? color[C_SELECTED|highlights[i]|C_BACKGROUND]
                                : color[C_NOT_SELECTED|highlights[i]|C_BACKGROUND] ;
          comp.setForeground( fgColor);
          comp.setBackground( bgColor);
//          log.debug("HTCR.getTableCellRendererComponent(): Row "+row+", col="+column
//            +", highlight="+i+", select="+isSelected+", foreground="+fgColor+", background="+bgColor
//            );
          return comp;
        }//if highlighted
      }//for

      // If we made it this far, this row wasn't highlighted.
      Color fgColor = isSelected
                            ? color[C_SELECTED|0/*C_NOT_HIGHLIGHTED*/|C_FOREGROUND]
                            : color[C_NOT_SELECTED|0/*C_NOT_HIGHLIGHTED*/|C_FOREGROUND] ;
      Color bgColor = isSelected
                            ? color[C_SELECTED|0/*C_NOT_HIGHLIGHTED*/|C_BACKGROUND]
                            : color[C_NOT_SELECTED|0/*C_NOT_HIGHLIGHTED*/|C_BACKGROUND] ;
      comp.setForeground( fgColor);
      comp.setBackground( bgColor);
//          log.debug("HTCR.getTableCellRendererComponent(): Row "+row+", col="+column
//            +", highlight=false, select="+isSelected+", foreground="+fgColor+", background="+bgColor
//            );
    }//if
    return comp;
  }
}
