/*------------------------------------------------------------------------
 * MAGNET User Interface components
 * $Id: TPTableModel.java,v 1.14.2.1 2002/05/05 19:42:02 ashutosh Exp $
 * Copyright (c) 2000 Regents of the University of Minnesota.
 * All Rights Reserved by the Regents of the University of Minnesota.
 * This file contains Original Code and/or Modifications of Original
 * Code as defined in the MAGNET Public Source License, which is
 * available at http://www.cs.umn.edu/magnet/source_license_v1.txt.
 * You may not use this file except in compliance with the License.
 *------------------------------------------------------------------------
 * Revision History:
 *
 *
 * 18Feb2001 DFK - Initial Entry
 * 11Apr2001 DFK - added JTable- and graphic- position-related
 *                  methods (eg getCellRect()).
 * 17Apr2001 DFK - added additional table column enums for JobInfo fields.
 * 07May2001 DFK - temp comment out System.out.printlns.. :-P  needs log4j!!
 * 13May2001 DFK - Bugs #0000052,0000053: Add support for highlighting table rows
 *                   and adding Gantt bid objs.
 * 15May2001 DFK - Bug #0000079: Convert Date/Time formats to short format.
 * Wed May 23 11:48:37 2001  John Collins - Set up for logging.
 * 27May2001 DFK - Bug #0000120: Fix sortedList -related positional error stuff.
 */

package edu.umn.magnet.client.userif;

import javax.swing.table.*;
import javax.swing.JTable;  // for javadoc
import java.awt.Rectangle;
import java.util.*;
import java.text.DateFormat;
import javax.swing.event.EventListenerList;  // for listeners

import org.apache.log4j.Category;

import edu.umn.magnet.protocol.cust.PlanListener;
import edu.umn.magnet.protocol.cust.SubtaskRequest;
import edu.umn.magnet.protocol.cust.TaskPlan;
import edu.umn.magnet.protocol.cust.TaskWithPrecedence;
import edu.umn.magnet.protocol.cust.TPEvent;
import edu.umn.magnet.protocol.JobInfo;

/**
 * Provides a "wrapping" around the {@link TaskPlan} data object,
 * suitable for supporting a {@link JTable}.
 * <p>
 * It's implied that the following "columns" (types of {@link TaskPlan} data)
 * will be accessible:
 * <ul>
 *    <li> Task number (calculated from precedence ordering)
 *    <li> Task ID
 *    <li> Task name (user-provided)
 *    <li> Predecessors to this task
 *    <li> Early start time
 *    <li> Late start time
 *    <li> Deadline time
 *    <li> Duration
 *    <li> Early finish time
 *    <li> Late finish time
 *    <li> Description
 *    <li> Comments (user-provided)
 * </ul>
 * (Feel free to add more as you need 'em..)
 */
public class TPTableModel extends AbstractTableModel
  implements PlanListener, TableModel, MagnetClientConstants
{
  private static Category log
    = Category.getInstance(TPTableModel.class.getName());

  // Enums for columns
    /** Enum for table column for Task Number
     *  (calculated from precedence ordering)  */
  public static final int COL_TASK_NUM              = 0;
    /** Enum for table column for Task ID string */
  public static final int COL_TASK_ID               = 1;
    /** Enum for table column for Task Name */
  public static final int COL_TASK_NAME             = 2;
    /** Enum for table column for this task's Predecessors */
  public static final int COL_TASK_PREDS            = 3;
    /** Enum for table column for Early Start Time */
  public static final int COL_TASK_ESTART           = 4;
    /** Enum for table column for Late Start Time */
  public static final int COL_TASK_LSTART           = 5;
    /** Enum for table column for this task's Duration */
  public static final int COL_TASK_DURATION         = 6;
    /** Enum for table column for Early Finish Time */
  public static final int COL_TASK_EFINISH          = 7;
    /** Enum for table column for Late Finish Time */
  public static final int COL_TASK_LFINISH          = 8;
    /** Enum for table column for Task Description string */
  public static final int COL_TASK_DESCRIPTION      = 9;
    /** Enum for table column for Task Deadline time */
  public static final int COL_TASK_DEADLINE         = 10;
    /** Enum for table column for user-furnished Comments */
  public static final int COL_TASK_COMMENTS         = 11;
    /** Enum for table column for JobInfo Task Type */
  public static final int COL_JOBINFO_TASK_TYPE     = 12;
    /** Enum for table column for JobInfo Expected Duration */
  public static final int COL_JOBINFO_EXP_DURATION  = 13;
    /** Enum for table column for JobInfo Expected Price */
  public static final int COL_JOBINFO_EXP_PRICE     = 14;
    /** Enum for table column for JobInfo description */
  public static final int COL_JOBINFO_DESCRIPTION   = 15;
  /** @todo Add the rest of the JobInfo columns here and in getValueAt(). */
    /** Keeps track of end of enums */
  public static final int COL_MAX                   = 16;

  /** Our handle to the {@link TaskPlan} obj we're wrapping. */
  private TaskPlan _taskPlan = null;

  /** An internal hashtable that maps the view's Task Number string
   *  to a Task ID string.
   *  @todo DAVE: This seems hacky & ugly...
   */
  private HashMap _mapTaskRow2TaskId = null;
  /** An internal hashtable that maps the view's Task ID string
   *  to a Task Number string.
   *  @todo DAVE: This seems hacky & ugly...
   */
  private HashMap _mapTaskId2TaskRow = null;
  /** Used by {@link #flattenTP(TaskPlan)} to
   *  hold a sorted list of tasks.
   */
  private AbstractList _sortedList = null;
  /** Maintains a vector list of column titles.
   *  @todo Stolen from DefaultTableModel; Untested as of 2/24/01; dfk
   */
  private Vector columnIdentifiers = null;

  /** Keep a handle to the current table view. */
  /*private*/ JTable _table = null;

  /** Precalc full date formatter, eg. "Jan 12, 1952 3:30:32pm". */
  private DateFormat dateFormat = DateFormat.getDateTimeInstance(DateFormat.SHORT, DateFormat.SHORT);
//  private DateFormat dateFormat = DateFormat.getDateTimeInstance(DateFormat.MEDIUM, DateFormat.LONG);

  /** List of listeners */
  protected EventListenerList _listenerList = null;
  /** An array of Vectors of highlighted rows.
   *  Each vector represents a highlight style, and each Vector's elts are TaskID Strings.
   *  You DON'T want to store positional numbers, because task additions/deletions
   *  can ruin the positionals.   Use the access methods instead, like getHighlightedRows()..
   */
  protected Vector[] highlightedRows = new Vector[5];


  /** Empty constructor.
   * Create a new instance of {@link TPTableModel} which
   *  doesn't yet wrap a {@link TaskPlan}, and supports a
   *  table of {@link #COL_MAX} columns, 0 rows.
   */
  public TPTableModel() {
    setColumnIdentifiers();
    _listenerList = new EventListenerList();
  }

  /** Create a new instance of {@link TPTableModel} which
   *  wraps the specified {@link TaskPlan} object.
   */
  public TPTableModel(TaskPlan tp) {
    setColumnIdentifiers();

    setTaskPlan(tp);
    _listenerList = new EventListenerList();
  }

  public int getColumnCount() {
    return COL_MAX;
  }

  public Object getValueAt(int row, int col) {
    if (null==_taskPlan) { return new String(""); }
    if (row < 0) { return new String(""); }

    // step 1: Get the SubtaskRequest
    String taskId = taskNum2TaskId(row);
    SubtaskRequest req = _taskPlan.getTaskData(taskId);
    // step 2: get/calc the proper field
    switch (col) {
      case COL_TASK_NUM:
          return new String(""+(row+1));
      case COL_TASK_ID:
          return new String( ((SubtaskRequest)_sortedList.get(row)).getId() );
      case COL_TASK_NAME:
          return new String( ((SubtaskRequest)_sortedList.get(row)).getTaskName() );
      case COL_TASK_PREDS:
          String preds = new String("");
          for (Iterator iter = _taskPlan.getPredecessors( (SubtaskRequest)_sortedList.get(row) );
                                    iter.hasNext(); )
          {
            if (preds.length() > 0) {  preds += ",";  }
            preds += ""+taskId2TaskNum( ((SubtaskRequest)iter.next()).getId() );
          }
          return preds;
      case COL_TASK_ESTART:
          Date myEStartDate = ((SubtaskRequest)_sortedList.get(row)).getEarliestTimeToStart();
          return new String( dateFormat.format(myEStartDate) );
      case COL_TASK_LSTART:
          return new String("");
      case COL_TASK_DEADLINE:
          Date myDeadlineDate = ((SubtaskRequest)_sortedList.get(row)).getDeadline();
          return new String( dateFormat.format(myDeadlineDate) );
      case COL_TASK_DURATION:
          String myDateString = new String("");
          Date myStartDate = ((SubtaskRequest)_sortedList.get(row)).getEarliestTimeToStart();
          Date myFinishDate = ((SubtaskRequest)_sortedList.get(row)).getDeadline();
          log.debug("getValAt: Duration: calc span betw start("
                +DateFormat.getInstance().format(myStartDate)+") and finish("
                +DateFormat.getInstance().format(myFinishDate)+")"
                );
          return new String(timeSpan( myStartDate, myFinishDate ));
      case COL_TASK_EFINISH:
          return new String("");
      case COL_TASK_LFINISH:
          Date myLFinishDate = ((SubtaskRequest)_sortedList.get(row)).getDeadline();
          return new String( dateFormat.format(myLFinishDate) );
      case COL_TASK_DESCRIPTION:
          return new String( ((SubtaskRequest)_sortedList.get(row)).getJobInfo().getTaskType() );
      case COL_TASK_COMMENTS:
          return new String( ((SubtaskRequest)_sortedList.get(row)).getTaskComments() );
      // All JobInfos are handled similarly;
      //  grab a valid instance, then invoke methods on it.
      /** @todo Add the rest of the JobInfo's..  Remove any prior redundancies.. */
      case COL_JOBINFO_TASK_TYPE:
      case COL_JOBINFO_EXP_DURATION:
      case COL_JOBINFO_EXP_PRICE:
      case COL_JOBINFO_DESCRIPTION:
          JobInfo ji = ((SubtaskRequest)_sortedList.get(row)).getJobInfo();
          if (null==ji)  {  return new String("");  }
          switch (col) {
            case COL_JOBINFO_TASK_TYPE:
                return new String(""+ji.getTaskType());
            case COL_JOBINFO_EXP_DURATION:
                return new String(""+ji.getExpectedDuration());
            case COL_JOBINFO_EXP_PRICE:
                return new String(""+ji.getExpectedPrice());
            case COL_JOBINFO_DESCRIPTION:
                return new String(""+ji.getDescription());
            default:
                return new String("");
          }//switch2
      default:
          return new String("");
    }//switch
    // step 3: return as string
  }

  /** Return the number of rows in this {@link TaskPlan},
   *  or 0 if no {@link TaskPlan} exists.
   */
  public int getRowCount() {
    if (null==_taskPlan) {
      return 0;
    } else {
      return _taskPlan.size();
    }
  }

  /** Dump a row of {@link TPTableModel} info to a String, for debugging. */
  public String dumpRow(int row) {
      String NL = new String("");  // sometimes it's "\n"
      String s = new String("");
      s += "\t COL_TASK_NUM: " +getValueAt(row,COL_TASK_NUM);
      s += NL+"\t COL_TASK_ID: " +getValueAt(row,COL_TASK_ID);
      s += NL+"\t COL_TASK_NAME: " +getValueAt(row,COL_TASK_NAME);
      s += NL+"\t COL_TASK_PREDS: " +getValueAt(row,COL_TASK_PREDS);
      s += NL+"\t COL_TASK_ESTART: " +getValueAt(row,COL_TASK_ESTART);
      s += NL+"\t COL_TASK_LSTART: " +getValueAt(row,COL_TASK_LSTART);
      s += NL+"\t COL_TASK_DEADLINE: " +getValueAt(row,COL_TASK_DEADLINE);
      s += NL+"\t COL_TASK_DURATION: " +getValueAt(row,COL_TASK_DURATION);
      s += NL+"\t COL_TASK_EFINISH: " +getValueAt(row,COL_TASK_EFINISH);
      s += NL+"\t COL_TASK_LFINISH: " +getValueAt(row,COL_TASK_LFINISH);
      s += NL+"\t COL_TASK_DESCRIPTION: " +getValueAt(row,COL_TASK_DESCRIPTION);
      s += NL+"\t COL_TASK_COMMENTS: " +getValueAt(row,COL_TASK_COMMENTS);
      s += NL;
      return s;
  }

  /** Given the view's task number, fetch that {@link SubtaskRequest}'s ID string.
   *  If that taskNum doesn't exist, return an empty String.
   */
  public String taskNum2TaskId(int taskNum) {
    if (1>taskNum  ||
        taskNum>getRowCount()  ||
        null==_mapTaskRow2TaskId  ||
        _mapTaskRow2TaskId.isEmpty()
        ) {
      return new String("");
    }
    // Use (taskNum-1) because taskNums start at 1, but rows start at 0.
    String s = (String)_mapTaskRow2TaskId.get(new String(""+(taskNum-1) ));
    if (null==s) { s = new String(""); }
    return s;
  }
  /** Given the {@link SubtaskRequest}'s ID string, fetch the view's task number.
   *  @return -1 if not found, or the position [1..??] in the list.
   */
  public int taskId2TaskNum(String taskId) {
    if (null==taskId  ||
        0==taskId.length()  ||
        null==_mapTaskId2TaskRow  ||
        _mapTaskId2TaskRow.isEmpty()
        ) {
      return -1;
    }
    int i;
    try {
      Integer tempI = new Integer((String)_mapTaskId2TaskRow.get(taskId));
      i = (null==tempI) ? -1 : tempI.intValue();
    } catch (Exception e) {
      return -1;
    }
    // Use (i+1) because rows start at 0, but taskNums start at 1.
    return (i+1);
  }

  /** Return a ref to the current {@link TaskPlan}. */
  public TaskPlan getTaskPlan() {
    return _taskPlan;
  }

  /** Set the current {@TaskPlan}.
   *  The task num <-> task Id mappings are automatically recalculated.
   */
  public void setTaskPlan(TaskPlan aTaskPlan) {
    if (null!=_taskPlan) {
      _taskPlan.removePlanListener(this);
    }
    _taskPlan = aTaskPlan;
    _taskPlan.addPlanListener(this);
    refreshModel();
  }

  /** Update model and view anytime data is changed/updated. */
  public void refreshModel() {
    recalcTaskNumIdMappings();
    fireTableDataChanged();     // notify all registered TableModelEvent listeners
  }

  /** Update/recalc the taskNum <-> taskId mapping hashtables.
   *  This should be done any time the {@link TaskPlan} is set
   *  or modified.
   */
  private void recalcTaskNumIdMappings() {
    if (null==_taskPlan) { return; }

    _mapTaskRow2TaskId = new HashMap();
    _mapTaskId2TaskRow = new HashMap();
    _sortedList = flattenTP( _taskPlan );
    for (int i = 0; i < _sortedList.size(); i++ ) {
      SubtaskRequest req = (SubtaskRequest)_sortedList.get(i);
      String reqId = req.getId();
      _mapTaskRow2TaskId.put(new String(""+i), reqId);
      _mapTaskId2TaskRow.put(reqId, new String(""+i));
    }
  }

  /** Given a {@link TaskPlan}, "flatten" it out into an ordered list
   *  of tasks suitable for printing or Gantt'ing.
   *  @return An ordered list of {@link SubtaskRequest} objects.
   */
  synchronized public AbstractList flattenTP(TaskPlan tp) {
    ArrayList sortedList = new ArrayList();
    addSubordinatesOf( null /*1st time*/, sortedList, tp);
    return sortedList;
  }

  /** Add this node's next-hop successors into the list.
   *  This method recurses at each node in the TaskPlan.
   *  @param req    The parent node to inspect, or <tt>null</tt> to signify
   *                we should grab the TaskPlan's successor roots (heads).
   *  @param sortedList The ordered list to which we insert tasks.
   *  @param tp     The {@link TaskPlan} containing all the
   *                {@link SubtaskRequest}s.
   */
  private void addSubordinatesOf(SubtaskRequest req,
                                 AbstractList sortedList,
                                 TaskPlan tp)
  {
    ArrayList children = new ArrayList();
    // Keep a mapping from a node to an ArrayList of its predecessors.
    HashMap parentsMap = new HashMap();

    //	a. From the current node, find all immediate child nodes
    //		(i.e. they're the next hop) and shove them into a Set.
    if (null==req) {
      // 1st time: just grab the TaskPlan's head roots
      for (Iterator iter = tp.getSuccessorRoots(); iter.hasNext(); ) {
        children.add( (SubtaskRequest)iter.next() );
      }
    } else {
      for (Iterator iter = tp.getSuccessors(req); iter.hasNext(); ) {
        SubtaskRequest nextReq = (SubtaskRequest)iter.next();
        //	b. If not all of each child's predecessors has been added
        //		to the list, cull them from the Set.
        ArrayList preds = new ArrayList();
        for (Iterator iter2 = tp.getPredecessors(nextReq); iter2.hasNext(); ) {
          preds.add( (SubtaskRequest)iter2.next() );
        }
        parentsMap.put(nextReq, preds);   // Remember this child's parents
        boolean hasAllPreds = true;
        for (int i=0; i < preds.size(); i++) {
          if (!sortedList.contains( preds.get(i))) {
            hasAllPreds = false;
            break;
          }
        }
        if (hasAllPreds) {  children.add( nextReq );  }
      }
    }
    //	c. Sort the Set (based on EarlyStart, then Deadline,
    //                  then Alphabetical, then ??).
    Object[] childrenArray = children.toArray();
    TaskComparator taskComparator = new TaskComparator();
    Arrays.sort(childrenArray, taskComparator);
//    // Also save sorted array as an ArrayList..
//    children = new ArrayList( Arrays.asList( childrenArray ));
    //	d. Insert them after their predecessor that occurs furthest
    //		down in the list.
    if (null==req) {
      for (int i=0; i < childrenArray.length; i++) {
        sortedList.add( childrenArray[i] );
      }
    } else {
      for (int i=0; i < childrenArray.length; i++) {
        int insertAt = -1;
        ArrayList nodeParents = (ArrayList)parentsMap.get( childrenArray[i] );
        if (null!=nodeParents) {
          for (int j=0; j < nodeParents.size(); j++) {
            int positionInList = sortedList.indexOf( nodeParents.get(j) );
            if (positionInList > insertAt) {  insertAt=positionInList;  }
          }
        }
        // Bug 120: also advance past any current peers that are already added
        for (int j=0; j < childrenArray.length; j++) {
          int positionInList = sortedList.indexOf( childrenArray[j] );
          if (positionInList > insertAt) {  insertAt=positionInList;  }
        }
        /** for this child, insertAt now points at the furthest parent (or peer)
         *  in <tt>sortedList</tt> (or <tt>-1</tt> if no matches).
         *  Advance the pointer and insert at the resulting location.
         */
        insertAt++;
        // Bug 120: it's adding dupes; check for !present before adding..
        if (!sortedList.contains(childrenArray[i])) {
          sortedList.add( insertAt, childrenArray[i] );
        }
      }
    }
    //	e. For each childnode in the Set, recurse (goto a.) at that node.
    for (int i=0; i < childrenArray.length; i++) {
      addSubordinatesOf( (SubtaskRequest)childrenArray[i], sortedList, tp);   //recurse
    }
  }//end addSubordinatesOf()


  /** Return the amt of time betw 2 dates, like "5y 3d 4h 2m 1.03s".
   *  Warning: This method is <b>ULTRA-SLOW</b>!!  Is there a better way of doing this?
   */
  public static String timeSpan(Date start, Date finish) {
    log.debug("timeSpan(): Entering with "
      +",start="+DateFormat.getInstance().format(start)
      +",finish="+DateFormat.getInstance().format(finish)
      );
    int multiplier=0;
    /** *** NOTE ***  dfk 18Feb2001
     *  Don't share the same dateFormat amongst Calendar objs !!!
     *  It seems to set them all equal to the same date  (ugh!!).
     */
    Calendar cStart = DateFormat.getInstance().getCalendar();
    Calendar cFinish = DateFormat.getInstance().getCalendar();
//    int years=-1, days=-1, hours=-1, minutes=-1, seconds=-1, millis=-1;
    int years=0, days=0, hours=0, minutes=0, seconds=0, millis=0;
    cStart.setTime(start);
    cFinish.setTime(finish);
    Calendar cPrev = DateFormat.getInstance().getCalendar();  cPrev.setTime(cStart.getTime());

    log.debug("timespan: new dateformat instances: before loops: "
      +"prev="+DateFormat.getInstance().format(cPrev.getTime())
      +",start="+DateFormat.getInstance().format(cStart.getTime())
      +",finish="+DateFormat.getInstance().format(cFinish.getTime())
      );

    /** I started using a "multiplier" around the potentially-big loops
     *  to take bigger steps initially,
     *  so it geometrically cuts down on looping time..
     *  First guess: multiplier/divisor combo that "fast-fails" in 3 rounds
     *  (eg 100,10,1  or  20,5,1)
     */

    // years
    for (multiplier=100; multiplier>0; multiplier /= 10) {
      cStart.setTime(cPrev.getTime());
      while (cFinish.after(cStart)) {
        cPrev.setTime( cStart.getTime() );
        cStart.add(Calendar.YEAR, multiplier);
        if (cFinish.after(cStart)) {
          years += multiplier;
        } else {
          break;  // out to for
        }
      }
    }
    // days
    for (multiplier=100; multiplier>0; multiplier /= 10) {
      cStart.setTime(cPrev.getTime());
      while (cFinish.after(cStart)) {
        cPrev.setTime( cStart.getTime() );
        cStart.add(Calendar.DAY_OF_YEAR, multiplier);
        if (cFinish.after(cStart)) {
          days += multiplier;
        } else {
          break;  // out to for
        }
      }
    }
    // hours
    for (multiplier=23; multiplier>0; multiplier /= 4) {
      cStart.setTime(cPrev.getTime());
      while (cFinish.after(cStart)) {
        cPrev.setTime( cStart.getTime() );
        cStart.add(Calendar.HOUR, multiplier);
        if (cFinish.after(cStart)) {
          hours += multiplier;
        } else {
          break;  // out to for
        }
      }
    }
    // minutes
    for (multiplier=20; multiplier>0; multiplier /= 4) {
      cStart.setTime(cPrev.getTime());
      while (cFinish.after(cStart)) {
        cPrev.setTime( cStart.getTime() );
        cStart.add(Calendar.MINUTE, multiplier);
        if (cFinish.after(cStart)) {
          minutes += multiplier;
        } else {
          break;  // out to for
        }
      }
    }
    // seconds
    for (multiplier=20; multiplier>0; multiplier /= 4) {
      cStart.setTime(cPrev.getTime());
      while (cFinish.after(cStart)) {
        cPrev.setTime( cStart.getTime() );
        cStart.add(Calendar.SECOND, multiplier);
        if (cFinish.after(cStart)) {
          seconds += multiplier;
        } else {
          break;  // out to for
        }
      }
    }
    // millis
    for (multiplier=100; multiplier>0; multiplier /= 10) {
      cStart.setTime(cPrev.getTime());
      while (cFinish.after(cStart)) {
        cPrev.setTime( cStart.getTime() );
        cStart.add(Calendar.MILLISECOND, multiplier);
        if (cFinish.after(cStart)) {
          millis += multiplier;
        } else {
          break;  // out to for
        }
      }
    }

    // Build up the return string..
    String dateString = new String("");
    if (years > 0) {  dateString += ""+years+"y ";  }
    if (days > 0) {  dateString += ""+days+"d ";  }
    if (hours > 0) {  dateString += ""+hours+"h ";  }
    if (minutes > 0) {  dateString += ""+minutes+"m ";  }
    if (seconds >= 0) { dateString += ""+seconds+".";  }
    if (millis >= 0) {  dateString += ""+millis+"s";  }

    return dateString;
  }

  /** Called by the {@link TaskPlan} we're currently listening to when the TaskPlan changes.
   */
  public void planChanged(TPEvent e) {
    refreshModel();
    firePlanDataChanged(e);
  }

  // --------------------------------------------------------
  // These are default methods in AbstractTableModel that I'm overriding..

  /** Return the String name for this column.
   *  If column cannot be found, returns an empty String.
   */
  public String getColumnName(int column) {
    if (0>column  ||  COL_MAX<=column) {
      return new String("");
    }
    switch (column) {
      case COL_TASK_NUM:
          return new String("Task #");
      case COL_TASK_ID:
          return new String( "Task ID" );
      case COL_TASK_NAME:
          return new String( "Task Name" );
      case COL_TASK_PREDS:
          return new String("Task Predecessors");
      case COL_TASK_ESTART:
          return new String("Task Early Start");
      case COL_TASK_LSTART:
          return new String("Task Late Start");
      case COL_TASK_DEADLINE:
          return new String("Task Deadline");
      case COL_TASK_DURATION:
          return new String("Task Duration");
      case COL_TASK_EFINISH:
          return new String("Task Early Finish");
      case COL_TASK_LFINISH:
          return new String("Task Late Finish");
      case COL_TASK_DESCRIPTION:
          return new String("Task Description");
      case COL_TASK_COMMENTS:
          return new String("Comments");
      case COL_JOBINFO_TASK_TYPE:
          return new String("Task Type");
      case COL_JOBINFO_EXP_DURATION:
          return new String("Expected Duration");
      case COL_JOBINFO_EXP_PRICE:
          return new String("Expected Price");
      case COL_JOBINFO_DESCRIPTION:
          return new String("Description");
      default:
          return new String("");
    }//switch
  }

  /** Returns the class of the data stored in this column.
   *  Currently always returns {@link String} class.
   */
  public Class getColumnClass(int columnIndex) {
    return new String().getClass();
  }

  /** Returns the column index for the (first) column named <tt>columnName</tt>.
   *  @return the column with <tt>columnName</tt>, or -1 if not found
   */
  public int findColumn(String columnName) {
    for (int i=0; i < COL_MAX; i++) {
      if (getColumnName(i).equals(columnName)) {
        return i;
      }
    }
    return -1;
  }
   // --------------------------------------------------------
  /** Set the column titles. */
  public void setColumnIdentifiers() {
    Vector vColTitles = new Vector();
    for (int i=0; i < COL_MAX; i++) {
      vColTitles.add(getColumnName(i));
    }
    setColumnIdentifiers(vColTitles);
  }

    /**
     * Replaces the column identifiers in the model.  If the number of
     * <code>newIdentifier</code>s is greater than the current number
     * of columns, new columns are added to the end of each row in the model.
     * If the number of <code>newIdentifier</code>s is less than the current
     * number of columns, all the extra columns at the end of a row are
     * discarded. <p>
     *
     * @param   newIdentifiers  vector of column identifiers.  If
     *				<code>null</code>, set the model
     *                          to zero columns
     * @see "javax.swing.table.DefaultTableModel#setRowCount(int)"
     * @todo Stolen from DefaultTableModel.  Untested as of 2/24/01; dfk
     */
    public void setColumnIdentifiers(Vector newIdentifiers) {
        if (newIdentifiers != null) {
            columnIdentifiers = newIdentifiers;
        }
        else {
            columnIdentifiers = new Vector();
        }

        // Generate notification
        fireTableStructureChanged();
    }

//    /**
//     * Replaces the column identifiers in the model.  If the number of
//     * <code>newIdentifier</code>s is greater than the current number
//     * of columns, new columns are added to the end of each row in the model.
//     * If the number of <code>newIdentifier</code>s is less than the current
//     * number of columns, all the extra columns at the end of a row are
//     * discarded. <p>
//     *
//     * @param   newIdentifiers  array of column identifiers.
//     *				If <code>null</code>, set
//     *                          the model to zero columns
//     * @see #setNumRows
//     * @todo Stolen from DefaultTableModel.  Untested as of 2/24/01; dfk
//     */
//    public void setColumnIdentifiers(Object[] newIdentifiers) {
//        setColumnIdentifiers(convertToVector(newIdentifiers));
//    }

  /** Get the {@link SubtaskRequest} from the specified row of the table model.
   *  Although here Model row and View row are consistent,
   *  the columns are (potentially) INconsistent or absent.
   */
  public SubtaskRequest getTask(int aRow) {
    try {
      return (SubtaskRequest)_sortedList.get(aRow);
    } catch (IndexOutOfBoundsException e) {
      return null;
    }
  }

  /** Keeps a ref to the "view" component.
   *  (Do I really need to do this?)
   */
  public void setView(JTable aTable) {
    _table = aTable;
  }
  /** Retrieves the current "view" component, or null if none exists. */
  public JTable getView() {
    return _table;
  }

  /** Returns a Rectangle enclosing at least 1 cell in this task's row.
   *  This is used by other displays to "line up" stuff w/ table rows, so
   *  y-coord and width are not necessarily accurate..
   *  @return null if not found, otherwise the bounding Rectangle coords from
   *    the cell lookup, as specified by {@link JTable#getCellRect(int,int,boolean)}.
   *  @todo Check out {@link javax.swing.SizeSequence} (this might've been useful...)
   */
  public Rectangle getRowRect(int aRow) {
    JTable t = getView();                   // First, get the table.
    if (null==t)  {  return null;  }
    Rectangle r = t.getCellRect(aRow, this.COL_TASK_ID, false);
    if(false) log.info("\tgetRowRect(): requested cell rect for (r,c) ("
          + aRow+","+this.COL_TASK_ID+"), got:  "
          + "x="+r.x+",y="+r.y+",w="+r.width+",h="+r.height);
    // GET RID OF THIS..  just debugging...
    Rectangle r2 = t.getBounds();
    if(false) log.debug("\t\tparent table rect x="+r2.x+",y="+r2.y+",w="+r2.width+",h="+r2.height);

    return r;
  }

  /** Returns a Rectangle enclosing at least 1 cell in this task's row.
   *  Lookup is done via taskID, and defers to the row-number-lookup version.
   *  This is used by other displays to "line up" stuff w/ table rows, so
   *  y-coord and width are not necessarily accurate..
   *  @return null if not found, otherwise the bounding Rectangle coords from
   *    the cell lookup, as specified by {@link JTable#getCellRect(int,int,boolean)}.
   */
  public Rectangle getRowRect(String taskID) {
    int aRow = taskId2TaskNum(taskID) - 1;  // DAVE: row# = TaskNum - 1...
    return getRowRect(aRow);
  }

  // -------------------------------------------------------------------------
  // BEGIN PlanListener notifier methods
  //
  // DFK - Methods to support/feed PlanListener instances.
  /** Adds a listener to the list that's notified each time a change to the data model occurs. */
  public void addPlanListener(PlanListener l) {
    _listenerList.add(PlanListener.class, l);
  }

  /** Remove a listener from the notify-list. */
  public void removePlanListener(PlanListener l) {
    _listenerList.remove(PlanListener.class, l);
  }

  /** Indicate (in a vague sense) that the plan has somehow been changed. */
  public void firePlanDataChanged() {
    TPEvent myTPEvent = new TPEvent(this, "TaskPlan has been modified");  // generic for now
    firePlanDataChanged(myTPEvent);
  }

  /** Indicate that the plan has somehow been changed.
   *  @param aMessage Some arbitrary message (currently not used for anything useful)
   */
  public void firePlanDataChanged(String aMessage) {
    TPEvent myTPEvent = new TPEvent(this, aMessage);  // generic for now
    firePlanDataChanged(myTPEvent);
  }

  /** Indicate that the plan has somehow been changed.
   *  @param evt The {@link TPEvent} to broadcast to all listeners.
   */
  public void firePlanDataChanged(TPEvent evt) {
    // Guaranteed to return a non-null array
    Object[] listeners = _listenerList.getListenerList();
    // Process the listeners last to first, notifying
    // those that are interested in this event
    for (int i = listeners.length-2; i>=0; i-=2) {
//      System.err.println("dfk: firePlanDataChanged(): listeners["+i+"]="+listeners[i]);
//       if (listeners[i] instanceof edu.umn.magnet.userif.PlanListener) {
       if (listeners[i] == PlanListener.class) {
//           System.err.println("dfk: firePlanDataChanged(): listeners["+i+"] is a PlanListener");
           ((PlanListener)listeners[i+1]).planChanged(evt);
       }
    }
  }
  // END PlanListener notifier methods
  // -------------------------------------------------------------------------

  /** Get the list of highlighted task rows.
   *  @param hType The type of highlight; use H_TYPE0 .. H_TYPE4.
   *  @return an int array of Task rows, or null if bad hType or none highlighted.
   */
  public int[] getHighlightedRows(int hType) {
    int[] rows=null;
    if (H_TYPE0<=hType  &&  H_TYPE4>=hType  &&
        null!=highlightedRows[hType]  &&  0<highlightedRows[hType].size()) {
      rows = new int[highlightedRows[hType].size()];
      for (int i=0; i<rows.length; i++) {
        String taskID = (String)highlightedRows[hType].elementAt(i);
        int taskNum = taskId2TaskNum(taskID);
        if (-1==taskNum) {
          log.info("TPTableModel.getHighlightedRows(): now row for ID "+taskID);
          return null;
        }
        rows[i] = taskNum-1;    // row = taskNum - 1
      }
    }
    return rows;
  }
  /** Indicates whether the specified row is highlighted.
   *  @param aRow The row to check
   *  @param hType The type of highlight; use H_TYPE0 .. H_TYPE4.
   */
  public boolean rowIsHighlighted(int aRow, int hType) {
    if (    H_TYPE0>hType  ||  H_TYPE4<hType  ||
            0>aRow  ||  aRow>=getRowCount()  ||
            null==highlightedRows[hType]  ||  0==highlightedRows[hType].size()) {
      return false;   // no items highlighted
    }
    int[] rows = getHighlightedRows(hType);
    if (null==rows)  {  return false;  }
    for (int i=0; i < rows.length; i++) {
      if (aRow==rows[i]) {  return true;  }
    }
    return false;   // this row wasn't in the list of highlighted items
  }
//  /** Get the index of the first highlighted task row.  (Ignore the rest.)
//   *  @param hType The type of highlight; use H_TYPE0 .. H_TYPE4.
//   *  @return an int task row number, or -1 if none highlighted.
//   */
//  public int getHighlightedRow(int hType) {
//    if (H_TYPE0>hType  ||  H_TYPE4<hType  ||
//        null==highlightedRows[hType]  ||  0==highlightedRows[hType].size()) {
//      return -1;
//    } else {
//      try {
//        return Integer.parseInt((String)highlightedRows[hType].elementAt(0));
//      } catch (NumberFormatException nfe) {
//        log.info("TPTableModel.getHighlightedRow(): "+nfe.toString());
//        return -1;
//      }
//    }
//  }

  /** Highlight a task row
   *  @param aRow The row to highlight
   *  @param hType The type of highlight; use H_TYPE0 .. H_TYPE4.
   */
  public void highlightRow(int aRow, int hType) {
    // Look up the taskID
    String taskID = taskNum2TaskId(aRow+1);
    // Add it
    highlightRow(taskID, hType);
  }

  /** Highlight a task row
   *  @param taskID The String ID of the task to highlight
   *  @param hType The type of highlight; use H_TYPE0 .. H_TYPE4.
   */
  public void highlightRow(String taskID, int hType) {
    int aRow = taskId2TaskNum(taskID) -1; /* "off-by-one" row-to-taskNum conversion */
    if (H_TYPE0>hType  ||  H_TYPE4<hType  ||  0>aRow  ||  aRow>=getRowCount())  {  return;  }
    if (null==taskID  ||  0==taskID.length())  {  return;  }
    if (null==highlightedRows[hType]) {
      highlightedRows[hType] = new Vector();
    }
    // First, add elt to list.
//    String taskID = taskNum2TaskId(aRow+1);
    highlightedRows[hType].addElement(taskID);
    // Next, trigger a table redraw.  (Let table renderer do the highlighting dirtywork..)
    fireTableRowsUpdated(aRow,aRow);
  }

  /** Un-highlight a task row */
  public void unhighlightRow(int aRow) {
    GuiUtil.inform(null, "TPTableModel.unhighlightRow(): not impl yet!");
  }
  /** Un-highlight all rows */
  public void unhighlightAllRows() {
    // GuiUtil.inform(null, "TPTableModel.unhighlightRows(): not impl
    //    yet!");
    log.debug("Attempting to unhighlight all rows");
    for (int i=0; i < highlightedRows.length; i++) {
      highlightedRows[i] = new Vector();
    }
    fireTableRowsUpdated(0, getRowCount());
  }
}
