/*------------------------------------------------------------------------
 * Magnet utility components
 * $Id: NormalVariable.java,v 1.1.2.1 2002/05/05 19:42:04 ashutosh Exp $
 * Copyright (c) 1998 Regents of the University of Minnesota.
 * All Rights Reserved by the Regents of the University of Minnesota.
 * This file contains Original Code and/or Modifications of Original
 * Code as defined in the MAGNET Public Source License, which is
 * available at http://www.cs.umn.edu/magnet/source_license_v1.txt.
 * You may not use this file except in compliance with the License.
 *------------------------------------------------------------------------
 * Revision History:
 * 
 * Wed Sep 23 08:29:15 1998  John E. Collins
 *   Initial Entry, copied from jec.util
 */

package edu.umn.magnet.util.rvars;

import java.util.Random;

// ==================================================================

/**
   Produce random values with a normal distribution and the given mean
   and standard deviation values.
*/
public class NormalVariable extends RandomVariable 
{
  private double m_mean = 0.0;
  private double m_sigma = 1.0;
  
  private double savedValue;
  private boolean secondValue = false;

  /**
     Create a new NormalVariable instance with a mean of 0.0 and a sigma
     of 1.0.
  */
  public NormalVariable () {
    super();
  }

  /**
     Create a new NormalVariable instance with the specified mean and
     sigma values.
  */
  public NormalVariable (RandomVariableFactory factory,
			 double mean, double sigma) {
    super(factory);
    m_mean = mean;
    m_sigma = sigma;
  }

  /**
     Return a string naming this variable type
  */
  public String getLabel () {
    return "Normal";
  }
  
  /**
     Return the mean of the distribution of this variable.
  */
  public double getMean () {
    return m_mean;
  }
  
  /**
     Return the standard deviation of the distribution of this variable.
  */
  public double getSigma () {
    return m_sigma;
  }

  /**
     Change the mean of this distribution.
  */
  public void setMean (double newMean) {
    m_mean = newMean;
  }

  /**
     Change the standard deviation of this variable.
  */
  public void setSigma (double newSigma) {
    m_sigma = newSigma;
  }

  /**
     Return a random value as a double with the specified distribution.
  */
  public double nextDouble () {
    if (secondValue) {
      secondValue = false;
      return savedValue;
    }
    else {
      double s1 = m_sigma * Math.sqrt(-2.0 * Math.log(m_factory.nextDouble()));
      double s2 = 2.0 * Math.PI * m_factory.nextDouble();
      
      savedValue = m_mean + s1 * Math.sin(s2);
      secondValue = true;
      return m_mean + s1 * Math.cos(s2);
    }
  }

  /**
     Return a random value as a float with the specified distribution.
  */
  public float nextFloat () {
    return (float)this.nextDouble();
  }

  /**
     Return a random value as an integer with the specified distribution.
  */
  public int nextInt () {
    return (int)Math.round(this.nextDouble());
  }

  /**
     Return a random value as a long with the specified distribution.
  */
  public long nextLong () {
    return (long)Math.round(this.nextDouble());
  }
}
