/*------------------------------------------------------------------------
 * Magnet search utility components
 * $Id: SearchQueue.java,v 1.5.2.3 2002/05/05 19:42:05 ashutosh Exp $
 * Copyright (c) 1998 Regents of the University of Minnesota.
 * All Rights Reserved by the Regents of the University of Minnesota.
 * This file contains Original Code and/or Modifications of Original
 * Code as defined in the MAGNET Public Source License, which is
 * available at http://www.cs.umn.edu/magnet/source_license_v1.txt.
 * You may not use this file except in compliance with the License.
 *------------------------------------------------------------------------
 * Revision History:
 * 
 * Tue Sep  4 11:10:11 2001  John Collins - bug #0000158
 *   Set up SearchQueue factory, move Best First search to subclass
 *   that can handle the efficiency problems better.
 * 
 * Thu Dec 31 12:37:54 1998  John E. Collins
 *   Fill in code.
 * 
 * Thu Dec 31 09:07:15 1998  John E. Collins
 *   Initial Entry
 */

package edu.umn.magnet.util.search;

import org.apache.log4j.Category;

import java.util.Comparator;

/**
 * A SearchQueue is a queue abstraction for managing generalized
 * systematic and stochastic searches.  Nodes in the search must
 * implement the SearchNode interface.  You add nodes with
 * <tt>insert</tt>, you remove nodes with <tt>extract</tt>, and you
 * pick nodes without removing them using <tt>select</tt>.  The
 * following search disciplines are supported:
 * <ul>
 * <li> simple depth-first, in which the last node inserted is the
 *      next one extracted.
 * <li> simple breadth-first, in which the last node inserted is the
 *      last one extracted.
 * <li> best-first, in which the node with the "best" evaluation (as
 *      determined by a Comparator) is the next one extracted.
 * <li> beam, in which the number of nodes kept on the queue is
 *      limited, either by count or by failing a comparison against
 *      the "best" node using a RangeComparator.  It probably doesn't
 *      make sense to limit the node count for depth first or breadth
 *      first searches.
 * <li> stochastic, in which a double in the range [0.0 1.0] is used
 *      to make a selection, either by range (using a RangeComparator)
 *      or by node count.
 * </ul>
 * @author John Collins
 */
public class SearchQueue
{
  private static Category log =
    Category.getInstance(SearchQueue.class.getName());

  /**
   * Queue search discipline constants
   */
  public static final int BEST_FIRST = 1;
  public static final int BREADTH_FIRST = 2;
  public static final int DEPTH_FIRST = 3;

  protected int m_searchDiscipline = DEPTH_FIRST;
  protected int m_maxSize = 0;
  protected int m_sizeTolerance = 0;
  private boolean m_wasTrimmed = false;

  protected NodeHolder m_head = null;
  protected NodeHolder m_tail = null;
  protected int m_nodeCount = 0;
  
  /**
   * Default constructor and standard accessors makes this a Bean, but
   * you have to create one using the static create() method, which
   * will return a SearchQueue of an appropriate class to handle the
   * indicated search discipline.
   */
  protected SearchQueue ()
  {
    super();
  }

  /**
   * Factory method for creating these things.  The class you get back
   * meets the SearchQueue contract, but may be specialized to handle
   * the indicated <tt>searchDiscipline</tt>.
   * <p>
   * Only values of <tt>BEST_FIRST</tt>,
   * <tt>BREADTH_FIRST</tt>, and <tt>DEPTH_FIRST</tt> are accepted.
   * If any other value is provided, the result will be the default
   * <tt>DEPTH_FIRST</tt>.  Note that if <tt>BEST_FIRST</tt> is given,
   * the <tt>nodeComparator</tt> must be set unless you are happy with
   * the default <tt>MinimumIsBest</tt> comparator. </p>
   * <p>
   * It is currently not supported to change the search discipline of
   * an existing SearchQueue. </p>
   */
  static public SearchQueue create (int searchDiscipline)
  {
    SearchQueue result = null;
    if (searchDiscipline == BREADTH_FIRST
	|| searchDiscipline == DEPTH_FIRST) {
      result = new SearchQueue();
      result.m_searchDiscipline = searchDiscipline;
    }
    else if (searchDiscipline == BEST_FIRST)
    {
      result = new BFSearchQueue();
    }
    else { //bad value for searchDiscipline
      log.error("SearchQueue bad search discipline " + searchDiscipline);
    }
    return result;
  }

  // ----------------- Queue control attributes -------------------

  /**
   * Return the value of searchDiscipline.
   * @return Value of searchDiscipline.
   */
  public int getSearchDiscipline ()
  {
    return m_searchDiscipline;
  }
  
  /**
   * Return the maximum size of the queue.  If this is 0, the queue is
   * unlimited; otherwise you get beam-search behavior where nodes can
   * fall off the end.
   * @return Current maximum size of the queue.
   */
  public int getMaxSize ()
  {
    return m_maxSize;
  }

  /**
   * Sets the maximum size of the queue.  This gives you beam search
   * behavior, and only makes sense with a best-first search
   * discipline.  The default value is 0, which makes the queue
   * unlimited.  If the current size of the queue is greater than the
   * new maximum, nodes will be dropped off the end until the queue is
   * no larger than the new maximum.
   * @param v  New maximum size of the queue.
   */
  public void setMaxSize (int v)
  {
    m_maxSize = v;
    m_sizeTolerance = v/10;
    if (v > 0 && m_nodeCount > v)
      trimNodes();
  }

  /**
   * True just in case the queue has been trimmed. This can only
   * happen if the maxSize has been set, and more than MaxSize
   * elements have been inserted.
   */
  public boolean wasTrimmed ()
  {
    return m_wasTrimmed;
  }

  /**
   * Sets the wasTrimmed flag.
   */
  protected void setWasTrimmed ()
  {
    m_wasTrimmed = true;
  }
 
  /**
   * Returns the Comparator used to order nodes for best-first search.
   * For the base class, this is always null.
   */
  public Comparator getComparator ()
  {
    return null;
  }

  /**
   * Set the Comparator to be used to order nodes.  This doesn't do
   * anything in the base class.
   * @param v  Value to assign to comparator.
   */
  public void setComparator (Comparator v)
  {
  }

  // ---------------- Derived attributes ------------------

  /**
   * Returns a count of the number of nodes in the queue.
   */
  public int nodeCount ()
  {
    return m_nodeCount;
  }

  /**
   * Returns the number of nodes in the queue.
   */
  public int size ()
  {
    return m_nodeCount;
  }

  /**
   * Returns the evaluation of the first node in the queue.  This
   * probably only makes sense for a best-first queue.  If the queue
   * is empty, return 0.0.
   */
  public double firstValue ()
  {
    if (m_head != null)
      return m_head.value();
    else
      return 0.0;
  }

  /**
   * Returns the evaluation of the last node in the queue.  This only
   * makes sense for a best-first queue.  If the queue is empty,
   * return 0.0.
   */
  public double lastValue ()
  {
    if (m_tail != null)
      return m_tail.value();
    else
      return 0.0;
  }

  /**
   * Return true just in case the queue is empty.
   */
  public boolean isEmpty ()
  {
    return m_head == null;
  }

  // ------------------ Queue operations -------------------

  /**
   * Add a node to the queue according to the current search
   * discipline.
   */
  public void insert (SearchNode newNode)
  {
    NodeHolder it = new NodeHolder(newNode);
    if (m_searchDiscipline == DEPTH_FIRST)
      insertFirst(it);
    else if (m_searchDiscipline == BREADTH_FIRST)
      insertLast(it);
    // increment the count and trim if necessary
    m_nodeCount++;
    if (m_maxSize != 0
	&& m_nodeCount > (m_maxSize + m_sizeTolerance))
      trimNodes();
  }

  /**
   * Remove the next node from the queue according to the current
   * search discipline.
   */
  public SearchNode extract ()
  {
    // the queue is ordered, so we just extract and return the first one.
    NodeHolder result = m_head;
    if (result != null)
    {
      m_head = result.getNext();
      m_nodeCount--;
      return result.getNode();
    }
    else
      return null;
  }

  /**
   * Clears out the entire queue
   */
  public void clear ()
  {
    m_head = null;
    m_tail = null;
    m_nodeCount = 0;
  }

  /**
   * Access the next node on the queue without removing it.
   */
  public SearchNode select ()
  {
    // the queue is ordered, so we just return the first one.
    NodeHolder result = m_head;
    if (result != null)
      return result.getNode();
    else
      return null;
  }

  /**
   * Access a node by position.  The parameter is multiplied by the
   * current node count to determine the position in the queue to return.
   */
  public SearchNode selectByPosition (double mult)
  {
    if (m_head == null)
      return null;

    int position = (int)Math.round(m_nodeCount * mult);
    if (position <= 0)
      return (m_head.getNode());
    else if (position >= m_nodeCount - 1)
      return (m_tail.getNode());
    else
    {
      NodeHolder place = m_head;
      while (position > 0)
      {
	place = place.getNext();
	position--;
      }
      return place.getNode();
    }
  }

  /**
   * Access a node by predicate.  The queue is walked from the
   * beginning, and the last node for which the predicate's test
   * method returns true is returned. 
   * <p>
   * This admittedly odd
   * specification is intended to allow a finite
   * probability that the first node will be returned if the test is
   * based on a comparison within the value range of the queue.  As a
   * consequence of this twist, the first node is actually not tested;
   * instead, if the second node fails the test, the first node is
   * returned.
   */
  public SearchNode selectByPredicate (Tester predicate)
  {
    if (m_head == null)
      return null;

    NodeHolder place = m_head;
    while (place.getNext() != null
	   && predicate.test(place.getNext().getNode()))
      place = place.getNext();
    return place.getNode();
  }

  /**
   * Removes the given node from the queue.  This is done by a linear
   * walk down the queue.
   */
  public void remove (SearchNode node)
  {
    NodeHolder place = m_head;
    if (place.getNode() == node) {
      m_head = m_head.getNext();
      m_nodeCount--;
      return;
    }
    while (place.getNext() != null) {
      if (place.getNext().getNode() == node) {
	place.setNext(place.getNext().getNext());
	m_nodeCount--;
	if (place.getNext() == null)
	  m_tail = place;
	return;
      }
      place = place.getNext();
    }
  }

  /**
   * Trims the tail of a sorted queue, removing all elements that meet
   * the predicate. On the basic search queue, this doesn't actually
   * do anything.
   */
  public void trimByPredicate (Tester predicate)
  {
  }

  // --------------- local insertion methods ---------------------

  /**
   * Insert a node at the head of the queue, stackwise
   */
  protected void insertFirst (NodeHolder it)
  {
    if (m_head == null)
      m_tail=it;
    it.setNext(m_head);
    m_head = it;
  }

  /**
   * Insert a node at the tail of the queue, queuewise
   */
  protected void insertLast (NodeHolder it)
  {
    if (m_head == null)
      m_head = it;
    else
      m_tail.setNext(it);
    m_tail = it;
  }
  
  /**
   * Remove the last element from the queue
   */
  protected void trimNodes ()
  {
    if (m_head != null)
    {
      NodeHolder place = m_head;
      for (int i = 1; i < m_maxSize && place != null; i++)
	place = place.getNext();
      if (place != null)
      {
	m_tail = place;
	m_tail.setNext(null);
	m_nodeCount = m_maxSize;
	setWasTrimmed();
      }
    }
  }

  /**
   * These things hold nodes in the Queue.
   */
  class NodeHolder implements SearchNode
  {
    private SearchNode m_node;
    private NodeHolder m_next = null;
    
    protected NodeHolder (SearchNode node)
    {
      super();
      m_node = node;
    }
    
    public double value ()
    {
      return m_node.value();
    }
    
    protected SearchNode getNode ()
    {
      return m_node;
    }
    
    protected NodeHolder getNext ()
    {
      return m_next;
    }
    
    protected void setNext (NodeHolder next)
    {
      m_next = next;
    }
  }
}  // ------------------ end of class SearchQueue ------------------------
