/*------------------------------------------------------------------------
 * Magnet Utilities
 * $Id: Histogram.java,v 1.2.2.2 2002/05/05 19:42:05 ashutosh Exp $
 * Copyright (c) 1998 Regents of the University of Minnesota.
 * All Rights Reserved by the Regents of the University of Minnesota.
 * This file contains Original Code and/or Modifications of Original
 * Code as defined in the MAGNET Public Source License, which is
 * available at http://www.cs.umn.edu/magnet/source_license_v1.txt.
 * You may not use this file except in compliance with the License.
 *------------------------------------------------------------------------
 * Revision History:
 * 
 * Thu Aug 27 22:01:22 1998  John E. Collins
 *   Initial entry, copied from jec.util
 */

package edu.umn.magnet.util.stats;

import java.util.Enumeration;
import java.util.Vector;
import java.lang.Integer;

import java.util.Hashtable;


// ==================================================================

/**
 * Builds a histogram of observations.
 */
public class Histogram extends DataCollector {

    private double m_binSize = 1.0;
    private Integer m_lowBin;
    private Integer m_highBin;
    public Hashtable m_bins = new Hashtable();

    /**
     * Default constructor creates a new histogram with a bin-size of 1.0.
     */
    public Histogram() {
        this(1.0);
    }

    /**
     * Creates a new Histogram with the specified bin size.
     */
    public Histogram(double binSize) {
        super();
        m_binSize = binSize;
    }

    /**
     * Records a new observation by incrementing the appropriate bin,
     * then invoking newObservation on the underlying DataCollector.
     */
    public void newObservation(double value) {
        recordObservation
                (new Integer((int)Math.floor(value/m_binSize)));
        super.newObservation(value);
    }

    /**
     * Records a new observation by incrementing the appropriate bin,
     * then invoking newObservation on the underlying DataCollector.
     */
    public void newObservation(long value) {
        recordObservation
                (new Integer((int)Math.floor(value/m_binSize)));
        super.newObservation(value);
    }

    /**
     * Records a new observation by incrementing the appropriate bin,
     * then invoking newObservation on the underlying DataCollector.
     */
    public void newObservation(int value) {
        recordObservation
                (new Integer((int)Math.floor(value/m_binSize)));
        super.newObservation(value);
    }

    /**
     * Records a new observation by incrementing the appropriate bin,
     * then invoking newObservation on the underlying DataCollector.
     */
    public void newObservation(Number value) {
        recordObservation
                (new Integer((int)Math.floor(value.doubleValue()/m_binSize)));
        super.newObservation(value);
    }

    /**
    Returns a count of the number of bins in the histogram, including
    empty ones, between lowBin and highBin.
    */
    public int getBinCount() {
        if (this.m_weight == 0)
            return 0;
        else
            return (m_highBin.intValue() - m_lowBin.intValue() + 1);
    }

    /**
    Returns an array of HistogramBin instances, sorted from low to high.
    the first and last bins are guaranteed to be non-null, but intermediate
    bins that are empty are represented by nulls in the array.
    */
    public HistogramBin[] getBins() {
        // Create an array of bins, sorted by base value
        HistogramBin[] sortedBins = new HistogramBin[getBinCount()];
        Enumeration unsortedBins = m_bins.elements();

        while (unsortedBins.hasMoreElements()) {
            HistogramBin nextBin = (HistogramBin)unsortedBins.nextElement();
            int binIndex = (int)Math.round(nextBin.getBaseValue()
                                           / nextBin.getBinRange()
                                           - m_lowBin.intValue());
            sortedBins[binIndex] = nextBin;
        }

        return sortedBins;
    }

    /**
    Returns the bin representing the lowest value observed.
    */
    public HistogramBin getLowBin() {
        return (HistogramBin)m_bins.get(m_lowBin);
    }

    /**
    Returns the bin representing the highest value observed.
    */
    public HistogramBin getHighBin() {
        return (HistogramBin)m_bins.get(m_highBin);
    }

    /**
    Returns the bin containing the largest number of observations.
    */
    public HistogramBin getLargestBin() {
        if (getBinCount() > 0) {
            HistogramBin hb = (HistogramBin)m_bins.get(m_lowBin);
            Enumeration theBins = m_bins.elements();

            while (theBins.hasMoreElements()) {
                HistogramBin hb1 = (HistogramBin)theBins.nextElement();
                if (hb1.getBinCount() > hb.getBinCount())
                    hb = hb1;
            }
            return hb;
        }
        else
            return null;
    }

    /**
    Returns the width of a bin.
    */
    public double getBinSize() {
        return m_binSize;
    }

    /**
    Changes the width of a bin. This should only be done when all bins
    are empty, because existing bins will not be updated.
    */
    public void setBinSize(double newSize) {
        m_binSize = newSize;
    }

    /**
    Re-initializes this histogram, deleting all bins and removing all
    observations.
    */
    public void reInitialize() {
        super.reInitialize();
        m_lowBin = null;
        m_highBin = null;
        m_bins.clear();
    }

    private void recordObservation(Integer binIndex) {
        // find the bin for this value and increment its count.
        HistogramBin hb = (HistogramBin)m_bins.get(binIndex);
        if (hb == null) {
            hb = new HistogramBin(m_binSize * binIndex.doubleValue(),
				  m_binSize);
           m_bins.put(binIndex, hb);
        }
        hb.newEntry();

        // If this is the first entry, we initialize the high and low
        // bin values, otherwise we keep track of the lowest and
        // highest yet seen. Note that this code depends on this method
        // being invoked prior to recording the observation in the
        // underlying DataCollection.

        if (m_weight == 0) {
            m_lowBin = binIndex;
            m_highBin = binIndex;
        }
        else if (binIndex.intValue() < m_lowBin.intValue())
            m_lowBin = binIndex;
        else if (binIndex.intValue() > m_highBin.intValue())
            m_highBin = binIndex;
    }
}
