/*------------------------------------------------------------------------
 * Magnet Utilities
 * $Id: StorageDataCollector.java,v 1.1.2.2 2002/05/05 19:42:06 ashutosh Exp $
 * Copyright (c) 1998 Regents of the University of Minnesota.
 * All Rights Reserved by the Regents of the University of Minnesota.
 * This file contains Original Code and/or Modifications of Original
 * Code as defined in the MAGNET Public Source License, which is
 * available at http://www.cs.umn.edu/magnet/source_license_v1.txt.
 * You may not use this file except in compliance with the License.
 *------------------------------------------------------------------------
 * Revision History:
 *
 * Wed Nov  4 14:34:25 CST 1998 Max Tsvetovat
 *   Initial entry
 */

package edu.umn.magnet.util.stats;

import java.util.Vector;

/**
 * Collects numeric observations, keeps summary statistics.
 */
public class StorageDataCollector implements Collector{

    protected double m_sum = 0.0;
    protected double m_sumsq = 0.0;
    protected long m_weight = 0;
    protected Vector data;

    /**
    Creates a new DataCollector instance.
    */
    public StorageDataCollector() {
	super();
	data=new Vector();
    }

    /**
    Returns a string representation of the summary data, giving observation
    count, mean, and standard deviation.
    */
    public String toString() {
        return "Observation Count = " + m_weight
                + ", mean = " + this.getMean()
                + ", sigma = " + this.getSigma();
    }

    /**
    Returns the mean of all observations
    */
    public double getMean() {
        if (m_weight > 0)
            return m_sum/m_weight;
        else
            return 0;
    }

    /**
    Returns the variance of the observations
    */
    public double getVariance() {
        if (m_weight > 0)
        {
            double m = getMean();
            return m_sumsq/m_weight - m*m;
        }
        else
            return 0;
    }

    /**
    Returns the standard deviation of all observations
    */
    public double getSigma() {
        return Math.sqrt(Math.abs(getVariance()));
    }

    /**
    Returns the number of observations made so far.
    */
    public long getObservationCount() {
        return m_weight;
    }

    /**
    Adds an observation, type = double
    */
    public void newObservation(double value) {
        m_sum += value;
        m_sumsq += value * value;
        m_weight++;
	data.addElement((Object)new Double(value));
    }

    /**
    Adds a weighted observation (v, r) where v is the value and
    r is the repeat count.
    */
    public void newObservation(double value, long repeatCount) {
        m_sum += value * repeatCount;
        m_sumsq += value * value * repeatCount;
        m_weight += repeatCount;
	for (int i=0; i<repeatCount; i++)
	    data.addElement((Object)new Double(value));
    }

    /**
    Adds an observation, type = long
    */
    public void newObservation(long value) {
        m_sum += value;
        m_sumsq += value * value;
        m_weight++;
	data.addElement((Object)new Long(value));
    }

    /**
    Adds an observation, type = int
    */
    public void newObservation(int value) {
        m_sum += value;
        m_sumsq += value * value;
        m_weight++;
	data.addElement((Object)new Integer(value));
    }

    /**
    Adds an observation, type = Number
    */
    public void newObservation(Number value) {
        double incr = value.doubleValue();
        m_sum += incr;
        m_sumsq += incr * incr;
        m_weight++;
	data.addElement((Object)value);
    }

    /**
    Re-initializes this DataCollector, removing all observations.
    */
    public void reInitialize() {
        m_sum = 0.0;
        m_sumsq = 0.0;
        m_weight = 0;
	data=new Vector();
    }

    public Vector getData() {
	return data;
    }

    public String toMatlabCode(String varName) {
	String matlabCode=getDataMatlab(varName+"_data") + "\n" +
	    getStatsMatlab(varName+"_stats");
	return matlabCode;
    }

    public String getDataMatlab(String matlabVar) {
	java.util.Enumeration e=data.elements();
	String matlabCode=matlabVar+"= [";
	while (e.hasMoreElements())
	    matlabCode=matlabCode+e.nextElement().toString()+";\n ";
	matlabCode=matlabCode+"];";
	return matlabCode;
    }

    public String getStatsMatlab(String matlabVar) {
	String matlabCode=matlabVar+"= ["+
	    String.valueOf(m_sum) + "; "+
	    String.valueOf(m_sumsq) + "; " +
	    String.valueOf(m_weight)+" ];";
	return matlabCode;
    }
    
    public boolean isStatOnly() {
	return(false);
    }
	

}

