/*------------------------------------------------------------------------
 * MAGNET utility test components
 * $Id: TimerTest.java,v 1.1.2.1 2002/05/05 19:42:07 ashutosh Exp $
 * Copyright (c) 2000 Regents of the University of Minnesota.
 * All Rights Reserved by the Regents of the University of Minnesota.
 * This file contains Original Code and/or Modifications of Original
 * Code as defined in the MAGNET Public Source License, which is
 * available at http://www.cs.umn.edu/magnet/source_license_v1.txt.
 * You may not use this file except in compliance with the License.
 *------------------------------------------------------------------------
 * Revision History:
 * 
 * Mon Apr  9 09:15:46 2001  John Collins
 *   Initial Entry
 */

package edu.umn.magnet.util.time;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestResult;
import junit.framework.TestSuite;

import java.util.Date;

/**
 * This is a tester for subclasses of
 * <tt>edu.umn.magnet.util.time.AbstractTimer</tt>
 * @author John Collins
 */
public class TimerTest extends TestCase
{
  static final long HOUR = 3600L * 1000L;
  static final long DAY = HOUR * 24L;
  static final Date BASE = new Date(new Date().getTime() - DAY);
  static final double RATE = 10.0; // agent clock runs 10 times faster
  static final double SLOP = 32.0; // measurement slop
  
  public TimerTest (String testName)
  {
    super(testName);
  }

  /**
   * You run the main method to exercise this thing.  There are no
   * command-line args.
   */
  public static void main (String[] args)
  {
    junit.textui.TestRunner.run(suite());
  }

  public static Test suite ()
  {
    // initialize classes to reduce timing variation
    try {
      Class foo = Class.forName("edu.umn.magnet.util.time.AbsoluteDelay");
      foo = Class.forName("edu.umn.magnet.util.time.Delay");
      foo = Class.forName("edu.umn.magnet.util.time.Timer");
    }
    catch (ClassNotFoundException cnf) {
      fail("Class problem...");
    }
    return new TestSuite(TimerTest.class);
  }

  // -------------- absolute delay --------------------
  
  public synchronized void testAbsoluteDelay ()
  {
    final TimerTest target = this;
    final long delay = 2000L;
    final long now = new Date().getTime();

    // do a short Thread.sleep() just to get classes loaded etc.
    try {
      Thread.sleep(10);
    }
    catch (InterruptedException ie) {
      fail("Unexpected Interrupt");
    }

    // now we can run the actual test
    AbstractTimer tm
	= new AbsoluteDelay (delay,
			     new TimerAction () {
				 public void action () {
				   target.processT1(now + delay);
				 }
			       });
    try {
      this.wait();
    }
    catch (InterruptedException ie) {
      fail("Unexpected interrupt");
    }
  }

  private synchronized void processT1 (long target)
  {
    long err = (long)Math.abs(target - new Date().getTime());
    try {
      assert("AbsoluteDelay error = " + err,
	     (err < SLOP));
    }
    finally {
      this.notify();
    }
  }

  // -------------- scaled delay --------------------
  
  private void setDate()
  {
    AgentTime.setDate(BASE, RATE);
  }
  
  public synchronized void testDelay ()
  {
    setDate();
    final TimerTest target = this;
    final long delay = 2000L;
    final long now = AgentTime.getDate().getTime();
    AbstractTimer tm
	= new Delay (delay,
		     new TimerAction () {
			 public void action () {
			   target.processT2(now + delay);
			 }
		       });
    try {
      this.wait();
    }
    catch (InterruptedException ie) {
      fail("Unexpected interrupt");
    }
  }

  private synchronized void processT2 (long target)
  {
    long err = (long)Math.abs(target - AgentTime.getDate().getTime());
    try {
      assert("Delay error = " + err,
	     (err < (SLOP * RATE)));
    }
    finally {
      this.notify();
    }
  }

  // -------------- test waitUntil --------------------
  
  public synchronized void testTimer ()
  {
    setDate();
    final TimerTest target = this;
    final long now = AgentTime.getDate().getTime();
    final long deadline = now + 2000L;
    AbstractTimer tm
	= new Timer (new Date(deadline),
		     new TimerAction () {
			 public void action () {
			   target.processT3(deadline);
			 }
		       });
    try {
      this.wait();
    }
    catch (InterruptedException ie) {
      fail("Unexpected interrupt");
    }
  }

  private synchronized void processT3 (long deadline)
  {
    long err = (long)Math.abs(deadline - AgentTime.getDate().getTime());
    try {
      assert("Timer error = " + err,
	     (err < (SLOP * RATE)));
    }
    finally {
      this.notify();
    }
  }

  // Other tests to try:
  // - zero and negative delays
  // - Timer in the past
  // - fractional clock rates
  // - interrupts
  // - concurrent timers
}
