/*------------------------------------------------------------------------
 * MAGNET xx component
 * $Id: ComputeMLE.java,v 1.3 2001/10/07 22:30:18 jcollins Exp $
 * Copyright (c) 2001 University of Minnesota
 *------------------------------------------------------------------------
 * Revision History:
 * 
 * Sat Sep  5 21:44:18 1998  John E. Collins
 *   Initial Entry
 */

//package edu.umn.magnet.xx;

//import edu.umn.magnet.xx;
//import org.log4j.Category;
import java.io.BufferedReader;
import java.io.FileReader;
import java.io.FileNotFoundException;
import java.util.StringTokenizer;
import java.util.NoSuchElementException;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.FileOutputStream;
import java.lang.InterruptedException;
import java.lang.Process;
import java.text.NumberFormat;
import java.util.TreeSet;
import java.util.Iterator;
import java.util.Arrays;

/**
 * This is just a quickie program to recompute MLE values for the SA
 * performance experiments
 * @author John Collins
 * @version $Revision: 1.3 $
 */
public class ComputeMLE
{
  // config parameters, hard-coded
  static int s_nTolerances = 3;
  static int s_nSearches = 200;
  static double s_defaultBinSize = 15.0;
  static double s_binSize = s_defaultBinSize;
  //static double ln2 = Math.log(2.0);
  static int s_datapoints = 50;

  public static void main (String[] args)
  {
    if (args.length < 1)
    {
      System.out.println("Usage: ComputeMLE <data file> [search-count]");
      System.exit(0);
    }

    //Open the data file
    String filename = args[0];
    BufferedReader in = null;
    try {
      in = new BufferedReader(new FileReader(filename));
    }
    catch (FileNotFoundException fnf) {
      System.out.println("Can't open file " + filename);
      System.exit(-1);
    }

    //compute the root filename
    int sep = filename.lastIndexOf('.');
    String fnroot = filename.substring(0, sep);
    //System.out.println("fnroot = " + fnroot);

    //pick off the search count, if given
    if (args.length > 1) {
      String scount = args[1];
      try {
	s_nSearches = Integer.parseInt(scount);
      }
      catch (NumberFormatException nfe) {
	System.out.println("Bad number format for search count " + args[1]);
	System.exit(-1);
      }
    }

    //for each set, read the SA data and compute the MLE
    boolean eof = false;
    String line = null;
    int lineNumber = 0;
    int runNumber = 0;
    try {
      while (!eof) {
	line = in.readLine();
	lineNumber += 1;
	if (line == null)
	  eof = true;
	else {
	  StringTokenizer st = new StringTokenizer(line, " ");
	  String token = st.nextToken();
	  if (!token.equals("IPCon"))
	    System.out.println("Bad input " + token
			       + " on line " + lineNumber);
	  runNumber += 1;
	  s_binSize = s_defaultBinSize;
	  //for each SA row, read the data
	  Calculator calc = new Calculator();
	  boolean inSA = true;
	  while (inSA) {
	    st = new StringTokenizer(in.readLine(), " ");
	    lineNumber += 1;
	    token = st.nextToken();
	    if (!token.equals("SA"))
	      // This should be the MLE token
	      inSA = false;
	    else {
	      // at this point, the current line has 3 pairs of numbers,
	      // some of which may be zero.  The first number in each
	      // pair is the search time, which we discard.  The
	      // second is the node count, which is what we're
	      // interested in.
	      for (int i = 0; i < 3; i++) {
		token = st.nextToken(); // time value - discard
		//System.out.print("Line " + lineNumber + ":" + token);
		token = st.nextToken();
		//System.out.println(":" + token + ":");
		calc.newCount(Integer.parseInt(token));
	      }
	      // move the calculator to the next row
	      calc.nextSearch();
	    }
	  }
	  //compute & report the MLE info
	  if (token.equals("Corrected")) {
	    // we have hand-corrected MLE data here.  We'll use that
	    // instead of computing it from the data
	    for (int i = 0; i < s_nTolerances; i++) {
	      // MLE data is in <alpha beta> pairs
	      token = st.nextToken();
	      calc.setWeibullAlpha(i, Double.parseDouble(token));
	      token = st.nextToken();
	      calc.setWeibullBeta(i, Double.parseDouble(token));
	    }
	  }
	  String datafileName = (fnroot + "_" + runNumber);
	  PrintWriter datafile
	      = new PrintWriter(new FileOutputStream(datafileName + ".plot"));
	  datafile.println("# Plot data for " + (fnroot + "_" + runNumber));
	  System.out.println("Summary data:");
	  for (int i = 0; i < 3; i++) {
	    System.out.print("Tolerance #" + (i + 1) + ": ");
	    System.out.print(" Alpha = ");
	    double alpha = calc.getWeibullAlpha(i);
	    System.out.print(alpha);
	    System.out.print(", Beta = ");
	    double beta = calc.getWeibullBeta(i);
	    System.out.println(beta);
	    System.out.println(" Sample: Valid = " + (int)calc.getValidRows(i)
			       + ", Mean = " + calc.getMean(i)
			       + ", Sigma = " + calc.getSigma(i));
	    System.out.println(" Distribution: Mean = " + calc.getWMean(i)
			       + ", Sigma = " + calc.getWSigma(i));
	    calc.printChiSq(i);
	    calc.dumpDensityData(i, datafile);
	    datafile.println(); datafile.println();
	    calc.dumpDistributionData(i, datafile);
	    datafile.println(); datafile.println();
	  }
	  datafile.close();
	  plotData(datafileName, calc, 3);
	  System.out.println("------");
	}
      }
    }
    catch (IOException ioe) {
      System.out.println("IO Exception");
    }
  }

  private static void plotData (String plotfile,
				Calculator calc,
				int nSets)
  {
    Process plotter = null;
    System.out.println(" Plotting " + plotfile);
    for (int i = 0; i < nSets; i++) {
      try {
	String fn = (plotfile + "_" + (i + 1));
	PrintWriter out
	    = new PrintWriter(new FileOutputStream(fn +".gp"));
	NumberFormat nf = NumberFormat.getInstance();
	//PrintWriter out = new PrintWriter(plotter.getOutputStream());
	out.println("set terminal postscript eps 24");
	out.println("# Density plot:");
	out.println("set output \"" + fn + "_dn.eps" + "\"");
	out.println("set xlabel \"Node Count / 1000\"");
	out.println("set ylabel \"Density * 1000\"");
	out.print("plot \"" + plotfile + ".plot\" index "
		  + (i * 4) + " title \"Actual\" with steps, ");
	out.print("\"" + plotfile + ".plot\" index "
		  + (i * 4 + 1) + " smooth csplines title \"Weibull(");
	nf.setMaximumFractionDigits(4);
	out.print(nf.format(calc.getWeibullAlpha(i)) + ", ");
	//nf.setMaximumFractionDigits(1);
	out.print(nf.format((long)Math.round(calc.getWeibullBeta(i))));
	out.println(")\"");

	out.println("# Distribution plot:");
	out.println("set output \"" + fn + "_ds.eps" + "\"");
	out.println("set ylabel \"Cumulative Distribution\"");
	out.println("set key left");
	out.println("set logscale x");
	out.print("plot \"" + plotfile + ".plot\" index "
		  + (i * 4 + 2) + " title \"Actual\" with points, ");
	out.print("\"" + plotfile + ".plot\" index "
		  + (i * 4 + 3) + " smooth csplines title \"Weibull(");
	nf.setMaximumFractionDigits(4);
	out.print(nf.format(calc.getWeibullAlpha(i)) + ", ");
	//nf.setMaximumFractionDigits(1);
	out.print(nf.format((long)Math.round(calc.getWeibullBeta(i))));
	out.println(")\"");
	
	out.close();
	//plotter = Runtime.getRuntime().exec("gnuplot " + fn + ".gp");
      }
      catch (IOException io) {
	System.out.println("Couldn't run gnuplot");
      }
      //finally {
      //try {
      //plotter.waitFor();
      //}
      //catch (InterruptedException ie) {
      // nothing to do here...
      //}
      //}
    }
  }

  /**
   * Inner class for monitoring search results.  This is assumed to be
   * used only when multiple search experiments are being run.
   */
  private static class Calculator
  {
    static final double EPSILON = 0.00001;
    int toleranceIndex = 0;
    int searchIndex = 0;
    CalcData[] data = null;

    // the results array is indexed first by tolerance values, and
    // second by search number.

    Calculator ()
    {
      if (s_nSearches > 1) {
	data = new CalcData[s_nTolerances];
	for (int i = 0; i < s_nTolerances; i++) {
	  data[i] = new CalcData(s_nSearches);
	}
      }
    }

    /**
     * Handles the input data
     * when a new search result is detected.
     */
    public void newCount (int nodeCount)
    {
      data[toleranceIndex++].results[searchIndex] = (double)nodeCount;
    }

    /**
     * Sets up for the next search.  This must be called regardless of
     * how the search ended.
     */
    void nextSearch ()
    {
      toleranceIndex = 0;
      searchIndex += 1;
    }

    /**
     * Ensures that the Maximum Likelihood Estimators have been
     * computed for a Weibull
     * distribution to match the node-count values in the results
     * arrays.  The output is stored in the weibullMLE array.
     */
    private void ensureWeibullMLE ()
    {
      if (data[0].weibullAlpha > 0.0)
	return;
      // compute alpha and beta values
      //weibullAlpha = new double[s_nTolerances];
      //weibullBeta = new double[s_nTolerances];
      //validRows = new double[s_nTolerances];
      for (int i = 0; i < s_nTolerances; i++) {
	//log.debug("Computing MLE values for tolerance " + s_tolerance[i]);
	data[i].weibullAlpha = computeWAlpha(data[i].results);
	data[i].weibullBeta = computeWBeta(data[i].results,
					   data[i].weibullAlpha);
      }
    }

    /**
     * Computes the Weibull Alpha value for an array of ints.  The
     * computation involves an initial estimate followed by a
     * relaxation based on Newton's method.  There is lots of summing
     * of the array and summing of elements of the array to the power
     * of the current estimate.  The source is Law & Kelton,
     * Simulation Modeling and Analysis, McGraw-Hill 1991, pp 334.
     */
    private double computeWAlpha (double[] values)
    {
      double n = 0.0;
      // first the initial estimate:
      double sumlog = 0.0;
      double sumlogsq = 0.0;
      for (int i = 0; i < values.length; i++) {
	if (values[i] > EPSILON) {
	  n += 1.0;
	  double vlog = Math.log(values[i]);
	  sumlog += vlog;
	  sumlogsq += vlog * vlog;
	}
      }
      double num = ((sumlogsq - (sumlog * sumlog / n))
		    * (6.0 / (Math.PI * Math.PI)));
      double alphak1 = Math.pow((num / (n - 1.0)), -0.5);
      //log.debug("initial Alpha: " + alphak1);
      
      // At this point, we have the initial alpha estimate.  Now
      // starts the fixed-point calculation.
      double alphak = alphak1 * 2;
      double a = sumlog / n;
      while (Math.abs(1.0 - (alphak / alphak1)) > EPSILON) {
	alphak = alphak1;
	double b = 0.0;
	double c = 0.0;
	double h = 0.0;
	for (int i = 0; i < values.length; i++) {
	  if (values[i] > EPSILON) {
	    double vlog = Math.log(values[i]);
	    double xalpha = Math.pow(values[i], alphak);
	    b += xalpha;
	    c += xalpha * vlog;
	    h += xalpha * vlog * vlog;
	  }
	}
	alphak1 = (alphak
		   + ((a + (1.0 / alphak) - (c / b))
		      / ((1.0 / (alphak * alphak))
			 + ((b * h - c * c) / (b * b)))));
      }
      //log.debug("final Alpha: " + alphak1);
      return alphak1;
    }

    /**
     * Computes the Weibull Beta value for the given array and alpha
     * value.
     */
    private double computeWBeta (double[] values, double alpha)
    {
      double n = 0.0;
      double beta = 0.0;
      double sumx = 0.0;
      for (int i = 0; i < values.length; i++) {
	if (values[i] > EPSILON) {
	  n += 1.0;
	  sumx += Math.pow(values[i], alpha);
	}
      }
      beta = Math.pow((sumx / n), (1.0 / alpha));
      //log.debug("final Beta: " + beta);
      return beta;
    }

    /**
     * Indexed setter for a Weibull alpha value.
     */
    public void setWeibullAlpha (int index, double value)
    {
      data[index].weibullAlpha = value;
    }

    /**
     * Indexed getter for a Weibull alpha value.
     */
    public double getWeibullAlpha (int index)
    {
      ensureWeibullMLE();
      return data[index].weibullAlpha;
    }

    /**
     * Indexed setter for a Weibull beta value.
     */
    public void setWeibullBeta (int index, double value)
    {
      data[index].weibullBeta = value;
    }

    /**
     * Indexed getter for a Weibull beta value.
     */
    public double getWeibullBeta (int index)
    {
      ensureWeibullMLE();
      return data[index].weibullBeta;
    }

    /**
     * Ensure that Weibull bins have been computed
     */
    public void ensureWeibullBins ()
    {
      if (data[0].bins != null)
	return;

      ensureWeibullMLE();
      for (int k = 0; k < s_nTolerances; k++) {
	CalcData d = data[k];
	double[] values = d.results;
	double n = d.getValidRows();

	// If the observation count is less than twice the bin size,
	// make the bin count half the observation count.
	if (n > 1 && n < (s_binSize * 2))
	  s_binSize = (n+1)/2;

	// compute number of bins and bin boundaries
	d.binCount = (int)Math.ceil(n / s_binSize);
	d.binBounds = new double[d.binCount + 1];
	d.bins = new double[d.binCount];
	d.binBounds[0] = 0.0;
	// the total area under the histogram is the ratio of the
	// solved attempts to total attempts
	double integral = (double)d.getValidRows() / (double)d.results.length;
	for (int i = 1; i <= d.binCount; i++) {
	  double f = (double)i * (integral / (double)(d.binCount));
	  d.binBounds[i] = (getWeibullBeta(k)
			     * Math.pow(- Math.log(1.0 - f),
					(1.0 / getWeibullAlpha(k))));
	  //System.out.println("Bound[" + k + "][" + i + "] = "
	  //		     + d.binBounds[i]);
	}
	
	// count observations/bin
	for (int i = 0; i < values.length; i++) {
	  if (values[i] > EPSILON) {
	    for (int j = d.binCount - 1; j >= 0; j--) {
	      if (values[i] > d.binBounds[j]) {
		d.bins[j] += 1.0;
		break;
	      }
	    }
	  }
	}
      }
    }

    /**
     * Indexed getter for Weibull bin data
     */
    public double[] getWeibullBins (int index)
    {
      ensureWeibullBins();
      double[] result = data[index].bins;
      for (int j = 0; j < result.length; j++) {
	System.out.print(" " + result[j]);
      }
      System.out.println("");
      return result;
    }

    /**
     * Indexed getter for Weibull chi-squared data
     */
    public double printChiSq (int index)
    {
      ensureWeibullBins();
      double[] values = data[index].results;
      double expected = 0.0;
      double result = 0.0;
      double[] bins = getWeibullBins(index);
      //expected = ((double)(data[index].binCount)) / ((double)(bins.length));
      expected = (data[index].getValidRows()
		  / ((double)(data[index].binCount)));
      for (int i = 0; i < bins.length; i++) {
	double diff = expected - bins[i];
	result += (diff * diff) / expected;
      }
      System.out.println (" ChiSq: dof = " + (bins.length - 1)
			  + ", value = " + result);
      return result;
    }

    public double getValidRows (int index)
    {
      return data[index].getValidRows();
    }

    public double getMean (int index)
    {
      return data[index].getMean();
    }

    public double getSigma (int index)
    {
      return data[index].getSigma();
    }

    public double getWMean (int index)
    {
      return data[index].getWMean();
    }

    public double getWSigma (int index)
    {
      return data[index].getWSigma();
    }

    public void dumpBinData (int i, String filename)
    {
      data[i].dumpBinData(filename);
    }

    public void dumpDensityData (int i, PrintWriter file)
    {
      file.println("# Density " + i);
      data[i].dumpDensityData(file);
    }

    public void dumpDistributionData (int i, PrintWriter file)
    {
      file.println("# Distribution " + i);
      data[i].dumpDistributionData(file);
    }

    static class CalcData
    {
      double[] results = null;
      double sum = 0.0;
      double sumsq = 0.0;
      double weibullAlpha = 0.0;
      double weibullBeta = 0.0;
      double validRows = 0.0;
      int binCount = 0;
      double[] bins = null;
      double[] binBounds = null;

      CalcData (int nSearches)
      {
	results = new double[nSearches];
      }

      double getValidRows ()
      {
	if (validRows == 0.0)
	{
	  // count the non-zero rows
	  double n = 0.0;
	  for (int i = 0; i < results.length; i++) {
	    if (results[i] > EPSILON) {
	      n += 1.0;
	      sum += results[i];
	      sumsq += results[i] * results[i];
	    }
	  }
	  validRows = n;
	}
	return validRows;
      }

      /**
       * Returns the sample mean
       */
      double getMean ()
      {
	// gotta do this in the right order, since getValidRows does
	// some lazy evaluation.
	double n = getValidRows();
	return (sum / n);
      }

      /**
       * Returns the standard deviation of the sample
       */
      double getSigma ()
      {
	double m = getMean();
	double var = (sumsq / getValidRows()) - ( m * m);
	return Math.sqrt(var);
      }

      /**
       * Returns the mean of the inferred distribution
       */
      double getWMean ()
      {
	return ((weibullBeta / weibullAlpha)
		* gamma(1.0 / weibullAlpha));
      }

      /**
       * Returns the standard deviation of the inferred distribution
       */
      double getWSigma ()
      {
	double ga = gamma(1.0 / weibullAlpha);
	return Math.sqrt((weibullBeta * weibullBeta / weibullAlpha)
			 * (2.0 * gamma(2.0 / weibullAlpha)
			    - ga * ga / weibullAlpha));
      }

      void dumpBinData (String filename)
      {
	try {
	  PrintWriter out = new PrintWriter(new FileOutputStream(filename));
	  double expected = (getValidRows() / (double)binCount);
	  for (int i = 0; i < binCount; i++) {
	    out.print("" + (binBounds[i] / 1000.0));
	    double width = (binBounds[i + 1] - binBounds[i]);
	    out.print(" " + (100.0 * expected / width));
	    out.println(" " + (100.0 * bins[i] / width));
	  }
	  out.close();
	}
	catch (FileNotFoundException fnf) {
	  System.out.println("Failed to create file " + filename);
	}
      }

      /**
       * Dumps two datasets, separated by a double blank line.  The
       * first is input density data, the second is a higher density
       * set suitable for plotting a smooth density curve.
       */
      void dumpDensityData (PrintWriter file)
      {
	double expected = (getValidRows() / (double)binCount);
	if (binCount > 1) {
	  for (int i = 0; i < binCount; i++) {
	    file.print("" + (binBounds[i] / 1000.0));
	    double width = (binBounds[i + 1] - binBounds[i]);
	    //file.print(" " + (100.0 * expected / width));
	    //file.println(" " + ((100.0 / expected) * bins[i] / width));
	    file.println(" " + ((1000.0 / results.length)
				* (bins[i] / width)));
	  }
	  file.println(); file.println();
	  double max = binBounds[binCount - 1];
	  double increment = max / (double)s_datapoints;
	  for (double pt = increment; pt < max; pt += increment) {
	    file.print("" + (pt / 1000.0));
	    file.println(" " + (weibullAlpha * 1000.0
				* Math.pow(weibullBeta, -weibullAlpha)
				* Math.pow(pt, (weibullAlpha - 1.0))
				* Math.pow(Math.E,
					   -Math.pow((pt / weibullBeta),
						     weibullAlpha))));
	  }
	}
	else {
	  file.println("0.0 0.0");
	  file.println(); file.println();
	  file.println("0.0 0.0");
	}
      }

      void dumpDistributionData (PrintWriter file)
      {
	double points[] = new double[(int)getValidRows()];
	int i = 0;
	for (int j = 0; j < results.length; j++) {
	  if (results[j] > EPSILON)
	    points[i++] = results[j];
	}
	Arrays.sort(points);
	for (i = 0; i < points.length; i++) {
	  file.println("" + (points[i] / 1000.0)
		       + " " + ((double)i / results.length));
	}
	file.println(); file.println();
	if (binCount > 1) {
	  double pt = points[0];
	  double max = points[points.length - 1];
	  double increment = max / (double)s_datapoints;
	  double value;
	  //file.println("0.0 0.0");
	  //for (double pt = min; pt < max; pt += increment) {
	  do {
	    value = (1.0 - Math.pow(Math.E,
				    -Math.pow((pt / weibullBeta),
					      weibullAlpha)));
	    file.print("" + (pt / 1000.0));
	    file.println(" " + value);
	    pt += increment;
	  } while (value < 0.99);
	}
	else
	  file.println("0.0 0.0");
      }

      /**
       * Computes the gamma function of its (positive) argument
       */
      double gamma(double x)
      {
        double[] cof = new double[6];
        cof[0] = 76.18009172947146;
        cof[1] = -86.50532032941677;
        cof[2] = 24.01409824083091;
        cof[3] = -1.231739572450155;
        cof[4] = 0.1208650973866179e-2;
        cof[5] = -0.5395239384953e-5;
        double xx = x;
        double yy = x;
        double tmp = xx + 5.5 - (xx + 0.5) * Math.log(xx + 5.5);
        double ser = 1.000000000190015;
        for(int i = 0; i <= 5; i++) {
	  yy += 1.0;
          ser += (cof[i] / yy);
	}
        double lng = (Math.log(2.5066282746310005 * ser / xx) - tmp);
	return Math.pow(Math.E, lng);
      }
    }
  }
}
